// request.h : header file
// declarations of the Nirva client request classes
// A request maintains a client context for sending and receiving commands

#ifndef _REQUEST_H
#define _REQUEST_H

#ifdef WIN32
#include "..\nvs\task.h"
#include "..\nvs\container.h"
#include "..\nvs\obj.h"
#include "..\nvc\error.h"
#include "..\nvc\command.h"
#else
#include "task.h"
#include "container.h"
#include "obj.h"
#include "error.h"
#include "command.h"
#endif
#include <string>

int NvCommand2(void* Request, const char* Command, char* Buffer, int BufferSize, int *DataSize, int iEncoding = ENCODING_NONE);


/////////////////////////////////////////////////////////////////////////////
// NvRequest class
class NvRequest : public NvObject
{
public:

	NvError GlobalError;				// Global Error object that has to be tested after each command (except the command that gets the error code)

	std::string m_sServer;				// Server address and port
	std::string m_sSocketLog;			// socket log file
	std::string m_sUser;				// User name
	std::string m_sPassword;			// password
	std::string m_sNewPassword;			// optional new password
	std::string m_sNewPasswordConfirm;	// optional new password confirm
	std::string m_sAppli;				// Application name
	std::string m_sOpen;				// Session open procedure
	std::string m_sClose;				// Session close procedure
	std::string m_sCertificate;			// Client certificate (ssl only)
	std::string m_sHash;				// Used for security control
	std::string m_sTempDir;				// Temporary directory
	int m_iMaxMessMemSize;				// Maximum memory for a message
	bool b_UseSSL;
	bool b_UseSSO;
	std::string SSOPrincipal;			// SSO principal name
	bool b_ClientUnicode;				// true when client is unicode
	bool bWithNewPassword;				// Do we have to set a new password during login or not
	bool bWithNewPasswordConfirm;		// Set to true when a new password confirm entry has been given

	std::string m_sSessionId;			// Session ID. If the Session ID is provided, the order tries to connect the given session
										// The session ID is also updated after an order successfully transmitted to the server
										// and also with successfull return (eventually with error)
	std::string m_sServerVersion;		// ServerVersion (fo v 4.7.010 minimum)

	void *CurrentConnection;			// Current connection
	bool m_bAutoClose;					// If set to yes, a close session will follow an order
	
	int m_iSessionTimeOut;
	int m_iServerConnectionTimeOut;

	NvLocalContainer m_cContainer;		// Local container
	NvObjIndStringList m_oVariables;	// Local variables

	std::vector<void *> m_vCommands;	// command list
	std::string m_sLastOutputBuffer;	// last output buffer

	NvRequest();
	virtual ~NvRequest();

	void Clear();
	void GetLastOutputBuffer(char *Buffer, int BufferSize);

	// Global request command entry point
	bool NvCommand(const char *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	
	// Local command entry point (called by NvCommand)
	bool ProcessLocalCommand(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	
	// Server command entry point (called by NvCommand)
	bool ProcessServerCommand(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	
	
	// Internal functions
	void GetStringValue(NvServerCommand *Command, const char *String, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	bool SendServerCommands(bool StopOnError = false, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	NvObj* GetObjectCheckType(const char *ObjName, int ObjType = -1);


	// Local commands
	
	
	// *********************
	// *** REQUEST CLASS ***
	// *********************
	//REQUEST RESET_COMMAND_BUFFER
	void ResetCommandBuffer();
	//REQUEST SEND_COMMANDS
	bool SendCommands(NvServerCommand *Command);
	//REQUEST GET_NUM_COMMANDS
	void GetNumCommands(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//REQUEST REMOVE_COMMAND
	bool RemoveCommand(NvServerCommand *Command);
	//REQUEST ERROR_INFO
	void GetError(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//REQUEST GET_SESSION_ID
	void GetSessionId(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//REQUEST GET_SERVER_VERSION
	void GetServerVersion(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//REQUEST SET_VARIABLE
	void SetVariable(NvServerCommand *Command);
	//REQUEST GET_VARIABLE
	void GetVariable(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//REQUEST REMOVE_VARIABLE
	void RemoveVariable(NvServerCommand *Command);
	//REQUEST VARIABLE EXIST
	void VariableExist(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//REQUEST XML_COMMAND
	bool XmlCommand(NvServerCommand *Command);

	// ********************
	// *** OBJECT CLASS ***
	// ********************
	//OBJECT CREATE
	bool CreateObject(NvServerCommand *Command);
	//OBJECT GET_NUM_OBJECTS
	void GetNumObjects(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT GET_NAME
	bool GetObjectName(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT GET_TYPE
	bool GetObjectType(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT CLEAR_ALL
	void ClearLocalContainer();
	//OBJECT COPY
	bool CopyObject(NvServerCommand *Command);
	//OBJECT SET_NAME
	bool SetObjectName(NvServerCommand *Command);
	//OBJECT REMOVE
	bool RemoveObject(NvServerCommand *Command);
	//OBJECT EXIST
	void IsObjectExist(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);

	// ****************************
	// *** OBJECT CLASS BOOLEAN ***
	// ****************************
	//OBJECT BOOLEAN_SET_VALUE
	bool ObjBooleanSetValue(NvServerCommand *Command);
	//OBJECT BOOLEAN_GET_VALUE
	bool ObjBooleanGetValue(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);

	// ****************************
	// *** OBJECT CLASS INTEGER ***
	// ****************************
	//OBJECT INTEGER_SET_VALUE
	bool ObjIntegerSetValue(NvServerCommand *Command);
	//OBJECT INTEGER_GET_VALUE
	bool ObjIntegerGetValue(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);

	// ***************************
	// *** OBJECT CLASS STRING ***
	// ***************************
	//OBJECT STRING_SET_VALUE
	bool ObjStringSetValue(NvServerCommand *Command);
	//OBJECT STRING_GET_VALUE
	bool ObjStringGetValue(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);

	// *******************************
	// *** OBJECT CLASS STRINGLIST ***
	// *******************************
	//OBJECT STRINGLIST_SET_VALUE
	bool ObjStringListSetValue(NvServerCommand *Command);
	//OBJECT STRINGLIST_GET_VALUE
	bool ObjStringListGetValue(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT STRINGLIST_GET_SIZE
	bool ObjStringListGetSize(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT STRINGLIST_REMOVE
	bool ObjStringListRemove(NvServerCommand *Command);
	//OBJECT STRINGLIST_INSERT
	bool ObjStringListInsert(NvServerCommand *Command);
	//OBJECT STRINGLIST_ADD_STRINGLIST
	bool ObjStringListAddStringList(NvServerCommand *Command);
	//OBJECT STRINGLIST_SWAP
	bool ObjStringListSwap(NvServerCommand *Command);
	//OBJECT STRINGLIST_SORT
	bool ObjStringListSort(NvServerCommand *Command);
	//OBJECT STRINGLIST_GET_VALUES
	bool ObjStringListGetValues(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT STRINGLIST_SEARCH
	bool ObjStringListSearch(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);

	// **********************************
	// *** OBJECT CLASS INDSTRINGLIST ***
	// **********************************
	//OBJECT INDSTRINGLIST_SET_VALUE
	bool ObjIndStringListSetValue(NvServerCommand *Command);
	//OBJECT INDSTRINGLIST_GET_VALUE
	bool ObjIndStringListGetValue(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT INDSTRINGLIST_GET_VALUES
	bool ObjIndStringListGetValues(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT INDSTRINGLIST_GET_SIZE
	bool ObjIndStringListGetSize(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT INDSTRINGLIST_REMOVE
	bool ObjIndStringListRemove(NvServerCommand *Command);
	//OBJECT INDSTRINGLIST_ADD_INDSTRINGLIST
	bool ObjIndStringListAddIndStringList(NvServerCommand *Command);
	//OBJECT INDSTRINGLIST_KEY_EXIST
	bool ObjIndStringListKeyExist(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT INDSTRINGLIST_GET_KEY
	bool ObjIndStringListGetKey(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);

	// **************************
	// *** OBJECT CLASS TABLE ***
	// **************************
	//OBJECT TABLE_SET_DESCRIPTION
	bool ObjTableSetDescription(NvServerCommand *Command);
	//OBJECT TABLE_GET_DESCRIPTION
	bool ObjTableGetDescription(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_CLEAR
	bool ObjTableClear(NvServerCommand *Command);
	//OBJECT TABLE_CLEAR_DATA
	bool ObjTableClearData(NvServerCommand *Command);
	//OBJECT TABLE_SEARCH
	bool ObjTableSearch(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_SORT
	bool ObjTableSort(NvServerCommand *Command);
	//OBJECT TABLE_CREATE_FROM
	bool ObjTableCreateFrom(NvServerCommand *Command);
	//OBJECT TABLE_SAVE
	bool ObjTableSave(NvServerCommand *Command);
	//OBJECT TABLE_LOAD
	bool ObjTableLoad(NvServerCommand *Command);
	//OBJECT TABLE_IMPORT
	bool ObjTableImport(NvServerCommand *Command);
	//OBJECT TABLE_EXPORT
	bool ObjTableExport(NvServerCommand *Command);
	//OBJECT TABLE_SELECT_FROM
	bool ObjTableSelectFrom(NvServerCommand *Command);
	//OBJECT TABLE_JOIN
	bool ObjTableJoin(NvServerCommand *Command);
	//OBJECT TABLE_ADD_TABLE
	bool ObjTableAddTable(NvServerCommand *Command);
	//OBJECT TABLE_CLEAR_CELL
	bool ObjTableClearCell(NvServerCommand *Command);
	//OBJECT TABLE_SORT_CELL
	bool ObjTableSortCell(NvServerCommand *Command);
	//OBJECT TABLE_GET_CELL_NUM_LINES
	bool ObjTableGetCellNumLines(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_GET_CELL_LINE
	bool ObjTableGetCellLine(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_SET_CELL_LINE
	bool ObjTableSetCellLine(NvServerCommand *Command);
	//OBJECT TABLE_GET_CELL
	bool ObjTableGetCell(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_SET_CELL
	bool ObjTableSetCell(NvServerCommand *Command);
	//OBJECT TABLE_INSERT_CELL_LINE
	bool ObjTableInsertCellLine(NvServerCommand *Command);
	//OBJECT TABLE_REMOVE_CELL_LINE
	bool ObjTableRemoveCellLine(NvServerCommand *Command);
	//OBJECT TABLE_SWAP_CELL_LINES
	bool ObjTableSwapCellLines(NvServerCommand *Command);
	//OBJECT TABLE_GET_NUM_ROWS
	bool ObjTableGetNumRows(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_REMOVE_ROW
	bool ObjTableRemoveRow(NvServerCommand *Command);
	//OBJECT TABLE_REMOVE_ROWS
	bool ObjTableRemoveRows(NvServerCommand *Command);
	//OBJECT TABLE_CLEAR_ROW
	bool ObjTableClearRow(NvServerCommand *Command);
	//OBJECT TABLE_CLEAR_ROWS
	bool ObjTableClearRows(NvServerCommand *Command);
	//OBJECT TABLE_INSERT_ROWS
	bool ObjTableInsertRows(NvServerCommand *Command);
	//OBJECT TABLE_SWAP_ROWS
	bool ObjTableSwapRows(NvServerCommand *Command);
	//OBJECT TABLE_GET_ROWS
	bool ObjTableGetRows(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_GET_ROW
	bool ObjTableGetRow(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_GET_COLUMN
	bool ObjTableGetColumn(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_SET_COLUMN
	bool ObjTableSetColumn(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_GET_NUM_COLUMNS
	bool ObjTableGetNumColums(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_GET_COLUMN_NAME
	bool ObjTableGetColumnName(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_GET_COLUMN_DESCRIPTION
	bool ObjTableGetColumnDescription(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_IS_COLUMN_NUMERIC
	bool ObjTableIsColumnNumeric(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_GET_COLUMN_INDEX
	bool ObjTableGetColumnIndex(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_CLEAR_COLUMN
	bool ObjTableClearColumn(NvServerCommand *Command);
	//OBJECT TABLE_REMOVE_COLUMN
	bool ObjTableRemoveColumn(NvServerCommand *Command);
	//OBJECT TABLE_INSERT_COLUMN
	bool ObjTableInsertColumn(NvServerCommand *Command);
	//OBJECT TABLE_ADD_COLUMNS
	bool ObjTableAddColumns(NvServerCommand *Command);
	//OBJECT TABLE_MODIFY_COLUMN
	bool ObjTableModifyColumn(NvServerCommand *Command);
	//OBJECT TABLE_SWAP_COLUMNS
	bool ObjTableSwapColumns(NvServerCommand *Command);
	//OBJECT TABLE_FILTER_COLUMNS
	bool ObjTableFilterColumns(NvServerCommand *Command);
	//OBJECT TABLE_SET_PRIMARY_COLUMN
	bool ObjTableSetPrimaryColumn(NvServerCommand *Command);
	//OBJECT TABLE_SET_COLUMN_NAME
	bool ObjTableSetColumnName(NvServerCommand *Command);
	//OBJECT TABLE_SET_COLUMN_TYPE
	bool ObjTableSetColumnType(NvServerCommand *Command);
	//OBJECT TABLE_SET_COLUMN_DESCRIPTION
	bool ObjTableSetColumnDescription(NvServerCommand *Command);
	//OBJECT TABLE_GET_PRIMARY_INFO
	bool ObjTableGetPrimaryInfo(NvServerCommand *Command);
	//OBJECT TABLE_GET_ROW_INDEX
	bool ObjTableGetRowIndex(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT TABLE_SET_ROW
	bool ObjTableSetRow(NvServerCommand *Command);
	//OBJECT TABLE_ADD_ROWS
	bool ObjTableAddRows(NvServerCommand *Command);

	// *************************
	// *** OBJECT CLASS FILE ***
	// *************************
	//OBJECT FILE_SET_FILENAME
	bool ObjFileSetFileName(NvServerCommand *Command);
	//OBJECT FILE_SET_PERSIST
	bool ObjFileSetPersist(NvServerCommand *Command);
	//OBJECT FILE_GET_PATHNAME
	bool ObjFileGetPathName(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT FILE_GET_DIRNAME
	bool ObjFileGetDirName(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT FILE_GET_FILENAME
	bool ObjFileGetFileName(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT FILE_GET_EXTENSION
	bool ObjFileGetExtension(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT FILE_GET_SIZE
	bool ObjFileGetSize(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT FILE_REMOVE
	bool ObjFileRemove(NvServerCommand *Command);
	//OBJECT FILE_CREATE
	bool ObjFileCreate(NvServerCommand *Command);
	//OBJECT FILE_CLEAR
	bool ObjFileClear(NvServerCommand *Command);
	//OBJECT FILE_APPEND
	bool ObjFileAppend(NvServerCommand *Command);
	//OBJECT FILE_EXIST
	bool ObjFileExist(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT FILE_SET_ORIGIN
	bool ObjFileSetOrigin(NvServerCommand *Command);
	//OBJECT FILE_GET_ORIGIN
	bool ObjFileGetOrigin(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT FILE_SET_ASCII
	bool ObjFileSetAscii(NvServerCommand *Command);
	//OBJECT FILE_COMPRESS
	bool ObjFileCompress(NvServerCommand *Command);
	//OBJECT FILE_DECOMPRESS
	bool ObjFileDecompress(NvServerCommand *Command);

	// ***************************
	// *** OBJECT CLASS BINARY ***
	// ***************************
	//OBJECT BINARY_GET_SIZE
	bool ObjBinaryGetSize(NvServerCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, int iEncoding = ENCODING_NONE);
	//OBJECT BINARY_CLEAR
	bool ObjBinaryClear(NvServerCommand *Command);
	//OBJECT BINARY_APPEND
	bool ObjBinaryAppend(NvServerCommand *Command);
	//OBJECT BINARY_SAVE
	bool ObjBinarySave(NvServerCommand *Command);
	//OBJECT BINARY_LOAD
	bool ObjBinaryLoad(NvServerCommand *Command);
};


/////////////////////////////////////////////////////////////////////////////
// NvRequestList class
class NvRequestList : public NvObject
{

public:
	std::vector<void *> m_vRequests;			// request list
	
	NvRequestList();
	virtual ~NvRequestList();

	// Removes all requests from the list
	void Clear();
	
	// Creates a new request and add it to the list
	NvRequest* AddRequest();
	
	// Remove the given request from the list
	void RemoveRequest(NvRequest *Request);
	
	// Check if a given request is in the list
	bool CheckRequest(NvRequest *Request);
};




#endif //_REQUEST_H
