// nvcc.cpp : Defines the entry point for the console application.
//

#ifdef WIN32
#define _CRT_NONSTD_NO_DEPRECATE
#endif
#ifndef WIN32
#include <pthread.h>
#endif
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifndef MAC
#include <malloc.h>
#endif
#include <sys/types.h>
#ifdef WIN32
#include <windows.h>
#include <io.h>
#else
#include <unistd.h>
#include <ctype.h>
#endif
#include "../nvc/nvc.h"
#include "../nvs/task.h"
#include "util.h"


using namespace std;


#define NV_VERSION	"2.00"


// Local functions
int GetCommandLineOption(int argc, char* argv[], char option, int more);
int GetParametersFromCommandLine(int argc, char* argv[]);
void DisplayHelp();
void Cleanup(NVREQUEST Request);
bool CheckError(NVREQUEST Request, int Result);
int RunTestMode(int argc, char* argv[]);
void ReadLine(char *Buffer, int size);
#ifndef WIN32
unsigned int GetMilliSeconds();
unsigned int GetSeconds();
unsigned int InitialSeconds = 0;
#endif


class TestRequest : public NvTask
{
// Attributes
public:
	vector<string> m_CommandParameters;		// Command parameters
	string m_InputFile;
	string m_Procedure;
	string m_Server;
	string m_Certificate;
	string m_SessionId;
	string m_Application;
	string m_User;
	string m_Password;
	string m_NewPassword;
	string m_Open;
	string m_Close;
	bool m_DisplayCommands;
	bool m_DisplayErrors;
	bool m_DisplayPrint;
	bool m_CloseSession;
	bool m_Unicode;
	bool m_StopOnError;
	int m_NbLoops;
	NVREQUEST Request;

// Overridable callbacks
protected:
	virtual int Main();
	virtual void OnException(std::string sInfo);
	virtual bool OnInit();
	virtual void OnCleanup();

public:
	void SetFinished();
	void SetConnected();
	void Cleanup();
	bool CheckError(int Result);
	int Connect();
};


// Global variables
vector<string> CommandParameters;		// Command parameters
char OptionBuffer[2048];				// Contains option value
string InputFile;
string Procedure;
string DirectCommand;
string Server;
string Certificate;
string SessionId;
string Application;
string User;
string Password;
string Open;
string Close;
bool DisplayCommands;
bool DisplayErrors;
bool DisplayPrint;
bool CloseSession;
bool Unicode;
bool StopOnError;

// Global variables for test mode
string TestOptions;
NvMutex ProtectFinished;
NvMutex ProtectDeleted;
NvMutex ProtectConnected;
int Finished;
int Deleted;
int Connected;
NvEvent Tstart;
NvEvent Tend;


int main(int argc, char* argv[])
{
	char Buffer[2048];		// General buffer
	
	InputFile = "input.txt";
	Procedure = "";
	DirectCommand = "";
	Server = "127.0.0.1";
	Certificate = "";
	SessionId = "";
	Application = "";
	User = "";
	Password = "";
	Open = "";
	Close = "";
	TestOptions = "";
	DisplayCommands = false;
	DisplayErrors = true;
	DisplayPrint = true;
	CloseSession = true;
	StopOnError = true;
	Unicode = false;

	if((argc == 2) && (strlen(argv[1]) >= 2))
	{
		strcpy(Buffer, argv[1]);
		if((Buffer[0] == '-') && (Buffer[1] == 'h'))
		{
			DisplayHelp();
			return 0;
		}
	}
	
	// Check for display mode
	int Error = GetCommandLineOption(argc, argv, 'd', 1);
	if(Error == 0)
	{
		string Result = OptionBuffer;
		MakeUpper(Result);
		if(Result.find('C') != string::npos)
			DisplayCommands = true;
		DisplayErrors = false;
		if(Result.find('E') != string::npos)
			DisplayErrors = true;
		DisplayPrint = false;
		if(Result.find('P') != string::npos)
			DisplayPrint = true;
		if(Result.find('N') != string::npos)	// silent mode
		{
			DisplayCommands = false;
			DisplayErrors = false;
			DisplayPrint = false;
		}
	}


	// Check for server address
	Error = GetCommandLineOption(argc, argv, 'a', 1);
	if(Error == 0)
		Server = OptionBuffer;


	// Check for session ID
	Error = GetCommandLineOption(argc, argv, 's', 1);
	if(Error == 0)
		SessionId = OptionBuffer;


	// Check for application
	Error = GetCommandLineOption(argc, argv, 'p', 1);
	if(Error == 0)
		Application = OptionBuffer;


	// Check for user
	Error = GetCommandLineOption(argc, argv, 'u', 1);
	if(Error == 0)
		User = OptionBuffer;


	// Check for password
	Error = GetCommandLineOption(argc, argv, 'w', 1);
	if(Error == 0)
		Password = OptionBuffer;


	// Check for session open and close procedures
	Error = GetCommandLineOption(argc, argv, 'o', 1);
	if(Error == 0)
	{
		Open = OptionBuffer;
		string::size_type index = Open.find('!');
		if(index != string::npos)
		{
			Close = Open.substr(index+1);
			Open = Open.substr(0, index);
		}
		CleanString(Open);
		CleanString(Close);
	}
	
	// Check for noclose option
	Error = GetCommandLineOption(argc, argv, 'c', 0);
	if(Error == 0)
		CloseSession = false;

	
	// Check for unicode option
	Error = GetCommandLineOption(argc, argv, 'n', 0);
	if(Error == 0)
		Unicode = false;
	
	// Check for no stop on error mode
	Error = GetCommandLineOption(argc, argv, 't', 0);
	if(Error == 0)
		StopOnError = false;

	// Check for input file
	Error = GetCommandLineOption(argc, argv, 'i', 1);
	if(Error == 0)
		InputFile = OptionBuffer;

	// Check for procedure
	Error = GetCommandLineOption(argc, argv, 'r', 1);
	if(Error == 0)
		Procedure = OptionBuffer;
	if(Procedure.length())
		InputFile = "";	// Procedure has priority on input file

	// Check for certificate
	Error = GetCommandLineOption(argc, argv, 'k', 1);
	if(Error == 0)
		Certificate = OptionBuffer;

	// Check for direct command
	Error = GetCommandLineOption(argc, argv, 'z', 1);
	if(Error == 0)
		DirectCommand = OptionBuffer;
	if(DirectCommand.length())
	{
		InputFile = "";	// Procedure has priority on input file
		Procedure = "";	// Procedure has priority on procedure
	}

	if((!InputFile.length()) && (!Procedure.length()) && (!DirectCommand.length()))
	{
		if(DisplayErrors)
			printf("\nNo input file, procedure or command given");
		return -1;
	}

	// Check for test mode
	Error = GetCommandLineOption(argc, argv, 'b', 1);
	if(Error == 0)
		TestOptions = OptionBuffer;

	Error = GetParametersFromCommandLine(argc, argv);
	if(Error != 0)
	{
		if(DisplayErrors)
			printf("\nCannot get command line parameters");
		return -1;
	}

	if(TestOptions.length())
		return RunTestMode(argc, argv);

	FILE *stream = NULL;
	if(InputFile.length())
	{
		stream = fopen(InputFile.c_str(), "rb");
		if(stream == NULL)
		{
			if(DisplayErrors)
				printf("\nCannot open input file");
			return -1;
		}
	}

	bool UseSSL = false;
	string RealServer = Server;
	MakeUpper(Server);
	string::size_type index = Server.rfind("(SSL)");
	if(index == RealServer.length() - 5)
	{
		RealServer = RealServer.substr(0, index);
		UseSSL = true;
	}

	bool UseSSO = false;
	Server = RealServer;
	MakeUpper(Server);
	index = Server.rfind("(SSO");
	string SSOPrincipal = "";
	if(index != string::npos)
	{
		// Search optional pricipal name (ex (SSO:MyPrincipal))
		string::size_type index2 = Server.find(')', index);
		if(index2 != string::npos)
		{
			string SSOString = RealServer.substr(index+1, index2-index-1);
			RealServer.erase(index, index2-index+1);
			UseSSO = true;
			index = SSOString.find(':');
			if(index != string::npos)
				SSOPrincipal = SSOString.substr(index+1);
		}
	}
	CleanString(SSOPrincipal);

	string ConnectString = "Server=\"";
	ConnectString += RealServer;
	ConnectString += "\"";
	if(UseSSL)
		ConnectString += " Ssl=\"YES\"";
	if(UseSSO)
		ConnectString += " Sso=\"YES\"";
	if(SSOPrincipal.length())
	{
		ConnectString += " SsoPrincipal=\"";
		ConnectString += SSOPrincipal;
		ConnectString += "\"";
	}
	if(SessionId.length())
	{
		ConnectString += " Session=\"";
		ConnectString += SessionId;
		ConnectString += "\"";
	}
	if(Application.length())
	{
		ConnectString += " Application=\"";
		ConnectString += Application;
		ConnectString += "\"";
	}
	if(User.length())
	{
		ConnectString += " User=\"";
		ConnectString += User;
		ConnectString += "\"";
	}
	if(Password.length())
	{
		ConnectString += " Password=\"";
		ConnectString += Password;
		ConnectString += "\"";
	}
	if(Open.length())
	{
		ConnectString += " Open=\"";
		ConnectString += Open;
		ConnectString += "\"";
	}
	if(Close.length())
	{
		ConnectString += " Close=\"";
		ConnectString += Close;
		ConnectString += "\"";
	}
	if(Unicode)
		ConnectString += " Unicode=\"YES\"";

	if(Certificate.length())
	{
		ConnectString += " Certificate=\"";
		ConnectString += Certificate;
		ConnectString += "\"";
	}

	NVREQUEST Request = NvOpenRequest(ConnectString.c_str());
	if(Request == NULL)
	{
		if(stream != NULL)
			fclose(stream);
		if(DisplayErrors)
			printf("\nCannot open request");
		return -1;
	}

	char CommandBuffer[32768];
	int DataBufferSize = 32768;
	char *DataBuffer = (char *)malloc(DataBufferSize);
	strcpy(DataBuffer, "");

	int Result;

	// The eventual parameters are declared as local request variables named PARAM1, PARAM2, etc...
	int i;
	for(i=0; i<(int)CommandParameters.size(); i++)
	{
		string ParameterValue = CommandParameters[i];
		sprintf(Buffer, "PARAM%d", i+1);
		string ParameterName = Buffer;
		string Command = "NV_SERVICE=\"LOCAL\" NV_CLASS=\"VARIABLE\" NV_COMMAND=\"SET\" NAME=\"";
		Command += ParameterName;
		Command += "\" VALUE=\"";
		Command += ParameterValue;
		Command += "\"";
		Result = NvCommand(Request, Command.c_str(), NULL, 0, NULL);
		if((!CheckError(Request, Result)) && StopOnError)
		{
			if(stream != NULL)
				fclose(stream);
			if(DataBuffer != NULL)
				free(DataBuffer);
			Cleanup(Request);
			return -1;
		}
	}
	
	bool bCommandContinue = false;
	string PreviousLine = "";
	if(stream == NULL)
	{
		// Just executes the procedure if there is one
		int DataSize = 0;
		if(Procedure.length())
		{
			string Line = "NV_PROC=|";
			Line += Procedure;
			Line += "|";
			Result = NvCommand(Request, Line.c_str(), DataBuffer, DataBufferSize, &DataSize);	
			if((!CheckError(Request, Result)) && StopOnError)
			{
				if(DataBuffer != NULL)
					free(DataBuffer);
				Cleanup(Request);
				return -1;
			}
		}
		else
		{
			// Execute the direct command
			Result = NvCommand(Request, DirectCommand.c_str(), DataBuffer, DataBufferSize, &DataSize);	
			if((!CheckError(Request, Result)) && StopOnError)
			{
				if(DataBuffer != NULL)
					free(DataBuffer);
				Cleanup(Request);
				return -1;
			}
		}
		if((DataSize+1) > DataBufferSize)
		{
			char *NewDataBuffer = (char *)realloc(DataBuffer, DataSize+1);
			if(NewDataBuffer == NULL)
			{
				if(DataBuffer != NULL)
					free(DataBuffer);
				Cleanup(Request);
				return -1;
			}
			DataBuffer = NewDataBuffer;
			DataBufferSize = DataSize+1;
			strcpy(DataBuffer, "");

			// We get the last output buffer
			NvCommand(Request, "NV_CMD=|LOCAL:GET_LAST_OUTPUT_BUFFER|", DataBuffer, DataBufferSize, &DataSize);	
		}
	}
	else
	{
		// execute commands from the input file 
		while(freadline(CommandBuffer, sizeof(CommandBuffer), stream))
		{
			string Line = CommandBuffer;
			if((!Line.length()) || (Line[0] == ';') || (Line[0] == '/') || (Line[0] == '\\'))
			{
				PreviousLine = "";
				continue;
			}
			
			string::size_type index = Line.find("nvcc::printf");
			if(index == 0)
			{
				PreviousLine = "";
				// printf line
				Line.erase(0, 12);
				if(Line.length())	// We erase the first blank if there is one
					Line.erase(0, 1);
				if(DisplayPrint)
				{
					string LineToPrint = "";
					while(Line.length())
					{
						index = Line.find("\\n");
						string Provistr = Line;
						if(index == string::npos)
						{
							Line = "";
						}
						else
						{
							Provistr = Line.substr(0, index);
							Provistr += "\n";
							Line = Line.substr(index+2);
						}
						LineToPrint += Provistr;
					}
					Line = LineToPrint;
					LineToPrint = "";
					while(Line.length())
					{
						index = Line.find("\\t");
						string Provistr = Line;
						if(index == string::npos)
						{
							Line = "";
						}
						else
						{
							Provistr = Line.substr(0, index);
							Provistr += "\t";
							Line = Line.substr(index+2);
						}
						LineToPrint += Provistr;
					}
					printf(LineToPrint.c_str());
				}
				continue;
			}

			index = Line.find("nvcc::printdata");
			if(index == 0)
			{
				PreviousLine = "";
				// printf data buffer
				if(DisplayPrint && (DataBuffer != NULL))
					printf(DataBuffer);
				continue;
			}

			RemoveFirstAndEndBlanks(Line);
			if(!Line.length())
				continue;

			// replace tab with space
			while((index = Line.find('\t')) != string::npos)
				Line.replace(index, 1, 1, ' ');

			RemoveFirstAndEndBlanks(Line);
			if(!Line.length())
				continue;


			if(PreviousLine.length())
			{
				string Provistr = Line;
				Line = PreviousLine;
				Line += " ";
				Line += Provistr;
				PreviousLine = "";
			}

			if((Line[Line.length()-1] == '/') || (Line[Line.length()-1] == '\\'))
			{
				// Line continuation
				PreviousLine = Line.substr(0, Line.length()-1);
				
				// replace tab with space
				while((index = PreviousLine.find('\t')) != string::npos)
					PreviousLine.replace(index, 1, 1, ' ');
		
				RemoveFirstAndEndBlanks(PreviousLine);
				continue;
			}
			
			if(DisplayCommands)
			{
				printf("\nCommand : ");
				printf(Line.c_str());
			}
			
			int DataSize = 0;
			Result = NvCommand(Request, Line.c_str(), DataBuffer, DataBufferSize, &DataSize);	
			if((!CheckError(Request, Result)) && StopOnError)
			{
				if(stream != NULL)
					fclose(stream);
				if(DataBuffer != NULL)
					free(DataBuffer);
				Cleanup(Request);
				return -1;
			}
			if((DataSize+1) > DataBufferSize)
			{
				char *NewDataBuffer = (char *)realloc(DataBuffer, DataSize+1);
				if(NewDataBuffer == NULL)
				{
					if(stream != NULL)
						fclose(stream);
					if(DataBuffer != NULL)
						free(DataBuffer);
					if(DisplayErrors)
						printf("\nMemory allocation error");
					Cleanup(Request);
					return -1;
				}
				DataBuffer = NewDataBuffer;
				DataBufferSize = DataSize+1;
				strcpy(DataBuffer, "");

				// We get the last output buffer
				NvCommand(Request, "NV_CMD=|LOCAL:GET_LAST_OUTPUT_BUFFER|", DataBuffer, DataBufferSize, &DataSize);	
/*				Result = NvCommand(Request, Line.c_str(), DataBuffer, DataBufferSize, &DataSize);	
				if((!CheckError(Request, Result)) && StopOnError)
				{
					if(stream != NULL)
						fclose(stream);
					if(DataBuffer != NULL)
						free(DataBuffer);
					Cleanup(Request);
					return -1;
				}*/
			}
		}
	}

	if(DataBuffer != NULL)
		free(DataBuffer);
	if(stream != NULL)
		fclose(stream);
	Cleanup(Request);
	return 0;
}


void Cleanup(NVREQUEST Request)
{
	char CommandBuffer[2048];
	char DataBuffer[2048];
	if(CloseSession)
	{
		strcpy(CommandBuffer, "NV_SERVICE=\"Local\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"GET_SESSION_ID\"");
		int Result = NvCommand(Request, CommandBuffer, DataBuffer, sizeof(DataBuffer), NULL);	
		if((Result == 1) && strlen(DataBuffer))
		{
			sprintf(CommandBuffer, "NV_SERVICE=\"System\" NV_CLASS=\"Session\" NV_COMMAND=\"close\"");
			NvCommand(Request, CommandBuffer, NULL, 0, NULL);
		}
	}
	
	NvCloseRequest(Request);
	NvExitLib();

}

bool CheckError(NVREQUEST Request, int Result)
{
	if(Result == -1)
	{
		if(DisplayErrors)
			printf("\nInvalid request");
		return false;
	}
	if(Result == 0)
	{
		// Error
		if(DisplayErrors)
		{
			char ErrorString[256];
			NvCommand(Request, "NV_SERVICE=\"Local\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"ERROR_INFO\" ERROR_STRING=\"SERVICE\"", ErrorString, sizeof(ErrorString), NULL);
			printf("\n\tError service     : ");
			printf(ErrorString);
			NvCommand(Request, "NV_SERVICE=\"Local\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"ERROR_INFO\" ERROR_STRING=\"CLASS\"", ErrorString, sizeof(ErrorString), NULL);
			printf("\n\tError class       : ");
			printf(ErrorString);
			NvCommand(Request, "NV_SERVICE=\"Local\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"ERROR_INFO\" ERROR_STRING=\"CODE\"", ErrorString, sizeof(ErrorString), NULL);
			printf("\n\tError code        : ");
			printf(ErrorString);
			NvCommand(Request, "NV_SERVICE=\"Local\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"ERROR_INFO\" ERROR_STRING=\"INFO\"", ErrorString, sizeof(ErrorString), NULL);
			printf("\n\tError info        : ");
			printf(ErrorString);
			NvCommand(Request, "NV_SERVICE=\"Local\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"ERROR_INFO\" ERROR_STRING=\"DESCRIPTION\"", ErrorString, sizeof(ErrorString), NULL);
			printf("\n\tError description : ");
			printf(ErrorString);
		}
		return false;
	}

	return true;
}


int GetCommandLineOption(int argc, char* argv[], char option, int more)
{
	char Buffer[2048];		// General buffer

	int i;
	for(i=1; i<argc; i++)
	{
		strcpy(Buffer, argv[i]);
		if(Buffer[0] != '-')
			continue;
		if(strlen(Buffer) <= 1)
			continue;
		if(Buffer[1] != option)
			continue;
		
		// we found the option
		if(!more)			// if nothing more to get
			return 0;		// we leave

		i++;
		if(i == argc)		// error extension not found
			return 1;

		strcpy(OptionBuffer, argv[i]);
		if(OptionBuffer[0] == '-')		// error extension not found
			return 1;

		CleanString(OptionBuffer);

		return 0;		// we got the extension into Buffer
	}

	return 1;		// we didn't find the option
}



// Builds the list of parameters
int GetParametersFromCommandLine(int argc, char* argv[])
{
	char Buffer[1024];		// General buffer

	strcpy(Buffer, "");
	int i;
	for(i=1; i<argc; i++)
	{
		if(argv[i][0] != '-')
			break;
		if(argv[i][1] == 'c')	// test no close option
			continue;
		if(argv[i][1] == 't')	// test no stop on error option
			continue;
		if(argv[i][1] == 'n')	// test unicode option
			continue;
		i++;
		if(i == argc)		// error extension not found
			return 1;

		if(argv[i][0] == '-')
			i--;
	}

	for(; i<argc; i++)
	{
		strcpy(Buffer, argv[i]);
		// discard first and end double quotes
		if(strlen(Buffer) && (Buffer[0] == '\"'))
			strcpy(Buffer, argv[i]+1);
		if(strlen(Buffer) && (Buffer[strlen(Buffer)-1] == '\"'))
			Buffer[strlen(Buffer)-1] = '\0';

		CleanString(Buffer);

		if(!strlen(Buffer))
			continue;

		CommandParameters.push_back(Buffer);

	}


	return 0;		// al is ok
}

	

void DisplayHelp()
{
	printf("\n\nUtility for sending commands to a NIRVA server");
	printf("\nversion : ");
	printf(NV_VERSION);
	printf("\nCopyright NIRVA SYSTEMS 2002");
	printf("\n\n\nusage: nvcc [options] parameters");
	printf("\n\nparameters:");
	printf("\n          Any command line string that is not an option is considered as a");
	printf("\n          parameter. A parameter is automatically transformed by nvcc into a");
	printf("\n          request variable so a parameter can be accessed directly from a NIRVA");
	printf("\n          command.");
	printf("\n          The created request variables corresponding to the parameters are");
	printf("\n          named Param1, Param2, Param3, etc... where Param1 corresponds to the");
	printf("\n          first parameter, Param2 to the second and so on.");
	printf("\n\noptions:");
	printf("\n          -i <inputfile>");
	printf("\n             Name of the input file containing NIRVA commands.");
	printf("\n             If this parameter is not provided, nvcc uses a default name");
	printf("\n             input.txt.");
	printf("\n             The input file contains a succession of text lines corresponding to");
	printf("\n             NIRVA commands.");
	printf("\n             The input file can also contain some comments or some specific nvcc");
	printf("\n             print orders.");
	printf("\n             Please consult the NIRVA documentation for further information");
	printf("\n             about the nvcc input file format.");
	printf("\n          -r <procedure>");
	printf("\n             Name of Nirva procedure.");
	printf("\n             If this parameter is provided, nvcc doesn't use the input file but");
	printf("\n             directly asks the server to execute the given procedure.");
	printf("\n          -z <nirva command>");
	printf("\n             Nirva command.");
	printf("\n             If this parameter is provided, nvcc doesn't use the input file or");
	printf("\n             procedure but directly executes the given command.");
	printf("\n          -d <display mode>");
	printf("\n             Display mode is a string composed of letters c, e, p and n. If n is");
	printf("\n             in this string, nvcc runs in a complete silent mode. If c is in the");
	printf("\n             string, nvcc displays commands. If e is in the string, nvcc");
	printf("\n             displays errors. If p is in the string, nvcc displays the specific");
	printf("\n             nvcc print output found in the input file.");
	printf("\n             The default display mode is pe.");
	printf("\n          -a <NIRVA server address>");
	printf("\n             This is the TCP/IP address of the NIRVA server");
	printf("\n             This can be followed by :PortNumber if the port number of the NIRVA");
	printf("\n             server is different than the default one (1081).");
	printf("\n          -k <certificate;password>");
	printf("\n             Client certificate and otpional password when using SSL");
	printf("\n          -s <session ID>");
	printf("\n             NIRVA session to connect. If not provided, NIRVA creates a new");
	printf("\n             session.");
	printf("\n          -p <application name>");
	printf("\n             Name of the NIRVA application to connect. If not provided, nvcc");
	printf("\n             connects to the NIRVA default application.");
	printf("\n          -u <user name>");
	printf("\n             Application user name. If not provided, nvcc uses to the NIRVA");
	printf("\n             application default user name.");
	printf("\n          -o <open procedure!close procedure>");
	printf("\n             Name of the session open and close procedures. The open and close");
	printf("\n             procedure names are separated by the ! character. If one is omitted");
	printf("\n             or if the -o parameter is not provided NIRVA uses session_open and");
	printf("\n             session_close as default names");
	printf("\n          -w <user password>");
	printf("\n             Application user password. This may be followed by a new password");
	printf("\n             if a previous login required the user to change its password. At");
	printf("\n             this time both passwords are separated by a semicolon character.");
	printf("\n          -c");
	printf("\n             No close option. If this option is given, nvcc doesn't close the");
	printf("\n             NIRVA session when exiting.");
	printf("\n             The default is to close the session.");
	printf("\n          -t");
	printf("\n             No stop on error option. If this option is given, nvcc doesn't exit");
	printf("\n             in case of error on a command.");
	printf("\n             The default is to stop in case of error.");
	printf("\n          -n");
	printf("\n             Unicode option. If this option is given, all data sent to NIRVA are");
	printf("\n             supposed to be unicode encoded and all data coming from NIRVA are");
	printf("\n             unicode translated. File and binary data object are not translated.");
	printf("\n          -m <max message size>");
	printf("\n             Maximum message size in kilobytes. Nirva uses temporary files for");
	printf("\n             message data if the fixed limit overflows.");
	printf("\n          -b <test string>");
	printf("\n             Runs nvcc in test mode. This option allows to run several times the");
	printf("\n             given input file. The test string has the format");
	printf("\n             \"l=numloop:r=numthread\" where numthread is the number of");
	printf("\n             simultaneous threads to be created and numloop is the number of");
	printf("\n             times each thread will execute the input file.");
	printf("\n          -h");
	printf("\n             Display help.");
	printf("\n\n");
}
	


int RunTestMode(int argc, char* argv[])
{
	// For window, increase the maximum number of stdio file from default 512
	// to 2048 (this is the max, trying over crashes the prgm under win64).
#ifdef WIN32
	if(_setmaxstdio(2048) == -1)
		_setmaxstdio(1024);
#else
	// We do the same for unix. This includes all descriptors like sockets and pipes
	struct rlimit r;
	r.rlim_cur = r.rlim_max = 8192;
	if(setrlimit(RLIMIT_NOFILE,&r) != 0)	// first try with 8192 then with 4096, 2048 and 1024
	{
		r.rlim_cur = r.rlim_max = 4096;
		if(setrlimit(RLIMIT_NOFILE,&r) != 0)
		{
			r.rlim_cur = r.rlim_max = 2048;
			if(setrlimit(RLIMIT_NOFILE,&r) != 0)
			{
				r.rlim_cur = r.rlim_max = 1024;
				setrlimit(RLIMIT_NOFILE,&r);
			}
		}
	}
#endif

	int NumLoops = 1;
	int NumRequests = 1;

#ifndef WIN32
	InitialSeconds = GetSeconds();
#endif


	string::size_type index = TestOptions.find("l=");
	if(index != string::npos)
	{
		string sNumLoops = TestOptions.substr(index+2);
		index = sNumLoops.find(';');
		if(index != string::npos)
			sNumLoops = sNumLoops.substr(0, index);
		NumLoops = atoi(sNumLoops.c_str());
		if(NumLoops <= 0)
			NumLoops = 1;
	}
	
	index = TestOptions.find("r=");
	if(index != string::npos)
	{
		string sNumRequests = TestOptions.substr(index+2);
		index = sNumRequests.find(';');
		if(index != string::npos)
			sNumRequests = sNumRequests.substr(0, index);
		NumRequests = atoi(sNumRequests.c_str());
		if(NumRequests <= 0)
			NumRequests = 1;
	}

	// For now, the request are not finished
	Finished = 0;	
	Deleted = 0;	
	Tstart.Reset();
	Tend.Reset();
	
	printf("\n\n");
	printf("\nConnecting sessions...");
	// Start all threads and let them connecting every 10 ms

	vector<TestRequest *> requests;
	int i;
	for(i=0; i<NumRequests; i++)
	{
		TestRequest *tRequest = new TestRequest;
		tRequest->m_InputFile = InputFile;
		tRequest->m_Procedure = Procedure;
		tRequest->m_Server = Server;
		tRequest->m_Certificate = Certificate;
		tRequest->m_SessionId = SessionId;
		tRequest->m_Application = Application;
		tRequest->m_User = User;
		tRequest->m_Password = Password;
		tRequest->m_Open = Open;
		tRequest->m_Close = Close;
		string::size_type index = tRequest->m_Password.find(';');
		if(index != string::npos)
		{
			tRequest->m_NewPassword = tRequest->m_Password.substr(index+1);
			tRequest->m_Password = tRequest->m_Password.substr(0, index);
		}
		tRequest->m_DisplayCommands = DisplayCommands;
		tRequest->m_DisplayErrors = DisplayErrors;
		tRequest->m_DisplayPrint = DisplayPrint;
		tRequest->m_CloseSession = CloseSession;
		tRequest->m_StopOnError = StopOnError;
		tRequest->m_Unicode = Unicode;
		tRequest->m_NbLoops = NumLoops;
		int j;
		for(j=0; j<(int)CommandParameters.size(); j++)
			tRequest->m_CommandParameters.push_back(CommandParameters[j]);
		tRequest->Connect();
		requests.push_back(tRequest);
//		tRequest->Run(false, true);
//		SLEEP(25);
	}


	// Now wait for all threads to be connected
	printf("\nWaiting for all sessions to be connected...");
	while(1)
	{
		int ConnectedHere = 0;
		ProtectConnected.Lock();
		ConnectedHere = Connected;
		ProtectConnected.Unlock();
		if(ConnectedHere >= NumRequests)
			break;
		SLEEP(100);
	}

	for(i=0; i<(int)requests.size(); i++)
	{
		TestRequest *tRequest = requests[i];
		tRequest->Run(false, true);
		SLEEP(25);
	}

	printf("\nStart processing commands...");
	// signal the threads to start processing so all threads will start exactly at the same time
#ifdef WIN32
	clock_t m_cStartProcessTime = clock();
#else
	unsigned int m_cStartProcessTime = GetMilliSeconds();
#endif
	Tstart.Signal();

	// Wait for all threads to have finished their processing
	printf("\nWaiting for end processing...");
	while(1)
	{
		int FinishedHere = 0;
		ProtectFinished.Lock();
		FinishedHere = Finished;
		ProtectFinished.Unlock();
		if(FinishedHere >= NumRequests)
			break;
		SLEEP(10);
	}

#ifdef WIN32
	clock_t m_cProcessTime = clock();
	m_cProcessTime = m_cProcessTime - m_cStartProcessTime;
	int RealProcessTimeInMilliseconds = (int)((m_cProcessTime * 1000)/CLOCKS_PER_SEC);
#else
	unsigned int m_cProcessTime = GetMilliSeconds();
	m_cProcessTime = m_cProcessTime - m_cStartProcessTime;
	int RealProcessTimeInMilliseconds = m_cProcessTime;
#endif

	printf("\nProcessing has finished...");

	printf("\nDisconnecting sessions...");
	
	// signal the threads to leave their main loop and to auto delete
	Tend.Signal();

	printf("\nWaiting for all sessions to be disconnected...");
	// Wait for all threads to have be deleted
	while(1)
	{
		int DeletedHere = 0;
		ProtectDeleted.Lock();
		DeletedHere = Deleted;
		ProtectDeleted.Unlock();
		if(DeletedHere >= NumRequests)
			break;
		SLEEP(100);
	}

	printf("\n\n");
	printf("TEST RESULTS\n");

	printf("\nNumber of threads = %d", NumRequests);
	printf("\nNumber of loops = %d", NumLoops);
	printf("\nTotal number of requests = %d", NumLoops * NumRequests);
	printf("\nProcessing time = %d ms", RealProcessTimeInMilliseconds);
	if(RealProcessTimeInMilliseconds == 0)
		RealProcessTimeInMilliseconds = 1;
	printf("\nRate = %d requests / second", (NumLoops * NumRequests * 1000)/(RealProcessTimeInMilliseconds));

//	char Buffer[10];
//	printf("\n\nPress enter to continue");
//	ReadLine(Buffer, sizeof(Buffer));

	return 0;
}



int TestRequest::Main()
{
	// Do request connection here
//	char Buffer[1024];


	FILE *stream = NULL;
	if(m_InputFile.length())
	{
		stream = fopen(m_InputFile.c_str(), "rb");
		if(stream == NULL)
		{
			if(DisplayErrors)
				printf("\nCannot open input file");
			SetConnected();
			SetFinished();
			return -1;
		}
	}

/*	string ConnectString = "Server=\"";
	ConnectString += m_Server;
	ConnectString += "\"";
	if(m_SessionId.length())
	{
		ConnectString += " Session=\"";
		ConnectString += m_SessionId;
		ConnectString += "\"";
	}
	if(m_Application.length())
	{
		ConnectString += " Application=\"";
		ConnectString += m_Application;
		ConnectString += "\"";
	}
	if(m_User.length())
	{
		ConnectString += " User=\"";
		ConnectString += m_User;
		ConnectString += "\"";
	}
	if(m_Password.length())
	{
		ConnectString += " Password=\"";
		ConnectString += m_Password;
		ConnectString += "\"";
	}

	Request = NvOpenRequest(ConnectString.c_str());
	if(Request == NULL)
	{
		if(stream != NULL)
			fclose(stream);
		if(m_DisplayErrors)
			printf("\nCannot open request");
		SetConnected();
		SetFinished();
		return -1;
	}
*/
	char CommandBuffer[2048];
	int DataBufferSize = 2048;
	char *DataBuffer = (char *)malloc(DataBufferSize);
	strcpy(DataBuffer, "");

/*
	// The eventual parameters are declared as local request variables named PARAM1, PARAM2, etc...
	int i;
	for(i=0; i<m_CommandParameters.size(); i++)
	{
		string ParameterValue = m_CommandParameters[i];
		sprintf(Buffer, "PARAM%d", i+1);
		string ParameterName = Buffer;
		string Command = "NV_SERVICE=\"LOCAL\" NV_CLASS=\"VARIABLE\" NV_COMMAND=\"SET\" NAME=\"";
		Command += ParameterName;
		Command += "\" VALUE=\"";
		Command += ParameterValue;
		Command += "\"";
		int Result = NvCommand(Request, Command.c_str(), NULL, 0, NULL);
		if((!CheckError(Request, Result)) && m_StopOnError)
		{
			if(stream != NULL)
				fclose(stream);
			if(DataBuffer != NULL)
				free(DataBuffer);
			Cleanup(Request);
			SetConnected();
			SetFinished();
			return -1;
		}
	}

	// Send a NOP command to open the session
	int Result = NvCommand(Request, "NV_SERVICE=\"SYSTEM\" NV_CLASS=\"MISC\" NV_COMMAND=\"NOP\"", NULL, 0, NULL);
	if((!CheckError(Request, Result)) && m_StopOnError)
	{
		if(stream != NULL)
			fclose(stream);
		if(DataBuffer != NULL)
			free(DataBuffer);
		Cleanup(Request);
		SetConnected();
		SetFinished();
		return -1;
	}


	SetConnected();
*/
	Tstart.Lock();
	Tstart.Unlock();

	int Result = 0;
	// Do the loop for commands now

	int k;
	for(k=0; k<m_NbLoops; k++)
	{
		NvCommand(Request, "NV_SERVICE=\"LOCAL\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"RESET_COMMAND_BUFFER\"", NULL, 0, NULL);
//		NvCommand(Request, "NV_SERVICE=\"SYSTEM\" NV_CLASS=\"MISC\" NV_COMMAND=\"NOP\"", NULL, 0, NULL);
		// Go back to the start of the file
		if(stream == NULL)
		{
			// Just executes the procedure
			string Line = "NV_PROC=|";
			Line += Procedure;
			Line += "|";
			int DataSize = 0;
			Result = NvCommand(Request, Line.c_str(), DataBuffer, DataBufferSize, &DataSize);	
			if((!CheckError(Result)) && m_StopOnError)
			{
				if(DataBuffer != NULL)
					free(DataBuffer);
				Cleanup();
				SetFinished();
				return -1;
			}
		}
		else
		{
			fseek(stream, 0, SEEK_SET);
			bool bCommandContinue = false;
			string PreviousLine = "";
			while(freadline(CommandBuffer, sizeof(CommandBuffer), stream))
			{
				string Line = CommandBuffer;
				if((!Line.length()) || (Line[0] == ';') || (Line[0] == '/') || (Line[0] == '\\'))
				{
					PreviousLine = "";
					continue;
				}
				
				string::size_type index = Line.find("nvcc::printf");
				if(index == 0)
				{
					PreviousLine = "";
					// printf line
					Line.erase(0, 12);
					if(Line.length())	// We erase the first blank if there is one
						Line.erase(0, 1);
					if(m_DisplayPrint)
					{
						string LineToPrint = "";
						while(Line.length())
						{
							index = Line.find("\\n");
							string Provistr = Line;
							if(index == string::npos)
							{
								Line = "";
							}
							else
							{
								Provistr = Line.substr(0, index);
								Provistr += "\n";
								Line = Line.substr(index+2);
							}
							LineToPrint += Provistr;
						}
						Line = LineToPrint;
						LineToPrint = "";
						while(Line.length())
						{
							index = Line.find("\\t");
							string Provistr = Line;
							if(index == string::npos)
							{
								Line = "";
							}
							else
							{
								Provistr = Line.substr(0, index);
								Provistr += "\t";
								Line = Line.substr(index+2);
							}
							LineToPrint += Provistr;
						}
						printf(LineToPrint.c_str());
					}
					continue;
				}

				index = Line.find("nvcc::printdata");
				if(index == 0)
				{
					PreviousLine = "";
					// printf data buffer
					if(m_DisplayPrint && (DataBuffer != NULL))
						printf(DataBuffer);
					continue;
				}

				RemoveFirstAndEndBlanks(Line);
				if(!Line.length())
					continue;

				// replace tab with space
				while((index = Line.find('\t')) != string::npos)
					Line.replace(index, 1, 1, ' ');

				RemoveFirstAndEndBlanks(Line);
				if(!Line.length())
					continue;


				if(PreviousLine.length())
				{
					string Provistr = Line;
					Line = PreviousLine;
					Line += " ";
					Line += Provistr;
					PreviousLine = "";
				}

				if((Line[Line.length()-1] == '/') || (Line[Line.length()-1] == '\\'))
				{
					// Line continuation
					PreviousLine = Line.substr(0, Line.length()-1);
					
					// replace tab with space
					while((index = PreviousLine.find('\t')) != string::npos)
						PreviousLine.replace(index, 1, 1, ' ');
			
					RemoveFirstAndEndBlanks(PreviousLine);
					continue;
				}
				
				if(m_DisplayCommands)
				{
					printf("\nCommand : ");
					printf(Line.c_str());
				}
				
				int DataSize = 0;
				Result = NvCommand(Request, Line.c_str(), DataBuffer, DataBufferSize, &DataSize);	
				if((!CheckError(Result)) && m_StopOnError)
				{
					if(stream != NULL)
						fclose(stream);
					if(DataBuffer != NULL)
						free(DataBuffer);
					Cleanup();
					SetFinished();
					return -1;
				}
				if((DataSize+1) > DataBufferSize)
				{
					char *NewDataBuffer = (char *)realloc(DataBuffer, DataSize+1);
					if(NewDataBuffer == NULL)
					{
						if(stream != NULL)
							fclose(stream);
						if(DataBuffer != NULL)
							free(DataBuffer);
						if(m_DisplayErrors)
							printf("\nMemory allocation error");
						Cleanup();
						SetFinished();
						return -1;
					}
					DataBuffer = NewDataBuffer;
					DataBufferSize = DataSize+1;
					strcpy(DataBuffer, "");

					// We get the last output buffer
					NvCommand(Request, "NV_CMD=|LOCAL:GET_LAST_OUTPUT_BUFFER|", DataBuffer, DataBufferSize, &DataSize);	
/*					Result = NvCommand(Request, Line.c_str(), DataBuffer, DataBufferSize, &DataSize);	
					if((!CheckError(Result)) && m_StopOnError)
					{
						if(stream != NULL)
							fclose(stream);
						if(DataBuffer != NULL)
							free(DataBuffer);
						Cleanup();
						SetFinished();
						return -1;
					}*/
				}
			}
		}
	}

	// End real processing
	SetFinished();

	if(stream != NULL)
		fclose(stream);
	if(DataBuffer != NULL)
		free(DataBuffer);

	Tend.Lock();
	Tend.Unlock();

	Cleanup();

	return 0;
}


int TestRequest::Connect()
{
	// Do request connection here
	char Buffer[1024];

	bool UseSSL = false;
	string Server = m_Server;
	string RealServer = Server;
	MakeUpper(Server);
	string::size_type index = Server.rfind("(SSL)");
	if(index != string::npos)
	{
		RealServer.erase(index, 5);
		UseSSL = true;
	}
	bool UseSSO = false;
	Server = RealServer;
	MakeUpper(Server);
	index = Server.rfind("(SSO");
	string SSOPrincipal = "";
	if(index != string::npos)
	{
		// Search optional pricipal name (ex (SSO:MyPrincipal))
		string::size_type index2 = Server.find(')', index);
		if(index2 != string::npos)
		{
			string SSOString = RealServer.substr(index+1, index2-index-1);
			RealServer.erase(index, index2-index+1);
			UseSSO = true;
			index = SSOString.find(':');
			if(index != string::npos)
				SSOPrincipal = SSOString.substr(index+1);
		}
	}
	CleanString(SSOPrincipal);

	string ConnectString = "Server=\"";
	ConnectString += RealServer;
	ConnectString += "\"";
	if(UseSSL)
		ConnectString += " Ssl=\"YES\"";
	if(UseSSO)
		ConnectString += " Sso=\"YES\"";
	if(SSOPrincipal.length())
	{
		ConnectString += " SsoPrincipal=\"";
		ConnectString += SSOPrincipal;
		ConnectString += "\"";
	}
	if(m_SessionId.length())
	{
		ConnectString += " Session=\"";
		ConnectString += m_SessionId;
		ConnectString += "\"";
	}
	if(m_Application.length())
	{
		ConnectString += " Application=\"";
		ConnectString += m_Application;
		ConnectString += "\"";
	}
	if(m_User.length())
	{
		ConnectString += " User=\"";
		ConnectString += m_User;
		ConnectString += "\"";
	}
	if(m_Open.length())
	{
		ConnectString += " Open=\"";
		ConnectString += Open;
		ConnectString += "\"";
	}
	if(m_Close.length())
	{
		ConnectString += " Close=\"";
		ConnectString += Close;
		ConnectString += "\"";
	}
	if(m_Password.length())
	{
		ConnectString += " Password=\"";
		ConnectString += m_Password;
		ConnectString += "\"";
	}
	if(m_NewPassword.length())
	{
		ConnectString += " NewPassword=\"";
		ConnectString += m_NewPassword;
		ConnectString += "\"";
	}
	if(m_Unicode)
		ConnectString += " Unicode=\"YES\"";
	if(m_Certificate.length())
	{
		ConnectString += " Certificate=\"";
		ConnectString += m_Certificate;
		ConnectString += "\"";
	}


	Request = NvOpenRequest(ConnectString.c_str());
	if(Request == NULL)
	{
		if(m_DisplayErrors)
			printf("\nCannot open request");
		SetConnected();
		SetFinished();
		return -1;
	}

//	char CommandBuffer[2048];
	int DataBufferSize = 2048;
	char *DataBuffer = (char *)malloc(DataBufferSize);
	strcpy(DataBuffer, "");


	// The eventual parameters are declared as local request variables named PARAM1, PARAM2, etc...
	int i;
	for(i=0; i<(int)m_CommandParameters.size(); i++)
	{
		string ParameterValue = m_CommandParameters[i];
		sprintf(Buffer, "PARAM%d", i+1);
		string ParameterName = Buffer;
		string Command = "NV_SERVICE=\"LOCAL\" NV_CLASS=\"VARIABLE\" NV_COMMAND=\"SET\" NAME=\"";
		Command += ParameterName;
		Command += "\" VALUE=\"";
		Command += ParameterValue;
		Command += "\"";
		int Result = NvCommand(Request, Command.c_str(), NULL, 0, NULL);
		if((!CheckError(Result)) && m_StopOnError)
		{
			if(DataBuffer != NULL)
				free(DataBuffer);
			Cleanup();
			SetConnected();
			SetFinished();
			return -1;
		}
	}

	// Send a NOP command to open the session
	int Result = NvCommand(Request, "NV_SERVICE=\"SYSTEM\" NV_CLASS=\"MISC\" NV_COMMAND=\"NOP\"", NULL, 0, NULL);
	if((!CheckError(Result)) && m_StopOnError)
	{
		if(DataBuffer != NULL)
			free(DataBuffer);
		Cleanup();
		SetConnected();
		SetFinished();
		return -1;
	}


	SetConnected();

	if(DataBuffer != NULL)
		free(DataBuffer);
	return 0;
}



// called when an exception occurs (see class NvTask)
void TestRequest::OnException(string sInfo)
{
	// Do nothing special	
}


// called before the Main function (see class NvTask)
bool TestRequest::OnInit()
{
	NvTask::OnInit();	// call the base class
	// Do nothing special	
	
	return true;
}


// called after the Main function (see class NvTask)
void TestRequest::OnCleanup()
{
	ProtectDeleted.Lock();
	Deleted++;
	ProtectDeleted.Unlock();
	NvTask::OnCleanup();	// call the base class
}


void TestRequest::SetFinished()
{
	ProtectFinished.Lock();
	Finished++;
	ProtectFinished.Unlock();
}


void TestRequest::SetConnected()
{
	ProtectConnected.Lock();
	Connected++;
	ProtectConnected.Unlock();
}



void TestRequest::Cleanup()
{
	if(Request == NULL)
		return;
	char CommandBuffer[2048];
	char DataBuffer[2048];
	if(m_CloseSession)
	{
		strcpy(CommandBuffer, "NV_SERVICE=\"Local\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"GET_SESSION_ID\"");
		int Result = NvCommand(Request, CommandBuffer, DataBuffer, sizeof(DataBuffer), NULL);	
		if((Result == 1) && strlen(DataBuffer))
		{
			sprintf(CommandBuffer, "NV_SERVICE=\"System\" NV_CLASS=\"Session\" NV_COMMAND=\"close\"");
			NvCommand(Request, CommandBuffer, NULL, 0, NULL);
		}
	}
	
	NvCloseRequest(Request);
	Request = NULL;
}


bool TestRequest::CheckError(int Result)
{
	if(Result == -1)
	{
		if(m_DisplayErrors)
			printf("\nInvalid request");
		return false;
	}
	if(Result == 0)
	{
		// Error
		if(m_DisplayErrors)
		{
			char ErrorString[2048];
			NvCommand(Request, "NV_SERVICE=\"Local\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"ERROR_INFO\" ERROR_STRING=\"SERVICE\"", ErrorString, sizeof(ErrorString), NULL);
			printf("\n\tError service     : ");
			printf(ErrorString);
			NvCommand(Request, "NV_SERVICE=\"Local\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"ERROR_INFO\" ERROR_STRING=\"CLASS\"", ErrorString, sizeof(ErrorString), NULL);
			printf("\n\tError class       : ");
			printf(ErrorString);
			NvCommand(Request, "NV_SERVICE=\"Local\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"ERROR_INFO\" ERROR_STRING=\"CODE\"", ErrorString, sizeof(ErrorString), NULL);
			printf("\n\tError code        : ");
			printf(ErrorString);
			NvCommand(Request, "NV_SERVICE=\"Local\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"ERROR_INFO\" ERROR_STRING=\"INFO\"", ErrorString, sizeof(ErrorString), NULL);
			printf("\n\tError info        : ");
			printf(ErrorString);
			NvCommand(Request, "NV_SERVICE=\"Local\" NV_CLASS=\"REQUEST\" NV_COMMAND=\"ERROR_INFO\" ERROR_STRING=\"DESCRIPTION\"", ErrorString, sizeof(ErrorString), NULL);
			printf("\n\tError description : ");
			printf(ErrorString);
		}
		return false;
	}

	return true;
}




void ReadLine(char *Buffer, int size)
{
	fgets(Buffer, size, stdin);
	if(strlen(Buffer))
		if(Buffer[strlen(Buffer)-1] == '\n')
			Buffer[strlen(Buffer)-1] = '\0';
}


#ifndef WIN32
// get num milliseconds
unsigned int GetMilliSeconds() 
{
	struct timeval now;
	gettimeofday(&now, NULL);
	int RetVal = now.tv_sec - InitialSeconds;
	RetVal *= 1000;
	RetVal += now.tv_usec/1000;
	return RetVal;
}

unsigned int GetSeconds() 
{
	struct timeval now;
	gettimeofday(&now, NULL);
	int RetVal = now.tv_sec;
	return RetVal;
}
#endif
 



