// ***************************************************************************************
// File: container.h
// Description: declarations of the Nirva container classes
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: A container is a collection of Nirva objects. It can also contain
//           some other containers
// ***************************************************************************************

#ifndef _CONTAINER_H
#define _CONTAINER_H


// Classes declared in this file
class NvLocalContainer;
class NvContainer;
class NvRootContainer;
class NvRegistryContainer;
class NvDescriptionContainer;
class NvContainerList;
class NvRegistryContainerList;

#include "task.h"
#include "obj.h"
#include <vector>
#include <string>



#define NV_CONTAINER_DEFAULT_NAME				"nvdef"

// Container copy append modes
#define NV_CONTAINER_COPY_NO_APPEND							0
#define NV_CONTAINER_COPY_APPEND							1
#define NV_CONTAINER_COPY_APPEND_NO_REPLACE					2
#define NV_CONTAINER_COPY_APPEND_REPLACE_IF_BAD_TYPE		3

// defines for export
#define NV_CONTAINER_EXPORT_IDENTIFIER			"NIRVA Registry export file"
#define NV_CONTAINER_EXPORT_VERSION				"1.00"

// registry container synhronization mode
#define NV_CONTAINER_SYNCHRONIZE_BOTH				0
#define NV_CONTAINER_SYNCHRONIZE_MEMTODISK			1
#define NV_CONTAINER_SYNCHRONIZE_DISKTOMEM			2

// container levels
#define CONTAINER_LEVEL_APPLICATION					1
#define CONTAINER_LEVEL_SERVICE						2
#define CONTAINER_LEVEL_SYSTEM						3
#define CONTAINER_LEVEL_PARENT						4

// container lock modes
#define CONTAINER_LOCK_READWRITE					0
#define CONTAINER_LOCK_WRITE						1



/////////////////////////////////////////////////////////////////////////////
// NvLocalContainer class
// This is the base container class
// It's used directly by the Nirva client
class NvLocalContainer
{
public:

	std::string m_sName;
	std::map<std::string,void *> m_mObjects;	// object list
	std::string m_sRootDirectory;				// Used for registry containers only
	NvObj *LastObject;							// Last used object
	bool bUseLastObject;						// We must disable that for shared containers
	
	NvLocalContainer();
	NvLocalContainer(std::string Name);
	NvLocalContainer(const char *Name);
	virtual ~NvLocalContainer();

	void Clear();
	void ClearObjects(bool WithFiles = false);
	std::string GetName();
	bool SetName(std::string Name);
	bool SetName(const char *Name);
	void SetRootDirectory(const char *RootDirectory);
	std::string GetObjectFile(const char *Name, int Type);
#ifdef NV_SERVER
	void EncodeName(bool bSrcUnicode, bool bDestUnicode);
#endif

	NvObj* GetObject(const char *Name);
	void RemoveObject(const char *Name);
	bool CreateObject(const char *Name, int Type);
	bool PopulateObject(const char *Name, char *Data, unsigned long DataSize);
	bool PopulateObjectFromFile(const char *Name, const char *FileName, bool bUseDecodeFromFile = false);
	bool GetObjectData(const char *Name, char **Data, unsigned long *DataSize);
};



/////////////////////////////////////////////////////////////////////////////
// NvContainer class
// This is used in nirva server
// This is a hierachical container that can have parents and subcontainers
class NvContainer : public NvLocalContainer
{
public:

	std::map<std::string,void *> m_mSubcontainers;	// sub-container list
	NvContainer *m_cParent;
	NvContainer *LastSubContainer;				// Last used subcontainer
	int iLevel;									// Container level
	bool bUseLastSubContainer;					// We must disable that for shared containers
	
	NvContainer();
	NvContainer(std::string Name);
	NvContainer(const char *Name);
	NvContainer(std::string Name, NvContainer *Parent);
	NvContainer(const char *Name, NvContainer *Parent);
	NvContainer(NvContainer *Parent);
	virtual ~NvContainer();

	// These 2 functions are used for retrieving all subcontainers or object names
	// even those that are at file state but not into the container list
	void GetRealSubContainerNames(std::vector<std::string> *List);
	void GetRealObjectNames(std::vector<std::string> *List);

	void Clear();											// removes all objects and subcontainers
	bool ClearSubcontainers(bool WithFiles = false);		// removes all subcontainers and eventually their files for registry container
	void SetRootDirectory(const char *RootDirectory);
	void SetParent(NvContainer *Parent);
	bool CopyFrom(NvContainer *Source, int AppendMode = NV_CONTAINER_COPY_APPEND, const char *Objects = NULL, bool WithSubContainers = false, bool WithData = true, const char *FileDir = NULL, int PersistTime = 0, bool Recurse = true);
#ifdef NV_SERVER
	void Encode(bool bSrcUnicode, bool bDestUnicode, bool bWithSubcontainers, bool bWithNames);
#endif

	NvContainer* GetParent();
	NvContainer* GetSubContainer(const char *Name);
	bool CreateSubContainer(const char *Name);
	bool RemoveSubContainer(const char *Name);
#ifdef NV_SERVER
	bool GetXmlObjectDataToStream(const char *Name, FILE *stream, int NumTabs = 0, bool WithSubContainers = false, bool WithData = true, bool WithFileData = false, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, NvContainer *Model = NULL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	bool GetXmlObjectDataToString(const char *Name, std::string& Output, int NumTabs = 0, bool WithSubContainers = false, bool WithData = true, bool WithFileData = false, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, NvContainer *Model = NULL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	bool GetXmlObjectDataSize(const char *Name, NvInt64& Size, bool WithSubContainers = false, bool WithFileData = false, NvContainer *Model = NULL);
	bool SetSoapSample(std::string& Output, int NumTabs = 0, std::string NsPrefix = "", void *XmlPtr = NULL, bool bUnicode = false);
	void SetWsdlMessage(std::string& Output, int NumTabs, bool Input, void *XmlPtr = NULL);
#endif
	bool Import(FILE *stream, int AppendMode = NV_CONTAINER_COPY_APPEND, const char *FileDir = NULL, int PersistTime = 0, const char *Object = NULL, NvInt64 *Size = NULL, NvInt64 Offset = 0, NvInt64 BytesToCopy = -1);
	bool Export(FILE *stream, const char *Objects = NULL, bool WithSubContainers = true);
	bool Import(const char *FileName, int AppendMode = NV_CONTAINER_COPY_APPEND, const char *FileDir = NULL, int PersistTime = 0, const char *Object = NULL, NvInt64 *Size = NULL, NvInt64 Offset = 0, NvInt64 BytesToCopy = -1);
	bool Export(const char *FileName, const char *Objects = NULL, bool WithSubContainers = true);
	bool Dir(const char *DirName, const char *Files = "*", const char *SubDirs = "*", bool Recurse = true, int PersistTime = 0, bool NameStandard = false);
	bool Synchronize(int Mode = NV_CONTAINER_SYNCHRONIZE_BOTH);
	bool IsEmpty();
	bool View(NvObjTable *Content, bool WithRootName = true);
	bool MoveObject(NvContainer *DestinationContainer, const char *ObjName, bool Replace = true);
};



/////////////////////////////////////////////////////////////////////////////
// NvRootContainer class
// The root container is the same than a container but has a lock inside
// In fact, a container must be locked when working with it and unlocked after
// Only the root container is locked
// The lock mode can be read write (RW) or write only (W)
// in RW mode, only one thread at a time can access the container
// The thread that owns the container is in ThreadOwner
// The thread access required flag is set when a thread require an exclusive access
// to the container. This blocks all other threads to get the access (except the owner)
// while waiting for the container to be free.
// IntialRefCount is used only in W mode when a thread requires an exclusive access.
// It contains the actual number of access to the container in non exclusive mode at
// this time. it will be used when freeing the container in order to know when to
// return in W mode. The ActualLockMode contains the previous lock mode in this case.
class NvRootContainer : public NvContainer
{
public:
	NvMutex Protect;

	NvRootContainer();
	virtual ~NvRootContainer();
	void Clear();
	void Lock();
	void Unlock();
	long ThreadOwner;
	long ThreadAccessRequired;
	int InitialRefCount;
	std::map<long,int> m_mThreadRefCounts;
	int LockMode;
	int ActualLockMode;

	void IncRefCount(long ThreadId);
	void DecRefCount(long ThreadId, bool Force = false);
	int GetRefCount(long ThreadId);
};



/////////////////////////////////////////////////////////////////////////////
// NvRegistryContainer class
// This special root container is used for registry
// A registry container gets and save its data into physical files
// in a dedicated directory
class NvRegistryContainer : public NvRootContainer
{
public:

	int InUse;				// Number of users having opened the registry (used only
							// with user registries managed by the
							// NvRegistryContainerList class)
	bool bRemoveDirRequest;	// Request to remove the root directory (used with user registries)

	NvRegistryContainer();

	bool GetKey(const char *Key, NvContainer *Destination, int AppendMode = NV_CONTAINER_COPY_APPEND, const char *Entries = NULL, bool WithSubKeys = false, bool WithData = true, const char *FileDir = NULL, int PersistTime = 0, bool Recurse = true);
	bool SetKey(const char *Key, NvContainer *Source, int AppendMode = NV_CONTAINER_COPY_APPEND, const char *Entries = NULL, bool WithSubKeys = false, bool WithData = true);
	bool RemoveKey(const char *Key, const char *Entries = NULL, bool WithSubKeys = false);
	bool Import(const char *Key, const char *FileName, int AppendMode = NV_CONTAINER_COPY_APPEND);
	bool Export(const char *Key, const char *FileName, bool WithSubContainers = true);
};



/////////////////////////////////////////////////////////////////////////////
// NvDescriptionContainer class
// This special root container is used for system and service information
// that is loaded with the system or service
class NvDescriptionContainer : public NvRootContainer
{
public:

	bool Load(const char *FileName);
};


/////////////////////////////////////////////////////////////////////////////
// NvContainerList class
class NvContainerList : public NvObject
{
private:
	NvRootContainer* GetRootContainer(const char *Name, bool Erase = false);
	NvRootContainer* WaitSingleThread(const char *Name, long ThreadId);
	bool InternalCreateContainer(const char *Name, int LockMode = CONTAINER_LOCK_READWRITE);
public:
	bool m_bUseLocking;
	std::map<std::string,void *> m_mContainers;	// container list (This list contains only root containers)
	NvContainer *LastContainer;					// Last used container
	std::string LastContainerName;				// Last used container complete name
	int ListLevel;								// Session, system, application or service container list
	
	NvContainerList();
	virtual ~NvContainerList();

	// Removes all containers from the list
	void Clear();

	// Cleans a container name
	static std::string CleanContainerName(std::string Name);

	// used to get an existing container
	// The function locks the corresponding root container if locking is used
	NvContainer* GetContainer(const char *Name, void *Session = NULL, bool Exclusive = false);

	// used to get a container, if not existing, the container is created
	// The function locks the corresponding root container if locking is used
	NvContainer* GetCreateContainer(const char *Name, void *Session = NULL, bool Exclusive = false, int LockMode = CONTAINER_LOCK_READWRITE);

	// used to free an existing container if locking is used
	// The function unlocks the corresponding root container if locking is used
	void FreeContainer(const char *Name, void *Session = NULL, bool Force = false);

	
	// Creates the root container if it doesn't exist
	bool CreateContainer(const char *Name, int LockMode = CONTAINER_LOCK_READWRITE);
	
	
	// Remove the given container from the list
	void RemoveContainer(const char *Name, void *Session = NULL);

	// Get the root container name
	static std::string GetRootName(std::string Name);

	// Removes the root container name
	static std::string RemoveRootName(std::string Name);

	// Set the root directory for all containers of the list
	void SetRootDirectory(const char *RootDirectory);

	// Synchronize all disk containers
	bool Synchronize(int Mode = NV_CONTAINER_SYNCHRONIZE_BOTH);

	// Get container list
	void GetContainerNames(std::vector<std::string> *List);

};




/////////////////////////////////////////////////////////////////////////////
// NvRegistryContainerList class
class NvRegistryContainerList : public NvObject
{
public:
	std::map<std::string,NvRegistryContainer*> m_mRegistries;			// registries in this list
	std::map<std::string,std::string> m_mDirectories;	// directories ever mapped (to avoid 2 registries pointing to the same directory)
	
	NvRegistryContainerList();
	virtual ~NvRegistryContainerList();

	// Removes all registries from the list
	void Clear();

	// Cleans a registry name
	static std::string CleanRegistryName(std::string Name);

	// Opens and eventually create a user registry
	NvRegistryContainer* OpenRegistry(std::string Name, std::string Directory, bool Create = true);

	// Closes and eventually remove a user registry
	bool CloseRegistry(std::string Name, bool Remove = false, bool RemoveDir = false, int WaitTime = 10);

	// Gets an opened registry
	NvRegistryContainer* GetRegistry(std::string Name);
};



#endif //_CONTAINER_H
