// ***************************************************************************************
// File: log.cpp
// Description: nirva logs
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************


#include "nvs.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;




/////////////////////////////////////////////////////////////////////////////
// NvLogList

NvLogList::NvLogList()
{
	m_sAppName = "";
	m_sServiceName = "";
	LastLog = NULL;
}

NvLogList::~NvLogList()
{
	Clear();
}



void NvLogList::Clear()
{
	NV_LOCKOBJECT(this);
	map<string,void *>::iterator pos;
	for(pos = m_vLogs.begin(); pos != m_vLogs.end(); ++pos)
	{
		NvLog *Log = (NvLog *)(pos->second);
		if(Log != NULL)
			delete Log;
	}
	m_vLogs.clear();
	NV_UNLOCKOBJECT(this);
}

// Load
// Loads the log list from the registry
// Not locked because used at application, service or system load
bool NvLogList::Load()
{
	NvRegistryContainer *Registry;
	if(m_sAppName.length())
	{
		NvApplication *Application = (NvApplication *)Owner;
		Registry = &(Application->AppRegistry);
	}
	else if(m_sServiceName.length())
	{
		NvService *Service = (NvService *)Owner;
		Registry = &(Service->SrvRegistry);
	}
	else
		Registry = &(GetMain()->SystemRegistry);

	// load the log table
	// The log table is stored in a table object named "logs" in the
	// registry key system.logs
	NvContainer Container;
	if(Registry->GetKey("system.logs", &Container, NV_CONTAINER_COPY_NO_APPEND, "logs", false, true))
	{
		NvObj *Object = Container.GetObject("logs");
		if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_TABLE))
		{
			NvObjTable *LogTable = (NvObjTable*)Object;
			int index_Name = LogTable->GetColumnIndex("Name");
			int index_Desc = LogTable->GetColumnIndex("Description");
			int index_MaxFileTime = LogTable->GetColumnIndex("MaxTime");
			int index_SaveIt = LogTable->GetColumnIndex("SaveIt");
			int index_Level = LogTable->GetColumnIndex("Level");
			int index_MaxSaveDays = LogTable->GetColumnIndex("MaxSaveDays");
			int index_Flush = LogTable->GetColumnIndex("Flush");
			int row_index;
			for(row_index=0; row_index<LogTable->GetNumRow(); row_index++)
			{
				// Get log parameters
				string LogName = LogTable->GetCellLine(row_index, index_Name);
				NvLog::CleanLogName(LogName);
				if(!LogName.length()) // The log is identified by its name
					continue;
				if(InternalGetLog(LogName.c_str()) != NULL)	// Check doubles
					continue;
				string Description = LogTable->GetCellLine(row_index, index_Desc);
				string LogMaxFileTime = LogTable->GetCellLine(row_index, index_MaxFileTime);
				int iLogMaxFileTime = atoi(LogMaxFileTime.c_str());
				if(iLogMaxFileTime <= 0)
					iLogMaxFileTime = 1;
				string LogSaveIt = LogTable->GetCellLine(row_index, index_SaveIt);
				bool bLogSaveIt = true;
				if(LogSaveIt.length() && (atoi(LogSaveIt.c_str()) == 0))
					bLogSaveIt = false;
				string LogLevel = LogTable->GetCellLine(row_index, index_Level);
				int iLogLevel = atoi(LogLevel.c_str());
				if(iLogLevel < 0)
					iLogLevel = 0;
				string LogMaxSaveDays = "2";
				if(index_MaxSaveDays >= 0)
					LogMaxSaveDays = LogTable->GetCellLine(row_index, index_MaxSaveDays);
				int iLogMaxSaveDays = atoi(LogMaxSaveDays.c_str());
				if(iLogMaxSaveDays <= 0)
					iLogMaxSaveDays = 1;
				bool bLogFlush = true;
				if(index_Flush >= 0)
				{
					string LogFlush = LogTable->GetCellLine(row_index, index_Flush);
					if(LogFlush.length() && (atoi(LogFlush.c_str()) == 0))
						bLogFlush = false;
				}
				Create(LogName.c_str());
				SetDescription(LogName.c_str(), Description.c_str());
				SetLevel(LogName.c_str(), iLogLevel);
				SetSave(LogName.c_str(), bLogSaveIt);
				SetMaxTime(LogName.c_str(), iLogMaxFileTime);
				SetMaxSaveDays(LogName.c_str(), iLogMaxSaveDays);
				SetFlush(LogName.c_str(), bLogFlush);
			}
		}
		// Resave in case of change
		Save();
	}

	return true;
}


// Saves the log list to the system registry
// This function is called internally so its not locked
bool NvLogList::Save()
{
	NvRegistryContainer *Registry;
	if(m_sAppName.length())
	{
		NvApplication *Application = (NvApplication *)Owner;
		Registry = &(Application->AppRegistry);
	}
	else if(m_sServiceName.length())
	{
		NvService *Service = (NvService *)Owner;
		Registry = &(Service->SrvRegistry);
	}
	else
		Registry = &(GetMain()->SystemRegistry);

	// First create the registry key "system.logs" if it doesn't exist
	Registry->SetKey("system.logs", NULL);

	NvContainer Container;
	if(!Registry->GetKey("system.logs", &Container, NV_CONTAINER_COPY_NO_APPEND, NULL, false, true))
		return false; // cannot get the key. This should not occure here

	// Creates the logs object if it doesn't exist
	Container.CreateObject("logs", NV_OBJTYPE_TABLE);
	NvObj *Object = Container.GetObject("logs");
	if(Object == NULL)
		return false; // This should not occur here

	if(Object->GetType() != NV_OBJTYPE_TABLE)
	{
		Container.RemoveObject("logs");
		Container.CreateObject("logs", NV_OBJTYPE_TABLE);
		Object = Container.GetObject("logs");
		if(Object == NULL)
			return false; // This should not occur here
	}

	NvObjTable *LogTable = (NvObjTable*)Object;
	LogTable->Clear();
	LogTable->InsertColumn("Name");
	LogTable->InsertColumn("Description");
	LogTable->InsertColumn("MaxTime");
	LogTable->InsertColumn("SaveIt");
	LogTable->InsertColumn("Level");
	LogTable->InsertColumn("MaxSaveDays");
	LogTable->InsertColumn("Flush");
	int index_Name = LogTable->GetColumnIndex("Name");
	int index_Desc = LogTable->GetColumnIndex("Description");
	int index_MaxFileTime = LogTable->GetColumnIndex("MaxTime");
	int index_SaveIt = LogTable->GetColumnIndex("SaveIt");
	int index_Level = LogTable->GetColumnIndex("Level");
	int index_MaxSaveDays = LogTable->GetColumnIndex("MaxSaveDays");
	int index_Flush = LogTable->GetColumnIndex("Flush");

	int j;
	map<string,void *>::iterator pos;
	for(pos = m_vLogs.begin(),j=0; pos != m_vLogs.end(); ++pos)
	{
		NvLog *Log = (NvLog *)(pos->second);
		if(Log == NULL)
			continue;

		LogTable->InsertRow();
		LogTable->InsertCellLine(j, index_Name, Log->m_sName.c_str());
		LogTable->InsertCellLine(j, index_Desc, Log->m_sDesc.c_str());
		char Buf[50];
		sprintf(Buf, "%d", Log->m_iMaxFileTime);
		LogTable->InsertCellLine(j, index_MaxFileTime, Buf);
		if(Log->m_bSaveIt)
			LogTable->InsertCellLine(j, index_SaveIt, "1");
		else
			LogTable->InsertCellLine(j, index_SaveIt, "0");
		sprintf(Buf, "%d", Log->m_iLevel);
		LogTable->InsertCellLine(j, index_Level, Buf);
		sprintf(Buf, "%d", Log->m_iMaxSaveDays);
		LogTable->InsertCellLine(j, index_MaxSaveDays, Buf);
		if(Log->m_bFlush)
			LogTable->InsertCellLine(j, index_Flush, "1");
		else
			LogTable->InsertCellLine(j, index_Flush, "0");
		j++;
	}
	
	if(!Registry->SetKey("system.logs", &Container))
		return false;

	return true;
}


NvLog* NvLogList::InternalGetLog(const char *Name)
{
	if(Name == NULL)
		return NULL;

	string NameHere = Name;
	NvLog::CleanLogName(NameHere);

	if(!NameHere.length())
		return NULL;

	if(LastLog != NULL)
	{
		if(LastLog->m_sName == NameHere)
			return LastLog;
	}

	if(m_vLogs.find(NameHere) == m_vLogs.end())
		return NULL;
	
	NvLog *FoundLog = (NvLog *)(m_vLogs[NameHere]);

	LastLog = FoundLog;
	return FoundLog;
}



bool NvLogList::Create(const char *Name, bool UseLocking, bool Erase, const char *ImportFile)
{
	if(Name == NULL)
		return false;

	string NameHere = Name;
	NvLog::CleanLogName(NameHere);

	if(UseLocking)
		NV_LOCKOBJECT(this);

	NvLog *Log = InternalGetLog(Name);
	if(Log != NULL)
	{
		// Log ever exists

		Log->CheckEmptyDirs();

		if(UseLocking)
			NV_UNLOCKOBJECT(this);
		return true;
	}
	
	Log = NVNEW NvLog;
	if(Log == NULL)
	{
		if(UseLocking)
			NV_UNLOCKOBJECT(this);
		return false;
	}

	// Create here the log directory
	string LogDir = GetMain()->GetLogHomeDir();
	AddDirSuffix(LogDir);
	if(m_sAppName.length())
	{
		NvApplication Application;
		Application.m_sName = m_sAppName;
		string AppNameForDir = Application.GetAppNameForFiles();
		MakeUpper(AppNameForDir);
		LogDir += "Applications";
		AddDirSuffix(LogDir);
		LogDir += AppNameForDir;
		AddDirSuffix(LogDir);
	}
	else if(m_sServiceName.length())
	{
		string ServiceNameHere = m_sServiceName;
		NvService::CleanSrvName(ServiceNameHere);
		LogDir += "Services";
		AddDirSuffix(LogDir);
		LogDir += ServiceNameHere;
		AddDirSuffix(LogDir);
	}

	LogDir += "Logs";
	if(!CreateDir(LogDir.c_str()))
	{
		delete Log;
		if(UseLocking)
			NV_UNLOCKOBJECT(this);
		return false;
	}

	AddDirSuffix(LogDir);
	LogDir += NameHere;

	if(Erase)
		RemoveDir(LogDir.c_str());

	if(!CreateDir(LogDir.c_str()))
	{
		delete Log;
		if(UseLocking)
			NV_UNLOCKOBJECT(this);
		return false;
	}
	AddDirSuffix(LogDir);

	Log->m_sLogDir = LogDir;
	Log->m_sName = NameHere;

	Log->CheckEmptyDirs();

	string ImportFileName = "";
	if(ImportFile != NULL)
		ImportFileName = ImportFile;
	CleanString(ImportFileName);
	if(ImportFileName.length())
	{
		// We must import the given file and save it as nvs1.log
		// in the log directory

		// First verify that it's a correct log file
		bool bValidFile = false;
		FILE *stream = Nvfopen(ImportFileName.c_str(), "r");
		if(stream != NULL)
		{
			char Buffer[256];
			string Line;
			if(freadline(Buffer, sizeof(Buffer), stream))
			{
				Line = Buffer;
				CleanString(Line);
				if(Line == NV_LOG_FILE_IDENTIFIER)
					bValidFile = true;
			}
			fclose(stream);	       
		}
		// We always remove the log directory when importing a file
		RemoveDir(LogDir.c_str(), false);
		string LogFile1 = LogDir;
		LogFile1 += "nvs1.log";
		CopyFileEx(ImportFileName.c_str(), LogFile1.c_str());
	}

	m_vLogs[NameHere] = (void *)Log;
	
	Save();
	
	if(UseLocking)
		NV_UNLOCKOBJECT(this);
	
	return true;
}



// SetDescription
bool NvLogList::SetDescription(const char *Name, const char *Description)
{
	if(Description == NULL)
		return false;

	NV_LOCKOBJECT(this);
	NvLog *Log = InternalGetLog(Name);
	if(Log == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	// Wait for log to be unused
	while(Log->GetInUse())
		SLEEP(10);
	Log->m_sDesc = Description;
	Save();

	NV_UNLOCKOBJECT(this);
	return true;
}


bool NvLogList::SetLevel(const char *Name, int NewLevel)
{
	NV_LOCKOBJECT(this);
	NvLog *Log = InternalGetLog(Name);
	if(Log == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	// Wait for log to be unused
	while(Log->GetInUse())
		SLEEP(10);
	Log->m_iLevel = NewLevel;
	if(Log->m_iLevel < 0)
		Log->m_iLevel = 0;

	Save();
	
	NV_UNLOCKOBJECT(this);

	return true;
}


int NvLogList::GetLevel(const char *Name)
{
	int RetVal = 0;
	NV_LOCKOBJECT(this);
	NvLog *Log = InternalGetLog(Name);
	if(Log == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return RetVal;
	}
	RetVal = Log->m_iLevel;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}

// Test if log exist
bool NvLogList::Exist(const char *Name)
{
	NV_LOCKOBJECT(this);
	NvLog *Log = InternalGetLog(Name);
	if(Log == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	NV_UNLOCKOBJECT(this);
	return true;
}

// SetLogSaveFlag
bool NvLogList::SetSave(const char *Name, bool SaveIt)
{
	NV_LOCKOBJECT(this);
	NvLog *Log = InternalGetLog(Name);
	if(Log == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	// Wait for log to be unused
	while(Log->GetInUse())
		SLEEP(10);
	Log->m_bSaveIt = SaveIt;
	Save();
	NV_UNLOCKOBJECT(this);
	return true;
}

// SetFlushFlag
bool NvLogList::SetFlush(const char *Name, bool Flush)
{
	NV_LOCKOBJECT(this);
	NvLog *Log = InternalGetLog(Name);
	if(Log == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	// Wait for log to be unused
	while(Log->GetInUse())
		SLEEP(10);
	Log->m_bFlush = Flush;
	Save();
	NV_UNLOCKOBJECT(this);
	return true;
}

// SetLogMaxTime
bool NvLogList::SetMaxTime(const char *Name, int MaxTime)
{
	NV_LOCKOBJECT(this);
	NvLog *Log = InternalGetLog(Name);
	if(Log == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	// Wait for log to be unused
	while(Log->GetInUse())
		SLEEP(10);
	Log->m_iMaxFileTime = MaxTime;
	if(Log->m_iMaxFileTime <= 0)
		Log->m_iMaxFileTime = 1;
	Save();
	NV_UNLOCKOBJECT(this);
	return true;
}

// SetLogMaxSaveDays
bool NvLogList::SetMaxSaveDays(const char *Name, int MaxDays)
{
	NV_LOCKOBJECT(this);
	NvLog *Log = InternalGetLog(Name);
	if(Log == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	// Wait for log to be unused
	while(Log->GetInUse())
		SLEEP(10);
	Log->m_iMaxSaveDays = MaxDays;
	if(Log->m_iMaxSaveDays <= 0)
		Log->m_iMaxSaveDays = 1;
	Save();
	NV_UNLOCKOBJECT(this);
	return true;
}



void NvLogList::RemoveLog(const char *Name, bool Erase, bool ProtectSystemLog)
{
	string NameHere = Name;
	NvLog::CleanLogName(NameHere);
	
	// We cannot remove the system log
	if(ProtectSystemLog)
		if(NameHere == "SYSTEM")
			return;

	NV_LOCKOBJECT(this);
	NvLog *Log = InternalGetLog(Name);
	if(Log == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return;
	}
	// Wait for log to be unused
	while(Log->GetInUse())
		SLEEP(10);

	m_vLogs.erase(NameHere);

	if(Erase)
	{
		RemoveDirSuffix(Log->m_sLogDir);
		RemoveDir(Log->m_sLogDir.c_str());
	}
	
	if(LastLog == Log)
		LastLog = NULL;
	delete Log;

	Save();
	NV_UNLOCKOBJECT(this);
}



bool NvLogList::GetInfo(const char *Name, NvObjTable *LogInfo)
{
	LogInfo->Clear();

	bool AllLogs = true;

	NV_LOCKOBJECT(this);

	NvLog *FoundLog = NULL;
	if((Name != NULL) && strlen(Name))
	{
		// There is a request on a specific log
		FoundLog = InternalGetLog(Name);
		if(FoundLog == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		AllLogs = false;
	}


	LogInfo->InsertColumn("NAME");
	LogInfo->InsertColumn("DESCRIPTION");
	LogInfo->InsertColumn("LEVEL");
	LogInfo->InsertColumn("MAXTIME");
	LogInfo->InsertColumn("SAVE");
	LogInfo->InsertColumn("MAXSAVEDAYS");
	LogInfo->InsertColumn("FLUSH");

	if(AllLogs)
	{
		int j;
		map<string,void *>::iterator pos;
		for(pos = m_vLogs.begin(), j=0; pos != m_vLogs.end(); ++pos)
		{
			NvLog *Log = (NvLog *)(pos->second);
			if(Log == NULL)
				continue;
			LogInfo->InsertRow();
			LogInfo->InsertCellLine(j, 0, Log->m_sName);
			LogInfo->InsertCellLine(j, 1, Log->m_sDesc);
			char Buffer[20];
			sprintf(Buffer, "%d", Log->m_iLevel);
			LogInfo->InsertCellLine(j, 2, Buffer);
			sprintf(Buffer, "%d", Log->m_iMaxFileTime);
			LogInfo->InsertCellLine(j, 3, Buffer);
			if(Log->m_bSaveIt)
				LogInfo->InsertCellLine(j, 4, "YES");
			else
				LogInfo->InsertCellLine(j, 4, "NO");
			sprintf(Buffer, "%d", Log->m_iMaxSaveDays);
			LogInfo->InsertCellLine(j, 5, Buffer);
			if(Log->m_bFlush)
				LogInfo->InsertCellLine(j, 6, "YES");
			else
				LogInfo->InsertCellLine(j, 6, "NO");
			j++;
		}
	}
	else
	{
		LogInfo->InsertRow();
		LogInfo->InsertCellLine(0, 0, FoundLog->m_sName);
		LogInfo->InsertCellLine(0, 1, FoundLog->m_sDesc);
		char Buffer[20];
		sprintf(Buffer, "%d", FoundLog->m_iLevel);
		LogInfo->InsertCellLine(0, 2, Buffer);
		sprintf(Buffer, "%d", FoundLog->m_iMaxFileTime);
		LogInfo->InsertCellLine(0, 3, Buffer);
		if(FoundLog->m_bSaveIt)
			LogInfo->InsertCellLine(0, 4, "YES");
		else
			LogInfo->InsertCellLine(0, 4, "NO");
		sprintf(Buffer, "%d", FoundLog->m_iMaxSaveDays);
		LogInfo->InsertCellLine(0, 5, Buffer);
		if(FoundLog->m_bFlush)
			LogInfo->InsertCellLine(0, 6, "YES");
		else
			LogInfo->InsertCellLine(0, 6, "NO");
	}
	NV_UNLOCKOBJECT(this);

	return true;
}



bool NvLogList::Write(const char *Name, const char *Message, const char *Extra, int Type, int Level, const char *Identifier, const char *Source, bool Flush)
{
	NV_LOCKOBJECT(this);
	NvLog *Log = InternalGetLog(Name);
	if(Log == NULL)
	{
		// Try reloading the log table
		Load();
		Log = InternalGetLog(Name);
	}
	if(Log == NULL)
	{
		// The log doesn't exist, we must create it
		if(!Create(Name, false))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		// Now we should be able to get it
		Log = InternalGetLog(Name);
		if(Log == NULL)
		{
			// Should not occur
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	Log->IncInuse();

	// The log is now in use so we can unlock
	NV_UNLOCKOBJECT(this);


	// Let's the Log do the writting of the message here
	Log->Write(Message, Extra, Type, Level, Identifier, Source, Flush);

	// We can now free the Log. This is made outside list locking because all list
	// function that modifies the log wait for it to be unused
	Log->DecInUse();

	return true;
}

bool NvLogList::WriteFlush(const char *Name, const char *Message, const char *Extra, int Type, int Level, const char *Identifier, const char *Source)
{
	return Write(Name, Message, Extra, Type, Level, Identifier, Source, true);
}


bool NvLogList::Search(const char *Name,
					   NvObjTable *LogView,
					   const char* FromDate,
					   const char* ToDate,
					   const char* FromHour,
					   const char* ToHour,
					   const char* Identifier,
					   int Type,
					   const char* Message,
					   const char* Extra,
					   const char* Source,
					   int Level,
					   const char* ExportFile,
					   int MaxFileSize)
{
	LogView->Clear();
	NV_LOCKOBJECT(this);
	NvLog *Log = InternalGetLog(Name);
	if(Log == NULL)
	{
		// The log should exist
		NV_UNLOCKOBJECT(this);
		return false;
	}

	Log->IncInuse();

	// The log is now in use so we can unlock
	NV_UNLOCKOBJECT(this);

	LogView->InsertColumn("DATE", "Date", true);
	LogView->InsertColumn("TIME", "Time", true);
	LogView->InsertColumn("IDENTIFIER", "Identifier");
	LogView->InsertColumn("TYPE", "Type", true);
	LogView->InsertColumn("MESSAGE", "Message");
	LogView->InsertColumn("EXTRA", "Extra");
	LogView->InsertColumn("SOURCE", "Source");
	LogView->InsertColumn("LEVEL", "Level", true);
	LogView->InsertColumn("THREAD", "Level", true);

	// Let's the Log do the search
	if(!Log->Search(LogView, FromDate, ToDate, FromHour, ToHour, Identifier, Type, Message, Extra, Source, Level, ExportFile, MaxFileSize))
	{
		Log->DecInUse();
		return false;
	}

	// We can now free the Log. This is made outside list locking because all list
	// function that modifies the log wait for it to be unused
	Log->DecInUse();

	return true;
}



bool NvLogList::RemoveLogData(const char *Name, const char* FromDate, const char* ToDate, bool ClearIt)
{
	NV_LOCKOBJECT(this);
	NvLog *Log = InternalGetLog(Name);
	if(Log == NULL)
	{
		// The log should exist
		NV_UNLOCKOBJECT(this);
		return false;
	}

	Log->IncInuse();

	// The log is now in use so we can unlock
	NV_UNLOCKOBJECT(this);

	// Let's the Log do the search
	if(!Log->RemoveLogData(FromDate, ToDate, true, ClearIt))
	{
		Log->DecInUse();
		return false;
	}

	// We can now free the Log. This is made outside list locking because all list
	// function that modifies the log wait for it to be unused
	Log->DecInUse();

	return true;
}



/////////////////////////////////////////////////////////////////////////////
// NvLog

NvLog::NvLog()
{
	InUse = 0;
	m_sName = "";
	m_sDesc = "";
	m_iMaxFileTime = 10;
	m_bSaveIt = false; 
	m_sLogDir = ""; 
	m_iLevel = 1; 
	m_iLogFileStartTime = 0;
	m_iMaxSaveDays = 2;
	logstream = NULL;
	m_bFlush = true; 
}

NvLog::~NvLog()
{
	NV_LOCKOBJECT(this);
	if(logstream != NULL)
		fclose(logstream);
	logstream = NULL;
	NV_UNLOCKOBJECT(this);
}


bool NvLog::IncInuse()
{
	NV_LOCKLOCK(&ProtectInUse);
	InUse++;
	NV_UNLOCKLOCK(&ProtectInUse);

	return true;
}

void NvLog::DecInUse()
{
	NV_LOCKLOCK(&ProtectInUse);
	if(InUse)
		InUse--;
	NV_UNLOCKLOCK(&ProtectInUse);
}

int NvLog::GetInUse()
{
	int RetVal = 0;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = InUse;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}




void NvLog::CleanLogName(string& Name)
{
	MakeUpper(Name);
	CleanString(Name, true);
	string::size_type index;
	// remove backslashes
	while((index = Name.find('\\')) != string::npos)
		Name.erase(index, 1);
	// remove slashes
	while((index = Name.find('/')) != string::npos)
		Name.erase(index, 1);
	// remove double quotes
	while((index = Name.find('\"')) != string::npos)
		Name.erase(index, 1);
}


bool NvLog::Write(const char *Message, const char *Extra, int Type, int Level, const char *Identifier, const char *Source, bool Flush)
{
	// The requested level must be less or equal to the actual log level
	if(Level <= 0)
		return true;

	if(Level > m_iLevel)
		return true;

    if(Message == NULL)		// must not be NULL
		return false;

	if(m_bFlush)		// We force flush if the option has been set
		Flush = true;
	
	// The writing is locked
	NV_LOCKOBJECT(this);

	string MessageHere = Message;
	// replace all \n with  character
	string::size_type index;
	while((index = MessageHere.find('\n')) != string::npos)
		MessageHere.replace(index, 1, 1, '');

	CleanString(MessageHere);
	string LogLine = "";

	// get current time and date
	timeval tv;
	Mygettimeofday(&tv, NULL);
	int milli = tv.tv_usec/1000;
	time_t CurrentTime = (time_t)(tv.tv_sec);
	int ActualCurrentTime = int(CurrentTime);
	struct tm tCurrentTime;
 	localtime_r(&CurrentTime, &tCurrentTime);

	char sBuf[50];
	// Write date
	sprintf(sBuf, "%04d%02d%02d",
					tCurrentTime.tm_year + 1900,
					tCurrentTime.tm_mon + 1,
					tCurrentTime.tm_mday); 
	LogLine += sBuf;
	LogLine += NV_LOG_FIELD_SEPARATOR;
	
	// Write hour
	sprintf(sBuf, "%02d%02d%02d.%03d",
					tCurrentTime.tm_hour,
					tCurrentTime.tm_min,
					tCurrentTime.tm_sec,milli); 
	LogLine += sBuf;
	LogLine += NV_LOG_FIELD_SEPARATOR;

	// Write identifier
	if(Identifier != NULL)
	{
		string IdentifierHere = Identifier;
		CleanString(IdentifierHere);
		LogLine += IdentifierHere;
	}
	LogLine += NV_LOG_FIELD_SEPARATOR;

	// Write type
	if(Type > NV_LOG_UNKNOWNTYPE)
		Type = NV_LOG_UNKNOWNTYPE;
	sprintf(sBuf, "%d", Type); 
	LogLine += sBuf;
	LogLine += NV_LOG_FIELD_SEPARATOR;

	// Write message
	LogLine += MessageHere;
	LogLine += NV_LOG_FIELD_SEPARATOR;

	// Write extra information
	if(Extra != NULL)
	{
		string ExtraHere = Extra;
		while((index = ExtraHere.find('\n')) != string::npos)
			ExtraHere.replace(index, 1, 1, '');
		CleanString(ExtraHere);
		LogLine += ExtraHere;
	}
	LogLine += NV_LOG_FIELD_SEPARATOR;

	// Write source
	if(Source != NULL)
	{
		string SourceHere = Source;
		CleanString(SourceHere);
		LogLine += SourceHere;
	}
	LogLine += NV_LOG_FIELD_SEPARATOR;
	
	// Write level
	sprintf(sBuf, "%d", Level); 
	LogLine += sBuf;
	
	// Write thread
	LogLine += NV_LOG_FIELD_SEPARATOR;
	sprintf(sBuf, "%lu", NvGetCurrentThreadId()); 
	LogLine += sBuf;

	if(LogLine.length() > (NV_LOG_MAX_LINE_LENGTH+2))
		LogLine = LogLine.substr(0, NV_LOG_MAX_LINE_LENGTH-2);

#ifdef WIN32
	LogLine += "\r";
#endif
	LogLine += "\n";

	string LogFile = m_sLogDir;
	LogFile += "nvs.log";
	if(m_iLogFileStartTime == 0)
	{
		// If this is the first call to the file
		// We must check if the file exist and get its start time if yes
		// By default, the start time is the same than the first write
		if(logstream != NULL)
			fclose(logstream);
		logstream = NULL;
		m_iLogFileStartTime = ActualCurrentTime;
		if(FileExist(LogFile.c_str()))
		{
			string sStartFromFile = GetLogHeader(LogFile.c_str(), "iStartTime");
			if(sStartFromFile.length())
				m_iLogFileStartTime = atoi(sStartFromFile.c_str());
		}
	}
		
	//PMARC Modif for max log file time
	if(m_iMaxFileTime > 0)
	{
		int TimeFromStartLog = ActualCurrentTime - m_iLogFileStartTime;
		int LogFileMaxTimeInSeconds = m_iMaxFileTime*60;
		if(TimeFromStartLog > LogFileMaxTimeInSeconds)
		{
			// Its time now to reset log file
			if(logstream != NULL)
				fclose(logstream);
			logstream = NULL;
			string LogFile1 = m_sLogDir;
			LogFile1 += "nvs1.log";
			unlink(LogFile1.c_str());	// removes the second log file
			if(FileExist(LogFile.c_str()))
			{
				rename(LogFile.c_str(), LogFile1.c_str());
				unlink(LogFile.c_str());	// to be sure
			}
			m_iLogFileStartTime = ActualCurrentTime;

			// We must set now the ENDDATETIME log file header with the correct value
			if(FileExist(LogFile1.c_str()))
			{
				FILE *stream = Nvfopen(LogFile1.c_str(), "r+b");
				if(stream == NULL)
				{
					NV_UNLOCKOBJECT(this);
					return false;
				}

				char Buffer[256];
				string Line;
				if(!freadline(Buffer, sizeof(Buffer), stream))
				{
					fclose(stream);	       
					NV_UNLOCKOBJECT(this);
					return false;
				}
				Line = Buffer;
				CleanString(Line);
				if(Line != NV_LOG_FILE_IDENTIFIER)
				{
					fclose(stream);	       
					NV_UNLOCKOBJECT(this);
					return false;
				}
				int Offset = ftell(stream);
				while(freadline(Buffer, sizeof(Buffer), stream))
				{
					Line = Buffer;
					CleanString(Line);
					if(!Line.length())
						break;
					index = Line.find('=');
					if(index == string::npos)
						break;
					string HeaderInFile = Line.substr(0, index);
					CleanString(HeaderInFile);
					MakeUpper(HeaderInFile);
					if(HeaderInFile == "ENDDATETIME")
						break;
					Offset = ftell(stream);
				}

				fseek(stream, Offset, SEEK_SET);
			
				string LineToWrite = "ENDDATETIME = ";
				// Write date
				sprintf(sBuf, "%04d%02d%02d%02d%02d%02d",
								tCurrentTime.tm_year + 1900,
								tCurrentTime.tm_mon + 1,
								tCurrentTime.tm_mday,
								tCurrentTime.tm_hour,
								tCurrentTime.tm_min,
								tCurrentTime.tm_sec); 
				LineToWrite += sBuf;
			#ifdef WIN32
				LineToWrite += "\r";
			#endif
				LineToWrite += "\n";
				fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), stream);

				fclose(stream);	    
			}		
			
			// Now save log file if requested
			if(m_bSaveIt && FileExist(LogFile1.c_str()))
			{
				char SaveLogFile[512];
				sprintf(SaveLogFile, "%sSaves", m_sLogDir.c_str());
				CreateDir(SaveLogFile);
#ifdef WIN32
				sprintf(SaveLogFile, "%sSaves\\%4d", m_sLogDir.c_str(), tCurrentTime.tm_year + 1900);
				CreateDir(SaveLogFile);
				sprintf(SaveLogFile, "%sSaves\\%4d\\%02d", m_sLogDir.c_str(), tCurrentTime.tm_year + 1900, tCurrentTime.tm_mon + 1);
				CreateDir(SaveLogFile);
				sprintf(SaveLogFile, "%sSaves\\%4d\\%02d\\%02d", m_sLogDir.c_str(), tCurrentTime.tm_year + 1900, tCurrentTime.tm_mon + 1, tCurrentTime.tm_mday);
				CreateDir(SaveLogFile);
				sprintf(SaveLogFile, "%sSaves\\%4d\\%02d\\%02d\\%02d%02d%02d.log", m_sLogDir.c_str(),
								tCurrentTime.tm_year + 1900,
								tCurrentTime.tm_mon + 1,
								tCurrentTime.tm_mday,
								tCurrentTime.tm_hour,
								tCurrentTime.tm_min,
								tCurrentTime.tm_sec); 
#else
				sprintf(SaveLogFile, "%sSaves/%4d", m_sLogDir.c_str(), tCurrentTime.tm_year + 1900);
				CreateDir(SaveLogFile);
				sprintf(SaveLogFile, "%sSaves/%4d/%02d", m_sLogDir.c_str(), tCurrentTime.tm_year + 1900, tCurrentTime.tm_mon + 1);
				CreateDir(SaveLogFile);
				sprintf(SaveLogFile, "%sSaves/%4d/%02d/%02d", m_sLogDir.c_str(), tCurrentTime.tm_year + 1900, tCurrentTime.tm_mon + 1, tCurrentTime.tm_mday);
				CreateDir(SaveLogFile);
				sprintf(SaveLogFile, "%sSaves/%4d/%02d/%02d/%02d%02d%02d.log", m_sLogDir.c_str(),
								tCurrentTime.tm_year + 1900,
								tCurrentTime.tm_mon + 1,
								tCurrentTime.tm_mday,
								tCurrentTime.tm_hour,
								tCurrentTime.tm_min,
								tCurrentTime.tm_sec); 
#endif
				CopyFileEx(LogFile1.c_str(), SaveLogFile);

				// After the save, we must cleanup the eventual too log information
				char Buf[10];
				sprintf(Buf,"-%d", m_iMaxSaveDays); 
				RemoveLogData("", Buf, false);
			}
		}
	}


	string sOpenType = "at";

	bool JustCreated = false;
	if(!FileExist(LogFile.c_str()))
		JustCreated = true;

	if(JustCreated)
		sOpenType = "w";

	if(logstream == NULL)
	{
		logstream = Nvfopen(LogFile.c_str(), sOpenType.c_str());
		if(logstream == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}


//	FILE *stream = Nvfopen(LogFile.c_str(), sOpenType.c_str());
//	if(stream == NULL)
//	{
//		NV_UNLOCKOBJECT(this);
//		return false;
//	}

	if(JustCreated)
	{
		string LineToWrite = NV_LOG_FILE_IDENTIFIER;
#ifdef WIN32
		LineToWrite += "\r";
#endif
		LineToWrite += "\n";
		fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), logstream);

		LineToWrite = "VERSION = ";
		LineToWrite += NV_LOG_FILE_VERSION;
#ifdef WIN32
		LineToWrite += "\r";
#endif
		LineToWrite += "\n";
		fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), logstream);
		
		LineToWrite = "STARTDATETIME = ";
		// Write date
		sprintf(sBuf, "%04d%02d%02d%02d%02d%02d",
						tCurrentTime.tm_year + 1900,
						tCurrentTime.tm_mon + 1,
						tCurrentTime.tm_mday,
						tCurrentTime.tm_hour,
						tCurrentTime.tm_min,
						tCurrentTime.tm_sec); 
		LineToWrite += sBuf;
#ifdef WIN32
		LineToWrite += "\r";
#endif
		LineToWrite += "\n";
		fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), logstream);
		
		// We reserve 14 spaces for the end date time that will be written when saving the log file
		LineToWrite = "ENDDATETIME =               ";
#ifdef WIN32
		LineToWrite += "\r";
#endif
		LineToWrite += "\n";
		fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), logstream);
		
		// ISTARTTIME is used for storing the value of the m_iLogFileStartTime member
		// This is usefull after log creation or when the server is starting
		LineToWrite = "ISTARTTIME = ";
		// Write date
		sprintf(sBuf, "%ld", (unsigned long)ActualCurrentTime); 
		LineToWrite += sBuf;
#ifdef WIN32
		LineToWrite += "\r";
#endif
		LineToWrite += "\n";
		fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), logstream);

		// Finally, we write a blank line that delimitates the start of the real log file data
		LineToWrite = "";
#ifdef WIN32
		LineToWrite += "\r";
#endif
		LineToWrite += "\n";
		fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), logstream);
		fflush(logstream);
	
	}

	fwrite(LogLine.c_str(), 1, LogLine.length(), logstream);
	if(Flush)
		fflush(logstream);
//	fclose(stream);	       

	NV_UNLOCKOBJECT(this);

	return true;
}



bool NvLog::Search(NvObjTable *LogView,
					   const char* FromDate,
					   const char* ToDate,
					   const char* FromHour,
					   const char* ToHour,
					   const char* Identifier,
					   int Type,
					   const char* Message,
					   const char* Extra,
					   const char* Source,
					   int Level,
					   const char* ExportFile,
					   int MaxFileSize)
{
	string LogFile = m_sLogDir;
	LogFile += "nvs.log";
	string LogFile1 = m_sLogDir;
	LogFile1 += "nvs1.log";

	string TempFile = "";
	GetTemporaryFile(TempFile, m_sLogDir, ".tmp", "nvslog", "");
	if(!TempFile.length())
		return false;

	// We must search the correct files following the date and hour requests
	if(!BuildLogFile(TempFile.c_str(), MaxFileSize, FromDate, ToDate, FromHour, ToHour))
	{
		RemoveFile(TempFile.c_str());
		return false;
	}

	NvObjTable ProviTable;
	ProviTable.CreateFrom(LogView);
	bool SearchAll = true;

	string IdentifierHere = "";
	if(Identifier != NULL)
		IdentifierHere = Identifier;
	if(IdentifierHere.length())
		SearchAll = false;

	string MessageHere = "";
	if(Message != NULL)
		MessageHere = Message;
	if(MessageHere.length())
		SearchAll = false;

	string ExtraHere = "";
	if(Extra != NULL)
		ExtraHere = Extra;
	if(ExtraHere.length())
		SearchAll = false;

	string SourceHere = "";
	if(Source != NULL)
		SourceHere = Source;
	if(SourceHere.length())
		SearchAll = false;

	if(Type != NV_LOG_ALL)
		SearchAll = false;
	
	if(Level != 0)
		SearchAll = false;

	NvObjTable *TableForImport = LogView;
	if(!SearchAll)
		TableForImport = &ProviTable;

	if(!TableForImport->Import(TempFile.c_str(), false, NV_LOG_FIELD_SEPARATOR, "", NV_LOG_MAX_LINE_LENGTH))
	{
		RemoveFile(TempFile.c_str());
		return false;
	}

	RemoveFile(TempFile.c_str());

	if(!SearchAll)
	{
		string Query = "";
		if(IdentifierHere.length())
		{
			Query += "IDENTIFIER=";
			Query += IdentifierHere;
		}
		if(Level != 0)
		{
			if(Query.length())
				Query += " AND ";
			Query += "LEVEL<=";
			char Buf[20];
			sprintf(Buf, "%d", Level);
			Query += Buf;
		}
		if(Type != NV_LOG_ALL)
		{
			if(Query.length())
				Query += " AND ";
			Query += "TYPE=";
			char Buf[20];
			sprintf(Buf, "%d", Type);
			Query += Buf;
		}
		if(MessageHere.length())
		{
			if(Query.length())
				Query += " AND ";
			Query += "MESSAGE=";
			Query += MessageHere;
		}
		if(ExtraHere.length())
		{
			if(Query.length())
				Query += " AND ";
			Query += "EXTRA=";
			Query += ExtraHere;
		}
		if(SourceHere.length())
		{
			if(Query.length())
				Query += " AND ";
			Query += "SOURCE=";
			Query += SourceHere;
		}

		if(TableForImport->Search(Query.c_str(), LogView) < 0)
			return false;
	}

	string ExportFileName = "";
	if(ExportFile != NULL)
		ExportFileName = ExportFile;
	if(ExportFileName.length())
	{
		// Export the result to the given file

		// We must prepare the first and end date time for this file

		time_t CurrentTime;
		time(&CurrentTime);
		struct tm tCurrentTime;
 		localtime_r(&CurrentTime, &tCurrentTime);
		char sBuf[30];
		sprintf(sBuf, "%04d%02d%02d%02d%02d%02d",
						tCurrentTime.tm_year + 1900,
						tCurrentTime.tm_mon + 1,
						tCurrentTime.tm_mday,
						tCurrentTime.tm_hour,
						tCurrentTime.tm_min,
						tCurrentTime.tm_sec); 
		string CurrentDateTime = sBuf;
		string StartDateTime = CurrentDateTime;
		string EndDateTime = CurrentDateTime;

		// The Start time is given by the first record of the LogView table
		int NumRecords = LogView->GetNumRow();
		if(NumRecords > 0)
		{
			StartDateTime = LogView->GetCellLine(0, 0);
			StartDateTime += LogView->GetCellLine(0, 1);
		}
		if(NumRecords > 1)
		{
			EndDateTime = LogView->GetCellLine(NumRecords-1, 0);
			EndDateTime += LogView->GetCellLine(NumRecords-1, 1);
		}

		// First write the file header
		FILE *stream = Nvfopen(ExportFileName.c_str(), "w+b");
		if(stream == NULL)
			return false;
		
		string LineToWrite = NV_LOG_FILE_IDENTIFIER;
#ifdef WIN32
		LineToWrite += "\r";
#endif
		LineToWrite += "\n";
		fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), stream);

		LineToWrite = "VERSION = ";
		LineToWrite += NV_LOG_FILE_VERSION;
#ifdef WIN32
		LineToWrite += "\r";
#endif
		LineToWrite += "\n";
		fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), stream);
		
		// Write start date
		LineToWrite = "STARTDATETIME = ";
		LineToWrite += StartDateTime;
#ifdef WIN32
		LineToWrite += "\r";
#endif
		LineToWrite += "\n";
		fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), stream);
		
		// Write start date
		LineToWrite = "ENDDATETIME = ";
		LineToWrite += EndDateTime;
#ifdef WIN32
		LineToWrite += "\r";
#endif
		LineToWrite += "\n";
		fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), stream);
		
		// ISTARTTIME is used for storing the value of the m_iLogFileStartTime member
		// This is usefull after log creation or when the server is starting
		// But for imported exported files, it's not important so we let it blank
		LineToWrite = "ISTARTTIME = ";
#ifdef WIN32
		LineToWrite += "\r";
#endif
		LineToWrite += "\n";
		fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), stream);

		// Finally, we write a blank line that delimitates the start of the real log file data
		LineToWrite = "";
#ifdef WIN32
		LineToWrite += "\r";
#endif
		LineToWrite += "\n";
		fwrite(LineToWrite.c_str(), 1, LineToWrite.length(), stream);
		fclose(stream);	       

		GetTemporaryFile(TempFile, m_sLogDir, ".tmp", "nvslog", "");
		if(!LogView->Export(TempFile.c_str(), NV_LOG_FIELD_SEPARATOR, "|", NV_LOG_MAX_LINE_LENGTH))
		{
			RemoveFile(TempFile.c_str());
			return false;
		}

		if(!AppendFileEx(TempFile.c_str(), ExportFileName.c_str()))
		{
			RemoveFile(TempFile.c_str());
			return false;
		}
		RemoveFile(TempFile.c_str());

	}

	return true;
}


string NvLog::GetLogHeader(const char *LogFile, const char *Header)
{
	if(LogFile == NULL)
		return "";

	string HeaderHere = Header;
	CleanString(HeaderHere);
	MakeUpper(HeaderHere);
	string Result = "";
	FILE *stream = Nvfopen(LogFile, "r");
	if(stream == NULL)
		return Result;

	char Buffer[256];
	string Line;
	if(!freadline(Buffer, sizeof(Buffer), stream))
	{
		fclose(stream);	       
		return Result;
	}
	Line = Buffer;
	CleanString(Line);
	if(Line != NV_LOG_FILE_IDENTIFIER)
	{
		fclose(stream);	       
		return Result;
	}
	while(freadline(Buffer, sizeof(Buffer), stream))
	{
		Line = Buffer;
		CleanString(Line);
		if(!Line.length())
			break;
		string::size_type index = Line.find('=');
		if(index == string::npos)
			break;
		string HeaderInFile = Line.substr(0, index);
		CleanString(HeaderInFile);
		MakeUpper(HeaderInFile);
		if(HeaderInFile == HeaderHere)
		{
			Result = Line.substr(index+1);
			CleanString(Result);
			break;
		}
	}

	fclose(stream);	    
	
	return Result;
}


int NvLog::GetLogStartOffset(const char *LogFile)
{
	int RetVal = -1;

	if(LogFile == NULL)
		return RetVal;

	FILE *stream = Nvfopen(LogFile, "r");
	if(stream == NULL)
		return RetVal;

	char Buffer[256];
	string Line;
	if(!freadline(Buffer, sizeof(Buffer), stream))
	{
		fclose(stream);	       
		return RetVal;
	}
	Line = Buffer;
	CleanString(Line);
	if(Line != NV_LOG_FILE_IDENTIFIER)
	{
		fclose(stream);	       
		return RetVal;
	}
	while(freadline(Buffer, sizeof(Buffer), stream))
	{
		Line = Buffer;
		CleanString(Line);
		if(!Line.length())
			break;
		string::size_type index = Line.find('=');
		if(index == string::npos)
			break;
	}

	RetVal = (int)ftell(stream);
	fclose(stream);	    
	
	return RetVal;
}


// This function is used internally to build a single log file list from the archive following a date and
// hour range.
// The default is to get the log for the last 15 minutes.
// MaxFileSize is in kilobytes
// The fromdate and todate must be on the form dd/mm/yyyy or dd.mm/yyyy
// The FromHour and ToHour must be on the form hhmmss or hh:mm:ss
bool NvLog::BuildLogFile(const char* FileOut, int MaxFileSize, const char* FromDate, const char* ToDate, const char* FromHour, const char* ToHour)
{
	int MaxFoundSize = 0;
	vector<string> List;
	char Buffer[1024];
	List.clear();

	string ToDateHere = "";
	if(ToDate != NULL)
		ToDateHere = ToDate;
	CleanString(ToDateHere);
	if(ToDateHere.length())
	{
		if(!ArrangeDate(ToDateHere))
			return false;
	}

	string FromDateHere = "";
	if(FromDate != NULL)
		FromDateHere = FromDate;
	CleanString(FromDateHere);
	if(FromDateHere.length())
	{
		if(!ArrangeDate(FromDateHere))
			return false;
	}
	else
		ToDateHere = "";


	int iMinHour = 0;
	int iMaxHour = 235959;
	string ToHourHere = "";
	if(ToHour != NULL)
		ToHourHere = ToHour;
	CleanString(ToHourHere);
	if(ToHourHere.length())
	{
		if(!ArrangeHour(ToHourHere))
			return false;
	}

	string FromHourHere = "";
	if(FromHour != NULL)
		FromHourHere = FromHour;
	CleanString(FromHourHere);
	if(FromHourHere.length())
	{
		if(!ArrangeHour(FromHourHere))
			return false;
	}


	int iFromDay = 0;
	int iFromMonth = 0;
	int iFromYear = 0;
	// By default, we get the actual day logs
	time_t tTime = time(NULL);
	struct tm tCurrentTime;
	localtime_r(&tTime, &tCurrentTime);
	iFromDay = tCurrentTime.tm_mday;
	iFromMonth = tCurrentTime.tm_mon + 1;
	iFromYear = tCurrentTime.tm_year + 1900;

	if(ToHourHere.length())
	{
		// If a up range is given, we take it
		iMaxHour = atoi(ToHourHere.c_str());
	}
	else
	{
		// Otherwise, the up range is the actual time if the FromDate is given
		if(!FromDateHere.length())
		{
			sprintf(Buffer, "%02d%02d%02d", tCurrentTime.tm_hour, tCurrentTime.tm_min, tCurrentTime.tm_sec);
			iMaxHour = atoi(Buffer);
		}
	}


	if(FromHourHere.length())
	{
		// If a down range is given, we take it
		// Otherwise, the down range is let to 0
		iMinHour = atoi(FromHourHere.c_str());
	}
	else
	{
		// If none of from hour and from date are given
		// we take 15 minutes before the actual time
		if(!FromDateHere.length())
		{
			// 15 minutes before
			tTime = tTime - 900;
			localtime_r(&tTime, &tCurrentTime);
			sprintf(Buffer, "%02d.%02d.%04d", tCurrentTime.tm_mday, tCurrentTime.tm_mon + 1, tCurrentTime.tm_year + 1900);
			FromDateHere = Buffer;
			sprintf(Buffer, "%02d%02d%02d", tCurrentTime.tm_hour, tCurrentTime.tm_min, tCurrentTime.tm_sec);
			FromHourHere = Buffer;
			iMinHour = atoi(FromHourHere.c_str());
			tTime = time(NULL);
			localtime_r(&tTime, &tCurrentTime);
		}
	}

	string FromDateHereHere = FromDateHere;
	if(FromDateHereHere.length())
	{
		string Day = "";
		string Month = "";
		string Year = "";
		string::size_type index;
		index = FromDateHereHere.find('.');
		if(index != string::npos)
		{
			Day = FromDateHereHere.substr(0, index);
			FromDateHereHere = FromDateHereHere.substr(index+1);
		}
		index = FromDateHereHere.find('.');
		if(index != string::npos)
		{
			Month = FromDateHereHere.substr(0, index);
			Year = FromDateHereHere.substr(index+1);
		}


		iFromDay = atoi(Day.c_str());
		iFromMonth = atoi(Month.c_str());
		iFromYear = atoi(Year.c_str());
	}


	int iToDay = iFromDay;
	int iToMonth = iFromMonth;
	int iToYear = iFromYear;
	string ToDateHereHere = ToDateHere;
	if(ToDateHereHere.length())
	{
		string Day = "";
		string Month = "";
		string Year = "";
		string::size_type index;
		index = ToDateHereHere.find('.');
		if(index != string::npos)
		{
			Day = ToDateHereHere.substr(0, index);
			ToDateHereHere = ToDateHereHere.substr(index+1);
		}
		index = ToDateHereHere.find('.');
		if(index != string::npos)
		{
			Month = ToDateHereHere.substr(0, index);
			Year = ToDateHereHere.substr(index+1);
		}


		iToDay = atoi(Day.c_str());
		iToMonth = atoi(Month.c_str());
		iToYear = atoi(Year.c_str());
	}

	string SaveLogDir = m_sLogDir;
	AddDirSuffix(SaveLogDir);
	SetDirName(SaveLogDir);
	SaveLogDir += "Saves";
	AddDirSuffix(SaveLogDir);

	sprintf(Buffer, "%04d%02d", iFromYear, iFromMonth);
	int iMinYearmonth = atoi(Buffer);
	sprintf(Buffer, "%04d%02d", iToYear, iToMonth);
	int iMaxYearmonth = atoi(Buffer);
	sprintf(Buffer, "%04d%02d%02d", iFromYear, iFromMonth, iFromDay);
	int iMinYearmonthday = atoi(Buffer);
	sprintf(Buffer, "%04d%02d%02d", iToYear, iToMonth, iToDay);
	int iMaxYearmonthday = atoi(Buffer);
	if(iMinYearmonthday > iMaxYearmonthday)
		return true;	// nothing to do
	if(iMinYearmonthday == iMaxYearmonthday)
		if(iMinHour > iMaxHour)
			return true;	// nothing to do


	vector<string> directories;
	vector<string> years;
	
	// We must lock as shorter as possible
	// but we do it for building the file list
	NV_LOCKOBJECT(this);
	DirList(SaveLogDir.c_str(), NULL, &years);
	int i;
	for(i=0; i<(int)(years.size()); i++)
	{
		// Eliminate bad directories
		string YearHere = years[i];
		if(YearHere.length() != 4)
			continue;
		int iYearHere = atoi(YearHere.c_str());
		if((iYearHere < iFromYear) || (iYearHere > iToYear))
			continue;
		if((iYearHere < 1900) || (iYearHere > 2100))
			continue;

		// Now check for months
		vector<string> months;
		string SaveLogDirYear = SaveLogDir;
		SaveLogDirYear += YearHere;
		AddDirSuffix(SaveLogDirYear);
		// We must lock as shorter as possible
		DirList(SaveLogDirYear.c_str(), NULL, &months);
		int j;
		for(j=0; j<(int)(months.size()); j++)
		{
			// Eliminate bad directories
			string MonthHere = months[j];
			if(MonthHere.length() != 2)
				continue;

			string YearMonthHere = YearHere;
			YearMonthHere += MonthHere;
			int iYearMonthHere = atoi(YearMonthHere.c_str());
			if((iYearMonthHere < iMinYearmonth) || (iYearMonthHere > iMaxYearmonth))
				continue;

			// Now check for days
			vector<string> days;
			string SaveLogDirYearMonth = SaveLogDirYear;
			SaveLogDirYearMonth += MonthHere;
			AddDirSuffix(SaveLogDirYearMonth);
			// We must lock as shorter as possible
			DirList(SaveLogDirYearMonth.c_str(), NULL, &days);
			int k;
			for(k=0; k<(int)(days.size()); k++)
			{
				// Eliminate bad directories
				string DayHere = days[k];
				if(DayHere.length() != 2)
					continue;

				string YearMonthDayHere = YearMonthHere;
				YearMonthDayHere += DayHere;
				int iYearMonthDayHere = atoi(YearMonthDayHere.c_str());
				if((iYearMonthDayHere < iMinYearmonthday) || (iYearMonthDayHere > iMaxYearmonthday))
					continue;

				// The directory is now candidate so we can add it to the list
				string DirToAdd = YearHere;
				AddDirSuffix(DirToAdd);
				DirToAdd += MonthHere;
				AddDirSuffix(DirToAdd);
				DirToAdd += DayHere;
				AddDirSuffix(DirToAdd);
				directories.push_back(DirToAdd);

			}
		}
	}

	// We don't need that any more
	years.clear();

	// Now in directories, we have all the candidate directories (starting for the Log Saves directory)
	// We must now check for the files (hour parameter)
	// We just need to check the first and last directories if there are ones

	NvObjStringList files;
	bool bNextFileInSameDir = false;	// Used to check if we have to search for the next file or not
										// In fact, we must do the search if the last choosen file is the last
										// of the last directory
	for(i=0; i<(int)(directories.size()); i++)
	{
		string FullDirName = m_sLogDir;
		AddDirSuffix(FullDirName);
		SetDirName(FullDirName);
		FullDirName += "Saves";
		AddDirSuffix(FullDirName);
		FullDirName += directories[i];
		AddDirSuffix(FullDirName);
		files.RemoveAll();
		int j;
		// We must lock as shorter as possible
		if(!DirExist(FullDirName.c_str()))
		{
			NV_UNLOCKOBJECT(this);
			continue;
		}
		
		FileList(FullDirName.c_str(), NULL, &(files.m_vList));

		// We must sort the file because the file names are the writting date
		files.Sort(true);
		if(i == 0)
		{
			// First dir
			// We must get files greater than the start time
			for(j=0; j<files.GetSize(); j++)
			{
				string FileNameHere = files.GetValue(j);
				if(FileNameHere.length() != 10)
					continue;	// The file name must be on the form HHMMSS
				int iHourHere = atoi(FileNameHere.c_str());
				if(iHourHere < iMinHour)
					continue;
				string DirFileToAdd = m_sLogDir;
				AddDirSuffix(DirFileToAdd);
				SetDirName(DirFileToAdd);
				DirFileToAdd += "Saves";
				AddDirSuffix(DirFileToAdd);
				DirFileToAdd += directories[i];
				AddDirSuffix(DirFileToAdd);
				DirFileToAdd += FileNameHere;
				List.push_back(DirFileToAdd);
				if(directories.size() == 1)
				{
					if(iHourHere > iMaxHour)
					{
						bNextFileInSameDir = true;
						break;
					}
				}
			}
			continue;
		}
		if(i == directories.size()-1)
		{
			// last dir
			// We must get files smaller or equal to the end time
			for(j=0; j<files.GetSize(); j++)
			{
				string FileNameHere = files.GetValue(j);
				if(FileNameHere.length() != 10)
					continue;	// The file name must be on the form HHMMSS
				int iHourHere = atoi(FileNameHere.c_str());
				if(directories.size() == 1)
				{
					if(iHourHere < iMinHour)
						continue;
				}
				string DirFileToAdd = m_sLogDir;
				AddDirSuffix(DirFileToAdd);
				SetDirName(DirFileToAdd);
				DirFileToAdd += "Saves";
				AddDirSuffix(DirFileToAdd);
				DirFileToAdd += directories[i];
				AddDirSuffix(DirFileToAdd);
				DirFileToAdd += FileNameHere;
				List.push_back(DirFileToAdd);
				if(iHourHere > iMaxHour)
				{
					bNextFileInSameDir = true;
					break;
				}
			}
			continue;
		}

		// For all other directories
		// We just have to add the files
		for(j=0; j<files.GetSize(); j++)
		{
			string FileNameHere = files.GetValue(j);
			if(FileNameHere.length() != 10)
				continue;	// The file name must be on the form HHMMSS
			string DirFileToAdd = m_sLogDir;
			AddDirSuffix(DirFileToAdd);
			SetDirName(DirFileToAdd);
			DirFileToAdd += "Saves";
			AddDirSuffix(DirFileToAdd);
			DirFileToAdd += directories[i];
			AddDirSuffix(DirFileToAdd);
			DirFileToAdd += FileNameHere;
			List.push_back(DirFileToAdd);
		}

	}

	// Now, we must also get the following file because
	// This file can also contain some information on the hour range requested
	// We do that only if the last file of the directory has been added to the list

	// We must lock as shorter as possible
	if(!bNextFileInSameDir)
	{
		// Search for the next file in the following days
		int NextYear = iToYear;
		int NextMonth = iToMonth;
		int NextDay = iToDay;
		sprintf(Buffer, "%04d%02d%02d", tCurrentTime.tm_year + 1900, tCurrentTime.tm_mon + 1, tCurrentTime.tm_mday);
		int ActualYearMonthDay = atoi(Buffer);
		while(1)
		{
			NextDay += 1;
			if(NextDay > 31)
			{
				NextDay = 1;
				NextMonth += 1;
			}
			if(NextMonth > 12)
			{
				NextMonth = 1;
				NextYear += 1;
			}
			sprintf(Buffer, "%04d%02d%02d", NextYear, NextMonth, NextDay);
			int NextYearMonthDay = atoi(Buffer);
			
			// We stop the search if the file is not found and the next day is a future day
			if(NextYearMonthDay > ActualYearMonthDay)
				break;

			// Now, we must search the first file
			sprintf(Buffer, "%s%04d/%02d/%02d", SaveLogDir.c_str(), NextYear, NextMonth, NextDay);
			string DirToSearch = Buffer;
			SetDirName(DirToSearch);
			if(!DirExist(DirToSearch.c_str()))
				continue;
			files.RemoveAll();
			int j;
			FileList(DirToSearch.c_str(), NULL, &(files.m_vList));
			if(!files.GetSize())
				continue;
			files.Sort(true);
			string FileFound = "";
			for(j=0; j<files.GetSize(); j++)
			{
				string FileNameHere = files.GetValue(j);
				if(FileNameHere.length() == 10)
				{
					// The file name must be on the form HHMMSS
					FileFound = FileNameHere;
					break;
				}
			}
			if(!FileFound.length())
				continue;

			string DirFileToAdd = DirToSearch;
			AddDirSuffix(DirFileToAdd);
			DirFileToAdd += FileFound;
			List.push_back(DirFileToAdd);
			break;
		}
	}

	// Now we have the file list

	// We must eventually add the nvs1.log if it's not ever in the list and if it's required
	// and also the nvs.log if it's required

	string LogFile = m_sLogDir;
	LogFile += "nvs.log";

	string LogFile1 = m_sLogDir;
	LogFile1 += "nvs1.log";

	// We work on file copies in order to not block a too long time
	// the processing

	string TempLogFile = "";
	GetTemporaryFile(TempLogFile, m_sLogDir, ".tmp", "nvslog", "");
	if(!TempLogFile.length())
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	string TempLogFile1 = "";
	GetTemporaryFile(TempLogFile1, m_sLogDir, ".tmp", "nvslog", "");
	if(!TempLogFile1.length())
	{
		NV_UNLOCKOBJECT(this);
		RemoveFile(TempLogFile.c_str());
		return false;
	}

	if(logstream != NULL)
		fclose(logstream);
	logstream = NULL;
	if(FileExist(LogFile.c_str()))
	{
		if(!CopyFileEx(LogFile.c_str(), TempLogFile.c_str()))
		{
			NV_UNLOCKOBJECT(this);
			RemoveFile(TempLogFile.c_str());
			RemoveFile(TempLogFile1.c_str());
			return false;
		}
	}

	if(FileExist(LogFile1.c_str()))
	{
		if(!CopyFileEx(LogFile1.c_str(), TempLogFile1.c_str()))
		{
			NV_UNLOCKOBJECT(this);
			RemoveFile(TempLogFile.c_str());
			RemoveFile(TempLogFile1.c_str());
			return false;
		}
	}



	// Check now the nvs1 log file (in fact a copy of it)
	string sStartFromNvs1File = "";
	string sEndFromNvs1File = "";
	sStartFromNvs1File = GetLogHeader(TempLogFile1.c_str(), "STARTDATETIME");
	sEndFromNvs1File = GetLogHeader(TempLogFile1.c_str(), "ENDDATETIME");

	NV_UNLOCKOBJECT(this);

	string sStartFromLastFile = "";
	string sEndFromLastFile = "";
	bool TakeIt = false;
	if(List.size())
	{
		sStartFromLastFile = GetLogHeader(List[List.size()-1].c_str(), "STARTDATETIME");
		sEndFromLastFile = GetLogHeader(List[List.size()-1].c_str(), "ENDDATETIME");
	}
	// For the nvs1.log file, we must check that we not have it yet in our file list

	if((sStartFromNvs1File != sStartFromLastFile) || (sEndFromNvs1File != sEndFromLastFile))
	{
		TakeIt = true;
		if((sStartFromNvs1File.length() != 14) || (sEndFromNvs1File.length() != 14))
			TakeIt = false;
		if(TakeIt)
		{
			// the nvs1 file is not in the file list
			// so we add it if required
			string sFileStartDate = sStartFromNvs1File.substr(0, 8);
			string sFileStartTime = sStartFromNvs1File.substr(8);
			int iFileStartDate = atoi(sFileStartDate.c_str());
			int iFileStartTime = atoi(sFileStartTime.c_str());
			string sFileEndDate = sEndFromNvs1File.substr(0, 8);
			string sFileEndTime = sEndFromNvs1File.substr(8);
			int iFileEndDate = atoi(sFileEndDate.c_str());
			int iFileEndTime = atoi(sFileEndTime.c_str());
			if(iFileStartDate > iMaxYearmonthday)
				TakeIt = false;
			if(TakeIt)
				if(iFileEndDate < iMinYearmonthday)
					TakeIt = false;
			if(TakeIt)
				if((iFileStartDate == iMaxYearmonthday) && (iFileStartTime > iMaxHour))
					TakeIt = false;
			if(TakeIt)
				if((iFileEndDate == iMinYearmonthday) && (iFileEndTime < iMinHour))
					TakeIt = false;
		}
	}
	if(TakeIt)
		List.push_back(TempLogFile1);


	TakeIt = true;
	// Check now the nvs log file (in fact a copy of it)
	string sStartFromNvsFile = "";
	sStartFromNvsFile = GetLogHeader(TempLogFile.c_str(), "STARTDATETIME");

	if(sStartFromNvsFile.length() != 14)
		TakeIt = false;
	if(TakeIt)
	{
		// the nvs file may be added to the file list
		// let's check that
		string sFileStartDate = sStartFromNvsFile.substr(0, 8);
		string sFileStartTime = sStartFromNvsFile.substr(8);
		int iFileStartDate = atoi(sFileStartDate.c_str());
		int iFileStartTime = atoi(sFileStartTime.c_str());
		if(iFileStartDate > iMaxYearmonthday)
			TakeIt = false;
		if(TakeIt)
			if((iFileStartDate == iMaxYearmonthday) && (iFileStartTime > iMaxHour))
				TakeIt = false;
	}
	if(TakeIt)
		List.push_back(TempLogFile);
	
	// We must remove the temporary files after finishing the job
	// This is done under the control of the lock
	// In order to not have problems with temporary file names



	// At this step, we have a complete list of log files in the List
	// We can now create the concatenation of these files
	// We must distinguish the first, middle and last files
	// The first one must be added starting from the requested start date and time
	// The middle files must be added in any case
	// The last one must be added until the requested end date and time


	// we can prepare the resulting log file
	char *LineBuffer = (char *)malloc(NV_LOG_MAX_LINE_LENGTH + 10);
	if(LineBuffer == NULL)
	{
		RemoveFile(TempLogFile.c_str());
		RemoveFile(TempLogFile1.c_str());
		return false;
	}

	bool bSameFirstAndLast = false;
	if(List.size() == 1)
		bSameFirstAndLast = true;
	for(i=0; i<(int)(List.size()); i++)
	{
		string FileName = List[i];
		int Offset = 0;
		int NumBytes = -1;
		// We must lock as shorter as possible
		NV_LOCKOBJECT(this);
		if(!FileExist(FileName.c_str()))
		{
			NV_UNLOCKOBJECT(this);
			continue;
		}
		Offset = GetLogStartOffset(FileName.c_str());
		if(i == 0)
		{
			FILE *stream = Nvfopen(FileName.c_str(), "r+b");
			if(stream == NULL)
			{
				RemoveFile(TempLogFile.c_str());
				RemoveFile(TempLogFile1.c_str());
				NV_UNLOCKOBJECT(this);
				free(LineBuffer);
				return false;
			}

			fseek(stream, Offset, SEEK_SET);

			while(fgets(LineBuffer, NV_LOG_MAX_LINE_LENGTH, stream) > 0)
			{
				// We must check the date and hour
				int LineLength = (int)(strlen(LineBuffer));
				char *Buf = strstr(LineBuffer, NV_LOG_FIELD_SEPARATOR);
				if(Buf == NULL)
				{
					Offset += LineLength;
					continue;
				}
				*Buf = '\0';
				string DateFromLine = LineBuffer;
				int iDateFromLine = atoi(DateFromLine.c_str());
				if(iDateFromLine < iMinYearmonthday)
				{
					Offset += LineLength;
					continue;
				}
				if(iDateFromLine > iMinYearmonthday)
					break;	// don't need to check time in this case
				
				// Now we must check time because the
				// date from line is equal to the requested date
				Buf += strlen(NV_LOG_FIELD_SEPARATOR);
				char *Buf2 = strstr(Buf, NV_LOG_FIELD_SEPARATOR);
				if(Buf2 == NULL)
				{
					Offset += LineLength;
					continue;
				}
				*Buf2 = '\0';
				string HourFromLine = Buf;
				int iHourFromLine = atoi(HourFromLine.c_str());
				if(iHourFromLine >= iMinHour)
					break;

				Offset += LineLength;
			}

			// We have now the offset for the first file

			if(bSameFirstAndLast)
			{
				// If we have the same first and last file
				// We must adjust the number of bytes to the required lines
				NumBytes = 0;
				fseek(stream, Offset, SEEK_SET);
				while(fgets(LineBuffer, NV_LOG_MAX_LINE_LENGTH, stream) > 0)
				{
					// We must check the date and hour
					int LineLength = (int)(strlen(LineBuffer));
					char *Buf = strstr(LineBuffer, NV_LOG_FIELD_SEPARATOR);
					if(Buf == NULL)
					{
						NumBytes += LineLength;
						continue;
					}
					*Buf = '\0';
					string DateFromLine = LineBuffer;
					int iDateFromLine = atoi(DateFromLine.c_str());
					if(iDateFromLine < iMaxYearmonthday)
					{
						NumBytes += LineLength;
						continue;
					}
					if(iDateFromLine > iMaxYearmonthday)
						break;	// don't need to check time in this case
					
					// Now we must check time because the
					// date from line is equal to the requested date
					Buf += strlen(NV_LOG_FIELD_SEPARATOR);
					char *Buf2 = strstr(Buf, NV_LOG_FIELD_SEPARATOR);
					if(Buf2 == NULL)
					{
						NumBytes += LineLength;
						continue;
					}
					*Buf2 = '\0';
					string HourFromLine = Buf;
					int iHourFromLine = atoi(HourFromLine.c_str());
					if(iHourFromLine > iMaxHour)
						break;

					NumBytes += LineLength;
				}
			}
			fclose(stream);
			
		}
		else if(i == (List.size() - 1))
		{
			if(!bSameFirstAndLast)
			{
				// We do the job only if it has not been ever done
				FILE *stream = Nvfopen(FileName.c_str(), "r+b");
				if(stream == NULL)
				{
					RemoveFile(TempLogFile.c_str());
					RemoveFile(TempLogFile1.c_str());
					NV_UNLOCKOBJECT(this);
					free(LineBuffer);
					return false;
				}

				fseek(stream, Offset, SEEK_SET);


				// We must adjust the number of bytes to the required lines
				NumBytes = 0;
				while(fgets(LineBuffer, NV_LOG_MAX_LINE_LENGTH, stream) > 0)
				{
					// We must check the date and hour
					int LineLength = (int)(strlen(LineBuffer));
					char *Buf = strstr(LineBuffer, NV_LOG_FIELD_SEPARATOR);
					if(Buf == NULL)
					{
						NumBytes += LineLength;
						continue;
					}
					*Buf = '\0';
					string DateFromLine = LineBuffer;
					int iDateFromLine = atoi(DateFromLine.c_str());
					if(iDateFromLine < iMaxYearmonthday)
					{
						NumBytes += LineLength;
						continue;
					}
					if(iDateFromLine > iMaxYearmonthday)
						break;	// don't need to check time in this case
					
					// Now we must check time because the
					// date from line is equal to the requested date
					Buf += strlen(NV_LOG_FIELD_SEPARATOR);
					char *Buf2 = strstr(Buf, NV_LOG_FIELD_SEPARATOR);
					if(Buf2 == NULL)
					{
						NumBytes += LineLength;
						continue;
					}
					*Buf2 = '\0';
					string HourFromLine = Buf;
					int iHourFromLine = atoi(HourFromLine.c_str());
					if(iHourFromLine > iMaxHour)
						break;

					NumBytes += LineLength;
				}
				fclose(stream);
			}
		}

		if(NumBytes == 0)
		{
			NV_UNLOCKOBJECT(this);
			continue;
		}

		// We must eventually control the out file max length
		bool bToMuch = false;
		if(MaxFileSize != -1)
		{
			int OutFileLength = 0;
			FILE *File = Nvfopen(FileOut, "rb");
			if(File != NULL)
			{
				fseek(File, 0, SEEK_END);
				OutFileLength = ftell(File);
				fclose(File);
			}
			int PieceFileLength = 0;
			File = Nvfopen(FileName.c_str(), "rb");
			if(File != NULL)
			{
				fseek(File, 0, SEEK_END);
				PieceFileLength = ftell(File);
				fclose(File);
			}
			int FuturSize = OutFileLength;
			if(NumBytes == -1)
				FuturSize += PieceFileLength - Offset;
			else
				FuturSize += NumBytes;

			int MaxFileSizeInBytes = MaxFileSize * 1024;

			if(FuturSize > MaxFileSizeInBytes)
			{
				// We must reduce num bytes to write
				bToMuch = true;
				NumBytes = MaxFileSizeInBytes - OutFileLength;

				// Now, we must be sure to get a complete log line
				// so we adjust to the next line
				File = Nvfopen(FileName.c_str(), "rb");
				if(File == NULL)
				{
					RemoveFile(TempLogFile.c_str());
					RemoveFile(TempLogFile1.c_str());
					NV_UNLOCKOBJECT(this);
					free(LineBuffer);
					return false;
				}
				fseek(File, NumBytes+Offset, SEEK_SET);
				if(fgets(LineBuffer, NV_LOG_MAX_LINE_LENGTH, File) > 0)
					NumBytes += (int)(strlen(LineBuffer));
				fclose(File);
			}

		}
		if(!AppendFileEx(FileName.c_str(), FileOut, (NvInt64)Offset, (NvInt64)NumBytes))
		{
			RemoveFile(TempLogFile.c_str());
			RemoveFile(TempLogFile1.c_str());
			NV_UNLOCKOBJECT(this);
			free(LineBuffer);
			return false;
		}
		NV_UNLOCKOBJECT(this);

		// Truncation due to the MaxFileSize parameter
		if(bToMuch)
			break;
	}


	RemoveFile(TempLogFile.c_str());
	RemoveFile(TempLogFile1.c_str());

	free(LineBuffer);

	return true;
}



// This function removes some log information from the archive following a date and
// hour range. The removing occurs only on the archived for date ranges and for the actual data
// if Clear flag is beeing used
bool NvLog::RemoveLogData(const char* FromDate, const char* ToDate, bool Lock, bool ClearIt)
{
	vector<string> List;
	char Buffer[1024];
	List.clear();

	if(Lock)
		NV_LOCKOBJECT(this);
	if(ClearIt)
	{
		if(logstream != NULL)
			fclose(logstream);
		logstream = NULL;
		string LogFile = m_sLogDir;
		AddDirSuffix(LogFile);
		LogFile += "nvs.log";
		string LogFile1 = m_sLogDir;
		AddDirSuffix(LogFile1);
		LogFile1 += "nvs1.log";
		RemoveFile(LogFile.c_str());
		RemoveFile(LogFile1.c_str());
	}

	string ToDateHere = "";
	if(ToDate != NULL)
		ToDateHere = ToDate;
	CleanString(ToDateHere);

	string FromDateHere = "";
	if(FromDate != NULL)
		FromDateHere = FromDate;
	CleanString(FromDateHere);

	if((!ToDateHere.length()) && (!FromDateHere.length()))
	{
		// nothing to do except checking empty directories
		CheckEmptyDirs();
		if(Lock)
			NV_UNLOCKOBJECT(this);
		return true;
	}
	
	if(ToDateHere.length())
	{
		if(!ArrangeDate(ToDateHere))
		{
			if(Lock)
				NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	if(FromDateHere.length())
	{
		if(!ArrangeDate(FromDateHere))
		{
			if(Lock)
				NV_UNLOCKOBJECT(this);
			return false;
		}
	}
	else
	{
		// If the From date is not provided, we necessarly have a to date so we
		// will remove all data until the given To date
		FromDateHere = "01.01.2000";
	}



	int iFromDay = 0;
	int iFromMonth = 0;
	int iFromYear = 0;
	// By default, we get the actual day logs
	time_t tTime = time(NULL);
	struct tm tCurrentTime;
	localtime_r(&tTime, &tCurrentTime);
	iFromDay = tCurrentTime.tm_mday;
	iFromMonth = tCurrentTime.tm_mon + 1;
	iFromYear = tCurrentTime.tm_year + 1900;


	string FromDateHereHere = FromDateHere;
	if(FromDateHereHere.length())
	{
		string Day = "";
		string Month = "";
		string Year = "";
		string::size_type index;
		index = FromDateHereHere.find('.');
		if(index != string::npos)
		{
			Day = FromDateHereHere.substr(0, index);
			FromDateHereHere = FromDateHereHere.substr(index+1);
		}
		index = FromDateHereHere.find('.');
		if(index != string::npos)
		{
			Month = FromDateHereHere.substr(0, index);
			Year = FromDateHereHere.substr(index+1);
		}


		iFromDay = atoi(Day.c_str());
		iFromMonth = atoi(Month.c_str());
		iFromYear = atoi(Year.c_str());
	}


	int iToDay = iFromDay;
	int iToMonth = iFromMonth;
	int iToYear = iFromYear;
	string ToDateHereHere = ToDateHere;
	if(ToDateHereHere.length())
	{
		string Day = "";
		string Month = "";
		string Year = "";
		string::size_type index;
		index = ToDateHereHere.find('.');
		if(index != string::npos)
		{
			Day = ToDateHereHere.substr(0, index);
			ToDateHereHere = ToDateHereHere.substr(index+1);
		}
		index = ToDateHereHere.find('.');
		if(index != string::npos)
		{
			Month = ToDateHereHere.substr(0, index);
			Year = ToDateHereHere.substr(index+1);
		}


		iToDay = atoi(Day.c_str());
		iToMonth = atoi(Month.c_str());
		iToYear = atoi(Year.c_str());
	}

	string SaveLogDir = m_sLogDir;
	AddDirSuffix(SaveLogDir);
	SetDirName(SaveLogDir);
	SaveLogDir += "Saves";
	AddDirSuffix(SaveLogDir);

	sprintf(Buffer, "%04d%02d", iFromYear, iFromMonth);
	int iMinYearmonth = atoi(Buffer);
	sprintf(Buffer, "%04d%02d", iToYear, iToMonth);
	int iMaxYearmonth = atoi(Buffer);
	sprintf(Buffer, "%04d%02d%02d", iFromYear, iFromMonth, iFromDay);
	int iMinYearmonthday = atoi(Buffer);
	sprintf(Buffer, "%04d%02d%02d", iToYear, iToMonth, iToDay);
	int iMaxYearmonthday = atoi(Buffer);
	if(iMinYearmonthday > iMaxYearmonthday)
	{
		CheckEmptyDirs();
		if(Lock)
			NV_UNLOCKOBJECT(this);
		return true;	// nothing to do
	}

	vector<string> directories;
	vector<string> years;
	
	// We must lock as shorter as possible
	// but we do it for building the file list
	DirList(SaveLogDir.c_str(), NULL, &years);
	int i;
	for(i=0; i<(int)(years.size()); i++)
	{
		// Eliminate bad directories
		string YearHere = years[i];
		if(YearHere.length() != 4)
			continue;
		int iYearHere = atoi(YearHere.c_str());
		if((iYearHere < iFromYear) || (iYearHere > iToYear))
			continue;
		if((iYearHere < 1900) || (iYearHere > 2100))
			continue;

		// Now check for months
		vector<string> months;
		string SaveLogDirYear = SaveLogDir;
		SaveLogDirYear += YearHere;
		AddDirSuffix(SaveLogDirYear);
		// We must lock as shorter as possible
		DirList(SaveLogDirYear.c_str(), NULL, &months);
		int j;
		for(j=0; j<(int)(months.size()); j++)
		{
			// Eliminate bad directories
			string MonthHere = months[j];
			if(MonthHere.length() != 2)
				continue;

			string YearMonthHere = YearHere;
			YearMonthHere += MonthHere;
			int iYearMonthHere = atoi(YearMonthHere.c_str());
			if((iYearMonthHere < iMinYearmonth) || (iYearMonthHere > iMaxYearmonth))
				continue;

			// Now check for days
			vector<string> days;
			string SaveLogDirYearMonth = SaveLogDirYear;
			SaveLogDirYearMonth += MonthHere;
			AddDirSuffix(SaveLogDirYearMonth);
			// We must lock as shorter as possible
			DirList(SaveLogDirYearMonth.c_str(), NULL, &days);
			int k;
			for(k=0; k<(int)(days.size()); k++)
			{
				// Eliminate bad directories
				string DayHere = days[k];
				if(DayHere.length() != 2)
					continue;

				string YearMonthDayHere = YearMonthHere;
				YearMonthDayHere += DayHere;
				int iYearMonthDayHere = atoi(YearMonthDayHere.c_str());
				if((iYearMonthDayHere < iMinYearmonthday) || (iYearMonthDayHere > iMaxYearmonthday))
					continue;

				// The directory is now candidate so we can add it to the list
				string DirToAdd = YearHere;
				AddDirSuffix(DirToAdd);
				DirToAdd += MonthHere;
				AddDirSuffix(DirToAdd);
				DirToAdd += DayHere;
				AddDirSuffix(DirToAdd);
				directories.push_back(DirToAdd);

			}
		}
	}

	// We don't need that any more
	years.clear();

	// Now in directories, we have all the candidate directories (starting for the Log Saves directory)
	for(i=0; i<(int)(directories.size()); i++)
	{
		string FullDirName = m_sLogDir;
		AddDirSuffix(FullDirName);
		SetDirName(FullDirName);
		FullDirName += "Saves";
		AddDirSuffix(FullDirName);
		FullDirName += directories[i];
		if(!RemoveDir(FullDirName.c_str()))
		{
			if(Lock)
				NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	CheckEmptyDirs();
	if(Lock)
		NV_UNLOCKOBJECT(this);

	return true;
}



// Must be locked when entering this function
// This function removes empty directories in the log archive
void NvLog::CheckEmptyDirs()
{
	string SaveLogDir = m_sLogDir;
	AddDirSuffix(SaveLogDir);
	SetDirName(SaveLogDir);
	SaveLogDir += "Saves";

	vector<string> years;
	DirList(SaveLogDir.c_str(), NULL, &years);
	int i, j, k;
	for(i=0; i<(int)(years.size()); i++)
	{
		string YearDir = SaveLogDir;
		AddDirSuffix(YearDir);
		YearDir += years[i];
		vector<string> months;
		DirList(YearDir.c_str(), NULL, &months);
		for(j=0; j<(int)(months.size()); j++)
		{
			string MonthDir = YearDir;
			AddDirSuffix(MonthDir);
			MonthDir += months[j];
			vector<string> days;
			DirList(MonthDir.c_str(), NULL, &days);
			for(k=0; k<(int)(days.size()); k++)
			{
				string DayDir = MonthDir;
				AddDirSuffix(DayDir);
				DayDir += days[k];
				vector<string> files;
				FileList(DayDir.c_str(), NULL, &files);
				if(!files.size())
					RemoveDir(DayDir.c_str());
			}
			days.clear();
			DirList(MonthDir.c_str(), NULL, &days);
			if(!days.size())
				RemoveDir(MonthDir.c_str());
		}
		months.clear();
		DirList(YearDir.c_str(), NULL, &months);
		if(!months.size())
			RemoveDir(YearDir.c_str());
	}

	RemoveFiles(SaveLogDir.c_str(), "nvslog");
}


