// ***************************************************************************************
// File: protocol.h
// Description: Communication protocol
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: This file defines the protocol classes.
//           There is one abstract class named NvProtocol that defines the
//           interface of the protocol to external system.
//           Then, the classes derived from this base class implement a
//           specific protocol. They only have to overload the Read and Write
//           functions of the base class.
//           The NvData class is used only by the NvProtocol class. It
//           encapsulate binary data information associated with a message. The
//           NvProtocol interface defines a collection of NvData objects.
// ***************************************************************************************


#ifndef _PROTOCOL_H
#define _PROTOCOL_H


// Classes declared in this file
class NvData;
class NvProtocol;
class NvHttpProtocol;

#include <vector>
#include <string>
#include <map>

#include "socket.h"

// HTTP Methods implemented
#define HTTP_METHOD_NONE	0
#define HTTP_METHOD_GET		1
#define HTTP_METHOD_PUT		2
#define HTTP_METHOD_POST	3

#define HTTP_VERSION_STRING	"HTTP/1.1"

#ifdef WIN32
#pragma warning(disable:4786)
#endif




// Class containing data
class NvData
{
// Attributes
public:
	std::string sInfo;									// contains global information on the data
	std::map<std::string, std::string> mHeaders;		// headers string1(index)=headname string2=headvalue
	unsigned long DataSize;								// associated data size
	char *Data;											// associated data
	std::string DataFile;								// associated data file
	bool bDeleteDataFile;								// flag for automatic removing of data file

// Operations
public:
	const NvData& operator = (const NvData& DataSrc);

// Overridable callbacks
protected:

// Implementation
public:
	NvData();
	virtual ~NvData();
	NvData(const NvData& DataSrc);

	void Clear();			// reset the class
	void FreeData();		// free associated data
};


// this is an abstract class for protocol implementation
// it completly defines the protocol interface
class NvProtocol
{
// Attributes
public:
	NvSocket *sSocket;
	
	std::string sPath;									// Path to service
	std::string sService;								// Service name
	std::string sCommand;								// Command
	std::map<std::string, std::string> mParameters;		// Command Parameters string1(index)=paramname string2=paramvalue
	std::vector<NvData*>	vDatas;						// A list of Datas containing information, headers and binary data
	bool bMultipart;									// Set to true if the content is multipart

protected:
	
	// 2 next functions are pure virtual functions that must be implemented in derived classes
	
	// Fills the data members from socket. Returns -1 if error, 0 if time out and 1 if successfull
	virtual int Read(unsigned short TimeOut = 0, bool UseDataFile = false, int DataFileMinSize= 1048576, const char *WorkDir = NULL) = 0;
	// Writes data members to socket. Returns 0 if error and 1 if successfull
	virtual int Write() = 0;
	virtual int WriteTest() = 0;

	// Implementation
public:
	NvProtocol();						// standard constructor
	NvProtocol(NvSocket *Socket);		// constructor with a socket pointer
	virtual ~NvProtocol();

	// resets the class memebers. this is not a pure virtual function
	// but derived classes should generally overwrite it
	virtual void Clear();
	
	// Add a header value to the given data index. If a header with the
	// same name exists, its value is replaced by the new one
	void AddHeader(std::string HeaderName, std::string HeaderValue, int index = 0, bool Encode = false);

	// Test if a parameter exist. Returns true if it exist and
	// false otherwise
	bool ParameterExist(std::string ParameterName);

	// Gets a parameter value. Returns the string value or an empty
	// string if the parameter doesn't exist.
	std::string GetParameter(std::string ParameterName);

	// Test if a header exist. Returns true if it exist and
	// false otherwise.
	bool HeaderExist(std::string HeaderName, int index = 0);

	// Gets a header value for the given data index. Returns the string value or an empty
	// string if the parameter doesn't exist or the index is invalid.
	std::string GetHeader(std::string HeaderName, int index = 0);

	// Sets the data bytes from a file for the given data index. Returns true if successful and false if error
	bool SetDataFromFile(const char *FileName, int index = 0);

	// Encode header
	void EncodeHeader(std::string& Value);

	// Decode header
	void DecodeHeader(std::string& Value);
};







// this class implements the http protocol
class NvHttpProtocol : public NvProtocol
{
// Attributes
public:
	int iMethod;				// Method only GET, PUT and POST for now
	std::string sHttpVersion;	// HTTP version
	int NumCr;					// Number of CRs before a /n

	char *ReadBuf;
	int ReadBufSize;
	int ReadBufStart;
	int ReadBufDataSize;

	std::string sSessionId;								// Used in nirva server to update session time-out
	NvObjIndStringList Aliases;				// HTTP Aliases

// Implementation
public:
	NvHttpProtocol();								// standard constructor
	NvHttpProtocol(NvSocket *Socket);				// constructor with a socket pointer
	virtual ~NvHttpProtocol();
	void Clear();									// resets the class members
	virtual int Read(unsigned short TimeOut = 0, bool UseDataFile = false, int DataFileMinSize = 1048576, const char *WorkDir = NULL);
	virtual int Write();
	virtual int WriteTest();
	
	// Read a line terminating with /n from the socket
	// The /n character is not included
	// If the last character is /r, it is removed
	// The max parameter tells the maximum number of characters
	// to read (-1 for the all line)
	// The function returns the number or read characters or -1 if error
	int ReadLine(std::string& line, int max = -1, bool CountCr = false);

	// Reads data from socket using read buffer
	int ReadData(char *Data, int iDataLength, bool withtimeout = false);

	// Returns the real size of the http body as it will be
	// sent by the Write function
	NvInt64 GetWriteBodySize();

	// Resolves aliases
	void GetAlias(std::string& Path);

private:
	void SetParameters(std::string line);				// extract parameters from the line

	// Reads the body part and write it in data memory or file if there is an attached file to the NvData
	int ReadBody(NvData *Data, NvInt64 bodysize);

	// Used in server for updating session time outs in case of a big message
	void UpdateSessionTimeOut();
};





#endif //_PROTOCOL_H

