// ***************************************************************************************
// File: request.cpp
// Description: communication from nirva to nirva
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************


#include "nvs.h"


#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;



/////////////////////////////////////////////////////////////////////////////
// NvRequestList

NvRequestList::NvRequestList()
{
}

NvRequestList::~NvRequestList()
{
	Clear();
}



void NvRequestList::Clear()
{
	Lock();
	map<string,NvRequest*>::iterator pos;
	for(pos = m_mRequests.begin(); pos != m_mRequests.end(); ++pos)
	{
		NvRequest *Request = (NvRequest *)(pos->second);
		if(Request != NULL)
			delete Request;
	}
	m_mRequests.clear();
	Unlock();
}



NvRequest* NvRequestList::AddRequest(const char* Name, bool Reopen)
{
	string NameHere = "";
	if(Name != NULL)
		NameHere = Name;
	CleanString(NameHere);
	MakeUpper(NameHere);
	Lock();
	if(m_mRequests.find(NameHere) != m_mRequests.end())
	{
		if(Reopen)
		{
			NvRequest *RequestInList = m_mRequests[NameHere];
			if(RequestInList != NULL)
				delete RequestInList;
			m_mRequests.erase(NameHere);
		}
		else
		{
			Unlock();
			return NULL;
		}
	}

	NvRequest *Request = NVNEW NvRequest;
	if(Request != NULL)
	{
		if(!NameHere.length())
		{
			char Buf[50];
			sprintf(Buf, "%lu", (unsigned long)Request);
			NameHere = Buf;
			MakeUpper(NameHere);
		}
		m_mRequests[NameHere] = Request;
	}
	Unlock();

	// Update connection list each time a new request is added
	GetMain()->Connections.Update();
	return Request;
}


void NvRequestList::RemoveRequest(const char* Name)
{
	string NameHere = "";
	if(Name != NULL)
		NameHere = Name;
	CleanString(NameHere);
	MakeUpper(NameHere);

	Lock();
	if(m_mRequests.find(NameHere) != m_mRequests.end())
	{
		NvRequest *RequestInList = m_mRequests[NameHere];
		if(RequestInList != NULL)
			delete RequestInList;
		m_mRequests.erase(NameHere);
	}
	Unlock();
}


NvRequest* NvRequestList::GetRequest(const char* Name)
{
	string NameHere = "";
	if(Name != NULL)
		NameHere = Name;
	CleanString(NameHere);
	MakeUpper(NameHere);

	Lock();
	NvRequest* RequestFound = NULL;
	if(m_mRequests.find(NameHere) != m_mRequests.end())
		RequestFound = m_mRequests[NameHere];
	Unlock();
	return RequestFound;
}



/////////////////////////////////////////////////////////////////////////////
// NvRequest

NvRequest::NvRequest()
{
	m_sServer = "127.0.0.1:1081";			// Default nvs server address and port
	m_sServerAddress = "127.0.0.1:1081";	// Used for log to avoid giving proxy password in log or console
	m_sUser = "";
	m_sHash = "";
	m_sPassword = "";
	m_sNewPassword = "";
	m_sNewPasswordConfirm = "";
	m_sAppli = "";
	m_sOpen = "";
	m_sClose = "";
	m_sSessionId = "";
	m_iAutoClose = NV_REQUEST_CLOSE_SESSION;
	m_iSessionTimeOut = 0;
	m_iServerConnectionTimeOut = 10;
	CurrentConnection = NULL;
	b_UseSSL = false;
	b_ClientUnicode = false;
	b_Hidden = false;
	bWithNewPassword = false;
	bWithNewPasswordConfirm = false;
	bControlPassword = true;
	Session = NULL;
	m_sSocketLogFile = "";
	bSensible = false;
	m_sServerVersion = "";
}

NvRequest::~NvRequest()
{
	Clear();
}

void NvRequest::Clear()
{
	if(m_sSessionId.length() && (m_iAutoClose != NV_REQUEST_CLOSE_MANUAL))
	{
		// We free the distant session if requested on exit
		NvCommand Command;
		Command.m_iCommandSource = NV_SOURCE_SERVER;
		Command.m_sService = "SYSTEM";
		Command.m_sClass = "SESSION";
		Command.m_sCommand = "CLOSE";
		Command.m_sInContainer = NV_CONTAINER_DEFAULT_NAME;
		Command.m_sOutContainer = NV_CONTAINER_DEFAULT_NAME;
		Command.Parameters.SetValue("NV_SESSION_ID", m_sSessionId);
		RequestCommand(&Command);
	}
	Session = NULL;
	if(CurrentConnection != NULL)
		GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
	CurrentConnection = NULL;
}



bool NvRequest::RequestCommand(NvCommand *Command, char *Buffer, int BufferSize, int *DataSize)
{
	if(DataSize != NULL)
		*DataSize = 0;
	if((Buffer == NULL) && (BufferSize > 0))
		Buffer[0] = 0;
	
	string Address = m_sServer;
	CleanString(Address);

	string Proxy = "";
	string ProxyUser = "";
	string ProxyPassword = "";

	// The proxy is eventually in the address
	// at this time the address starts with proxy::proxyaddress::...
	// proxyaddress can have some optional user and password information in brackets separated by a ; character
	string AddressHere = Address;
	MakeUpper(AddressHere);
	string::size_type index = AddressHere.find("PROXY::");
	if(index != string::npos)
	{
		Address = Address.substr(7);
		index = Address.find("::");
		if(index != string::npos)
		{
			Proxy = Address.substr(0, index);
			Address = Address.substr(index+2);
		}
		index = Proxy.find('(');
		if(index != string::npos)
		{
			ProxyUser = Proxy.substr(index+1);
			Proxy = Proxy.substr(0, index);
		}
		index = ProxyUser.find(')');
		if(index != string::npos)
			ProxyUser = ProxyUser.substr(0, index);
		
		index = ProxyUser.find(';');
		if(index != string::npos)
		{
			ProxyPassword = ProxyUser.substr(index+1);
			ProxyUser = ProxyUser.substr(0, index);
		}
	}
	CleanString(Proxy);
	CleanString(ProxyUser);
	CleanString(ProxyPassword);
	CleanString(Address);

	int Port = 1081;
	bool UseSsl = b_UseSSL;
	if(UseSsl)
		Port = 1082;
	index = Address.find(':');
	if(index != string::npos)
	{
		string Provistr = Address.substr(index+1);
		CleanString(Provistr);
		Port = atoi(Provistr.c_str());
		Address.erase(index);
		CleanString(Provistr);
	}

	if(!Address.length())
		Address = "127.0.0.1";

	string ConnectedServer = m_sServer; 
	string FinalAddress = Address;
	int FinalPort = Port;
	bool FinalUseSsl = UseSsl;
	bool bUseProxy = false;
	if(Proxy.length())
	{
		string ProxyHere = Proxy;
		MakeUpper(ProxyHere);
		CleanString(ProxyHere);
		index = ProxyHere.find("HTTP://");
		if(index != string::npos)
		{
			Proxy = Proxy.substr(index+7);
			bUseProxy = true;
			UseSsl = false;
		}
		if(!bUseProxy)
		{
			index = ProxyHere.find("HTTPS://");
			if(index != string::npos)
			{
				Proxy = Proxy.substr(index+8);
				bUseProxy = true;
				UseSsl = true;
			}
		}
		if(bUseProxy)
		{
			Address = Proxy;
			Port = 4480;
			index = Address.find(':');
			if(index != string::npos)
			{
				string Provistr = Address.substr(index+1);
				CleanString(Provistr);
				Port = atoi(Provistr.c_str());
				Address.erase(index);
			}
		}
	}

	int ProxyMode = NO_PROXY;
	if(bUseProxy)
	{
		if(FinalUseSsl)
			ProxyMode = PROXY_FINAL_SSL;
		else
			ProxyMode = PROXY_FINAL_NORMAL;
	}

	NvHttpConnection* Connection = (NvHttpConnection*)CurrentConnection;
	if(Connection == NULL)
	{
		Connection = GetMain()->Connections.GetConnection(Address.c_str(), Port, m_iServerConnectionTimeOut, UseSsl, ProxyMode, FinalAddress.c_str(), FinalPort);
		CurrentConnection = Connection;
	}
	if(Connection == NULL)
	{
		Command->Error.SetError("SYSTEM", "REQUEST", NV_REQUEST_CANNOT_GET_SERVER_CONNECTION, ConnectedServer);
		return false;
	}

	// We get a connected connection
	Connection->ProxyAuthInfo.m_sUser = ProxyUser;
	Connection->ProxyAuthInfo.m_sPassword = ProxyPassword;
	
	// PM Modif for proxy HTTPS
	if(bUseProxy && FinalUseSsl)
		Connection->b_ProxyHttps = true;
	else
		Connection->b_ProxyHttps = false;

	// Set an eventual log file
	Connection->sSocket.LogFile = m_sSocketLogFile;

	bool RetVal = false;
	
	// we clear the input message
	Connection->InMessage.Clear();
	
	// we clear the outpput message
	Connection->OutMessage.Clear();

	NvData *Data = NVNEW NvData;

	// For HTTP 1.1 we must include the host header
	char buffer[20];
	sprintf(buffer, "%d", FinalPort);
	FinalAddress += ":";
	FinalAddress += buffer;

	Data->sInfo = "POST ";
	// PM Modif for proxy HTTPS
	//if(bUseProxy)
	if(bUseProxy && (!FinalUseSsl))
	{
		Data->sInfo += "http://";
		Data->sInfo += FinalAddress;
	}
	Data->sInfo += "/NVS?Nirva";
	Data->sInfo += " HTTP/1.1";

	Connection->OutMessage.vDatas.push_back(Data);
	
	// by default, we have no content in the body
	Connection->OutMessage.AddHeader("Content-Length", "0");
	
	// For HTTP 1.1 we must include the host header
	Connection->OutMessage.AddHeader("Host", FinalAddress);

	Connection->OutMessage.AddHeader("Nv_Source", HTTP_CLIENT_STRING);

	// We prepare the output global headers
	
	// Set content-type
	string sContentType = "multipart/form-data; boundary=";
	sContentType += HTTP_SEND_BOUNDARY;
	Connection->OutMessage.AddHeader("Content-Type", sContentType);

	// Set Unicode
	if(b_ClientUnicode)
		Connection->OutMessage.AddHeader("Nv_Client_Unicode", "YES");

	if(!bSensible)
	{
		// Set NvMachineName
		Connection->OutMessage.AddHeader("Nv_Machine_Name", GetMain()->GetMachineName());

		// Set NvMachineUser
		Connection->OutMessage.AddHeader("Nv_Machine_User", GetMain()->GetMachineUser());
	}

	// Set NvSessionId
	Connection->OutMessage.AddHeader("Nv_Session_Id", m_sSessionId);

	// Send	control hash
	Connection->OutMessage.AddHeader("Nv_Control", m_sHash);

	if(!m_sSessionId.length())
	{
		// Send connection parameters
	
		// Set NvUser
		Connection->OutMessage.AddHeader("Nv_User", m_sUser);

		// Set NvPassword
		string PasswordHere = m_sPassword;
		if(PasswordHere.length() && (PasswordHere.length() < 256))
		{
			// The string must be encrypted
			string EncryptionKey = GetMain()->GetMachineName();
			EncryptionKey += "nvc";
			EncryptionKey += m_sAppli;
			MakeLower(EncryptionKey);

			int BufferLength = 100 + (4*((int)(PasswordHere.length())));
			char *Buffer = (char *)malloc(BufferLength);
			strcpy(Buffer, PasswordHere.c_str());
			char *Buffer2 = (char *)malloc(4*EncryptionKey.length());
			strcpy(Buffer2, EncryptionKey.c_str());
			EncryptString(Buffer2, Buffer, BufferLength);
			PasswordHere = Buffer;
			free(Buffer);
			free(Buffer2);
		}
		Connection->OutMessage.AddHeader("Nv_Password", PasswordHere);
		string m_sHash2 = m_sHash;
		m_sHash2 += "CTRL:";
		if(!bControlPassword)
			m_sHash2 += "NV_NO_CTRL";
		m_sHash2 = MD5_string(m_sHash2);
		MakeUpper(m_sHash2);

		Connection->OutMessage.AddHeader("Nv_Control2", m_sHash2);

		if(bWithNewPassword)
		{
			// Set NvNewPassword
			string NewPasswordHere = m_sNewPassword;
			if(NewPasswordHere.length() && (NewPasswordHere.length() < 256))
			{
				// The string must be encrypted
				string EncryptionKey = GetMain()->GetMachineName();
				EncryptionKey += "nvc";
				EncryptionKey += m_sAppli;
				MakeLower(EncryptionKey);

				int BufferLength = 100 + (4*((int)(NewPasswordHere.length())));
				char *Buffer = (char *)malloc(BufferLength);
				strcpy(Buffer, NewPasswordHere.c_str());
				char *Buffer2 = (char *)malloc(4*EncryptionKey.length());
				strcpy(Buffer2, EncryptionKey.c_str());
				EncryptString(Buffer2, Buffer, BufferLength);
				NewPasswordHere = Buffer;
				free(Buffer);
				free(Buffer2);
			}
			Connection->OutMessage.AddHeader("Nv_New_Password", NewPasswordHere);
			
			if(bWithNewPasswordConfirm)
			{
				// Set NvNewPasswordConfirm
				string NewPasswordConfirmHere = m_sNewPasswordConfirm;
				if(NewPasswordConfirmHere.length() && (NewPasswordConfirmHere.length() < 256))
				{
					// The string must be encrypted
					string EncryptionKey = GetMain()->GetMachineName();
					EncryptionKey += "nvc";
					EncryptionKey += m_sAppli;
					MakeLower(EncryptionKey);

					int BufferLength = 100 + (4*((int)(NewPasswordConfirmHere.length())));
					char *Buffer = (char *)malloc(BufferLength);
					strcpy(Buffer, NewPasswordConfirmHere.c_str());
					char *Buffer2 = (char *)malloc(4*EncryptionKey.length());
					strcpy(Buffer2, EncryptionKey.c_str());
					EncryptString(Buffer2, Buffer, BufferLength);
					NewPasswordConfirmHere = Buffer;
					free(Buffer);
					free(Buffer2);
				}
				Connection->OutMessage.AddHeader("Nv_New_Password_Confirm", NewPasswordConfirmHere);
			}
		}

		// Set NvApplication
		Connection->OutMessage.AddHeader("Nv_Application", m_sAppli);

		// Set NvTimeOut
		sprintf(buffer, "%d", m_iSessionTimeOut);
		Connection->OutMessage.AddHeader("Nv_TimeOut", buffer);

		// Set NvSessionOpen procedure
		if(m_sOpen.length())
			Connection->OutMessage.AddHeader("Nv_Session_Open", m_sOpen);

		// Set NvSessionClose procedure
		if(m_sClose.length())
			Connection->OutMessage.AddHeader("Nv_Session_Close", m_sClose);

		// Set hidden mode
		if(b_Hidden)
			Connection->OutMessage.AddHeader("NV_HIDDEN", "YES");
	}
	
	// Set the auto close session flag
	if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
		Connection->OutMessage.AddHeader("SESSION", "close");
	
		
	// Write now each command
	string sBoundary = "\r\n--";
	sBoundary += HTTP_SEND_BOUNDARY;

		
	Data = NVNEW NvData;
	Connection->OutMessage.vDatas.push_back(Data);
	Connection->OutMessage.vDatas[1]->sInfo = sBoundary;
	Connection->OutMessage.AddHeader("Nv_Service", Command->m_sService, 1);
	Connection->OutMessage.AddHeader("Nv_Class", Command->m_sClass, 1);
	Connection->OutMessage.AddHeader("Nv_Command", Command->m_sCommand, 1);
	Connection->OutMessage.AddHeader("Nv_In_Container", Command->m_sInContainer, 1);
	Connection->OutMessage.AddHeader("Nv_Out_Container", Command->m_sOutContainer, 1);
	Connection->OutMessage.AddHeader("Nv_Language", Command->m_sLanguage, 1);
	Connection->OutMessage.AddHeader("Content-Type", "nirva/command", 1);

	map<string,string>::iterator pos;
	for(pos = Command->Parameters.m_mList.begin(); pos != Command->Parameters.m_mList.end(); ++pos)
		Connection->OutMessage.AddHeader(pos->first, pos->second, 1, true);


	// We must set the data for the SendObject command of the system service

	string ServiceHere = Command->m_sService;
	MakeUpper(ServiceHere);			// The service is set to uppercase by the server
	CleanString(ServiceHere);

	string ClassHere = Command->m_sClass;
	MakeUpper(ClassHere);			// The class is set to uppercase by the server
	CleanString(ClassHere);

	string CommandHere = Command->m_sCommand;
	MakeUpper(CommandHere);
	CleanString(CommandHere);
	
	if((ServiceHere == "SYSTEM") && (ClassHere == "OBJECT") && (CommandHere == "SEND"))
	{

		// Get server object name that should be in parameters
		string ServerObjName = Command->Parameters.GetValue("NAME");
		CleanString(ServerObjName);
		if(!ServerObjName.length())
			ServerObjName = NV_OBJ_DEFAULT_NAME;

		// Get local object name that may be in parameters
		// If this parameter is not here or blank, the local object name is
		// the same than the server object name
		string LocalObjName = Command->Parameters.GetValue("LNAME");
		CleanString(LocalObjName);
		if(!LocalObjName.length())
			LocalObjName = ServerObjName;


		NvObj* Object = Session->OutContainer->GetObject(LocalObjName.c_str());
		if(Object == NULL)
		{
			Command->Error.SetError("SYSTEM", "OBJECT", NV_OBJECT_CANNOT_GET, LocalObjName);
			Connection->OutMessage.Clear();
			if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
			{
				if(CurrentConnection != NULL)
					GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
				CurrentConnection = NULL;
			}
			return false;
		}
		int iObjectType = Object->GetType();
		if(iObjectType == NV_OBJTYPE_FILE)
		{
			NvInt64 FileSize = ((NvObjFile *)Object)->GetSize();
			bool bFileMessage = false;
			if(FileSize > (NvInt64)(GetMain()->GetHttpMaxMemBody()))	// for files grater than 1024 kbytes
				bFileMessage = true;
			
			if(bFileMessage)
			{
				string ObjectFileName = ((NvObjFile *)Object)->GetFilePathName();
				Connection->OutMessage.vDatas[1]->DataFile = ObjectFileName;
				Connection->OutMessage.vDatas[1]->bDeleteDataFile = false;	// Don't remove the file because we directly use the object file
			}
			else
			{
				// Not a too big file so we use memory data
				if(!((NvObjFile *)Object)->GetData(&(Connection->OutMessage.vDatas[1]->Data), &(Connection->OutMessage.vDatas[1]->DataSize)))
				{
					Command->Error.SetError("SYSTEM", "OBJECT", NV_OBJECT_CANNOT_GET_DATA, LocalObjName);
					Connection->OutMessage.Clear();
					if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
					{
						if(CurrentConnection != NULL)
							GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
						CurrentConnection = NULL;
					}
					return false;
				}
			}
		}
		else
		{
			if(!(Session->OutContainer->GetObjectData(LocalObjName.c_str(), &(Connection->OutMessage.vDatas[1]->Data), &(Connection->OutMessage.vDatas[1]->DataSize))))
			{
				Command->Error.SetError("SYSTEM", "OBJECT", NV_OBJECT_CANNOT_GET_DATA, LocalObjName);
				Connection->OutMessage.Clear();
				if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
				{
					if(CurrentConnection != NULL)
						GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
					CurrentConnection = NULL;
				}
				return false;
			}
		}

		Connection->OutMessage.AddHeader("Content-Length", "0", 1);
		NvInt64 bodysize = Connection->OutMessage.vDatas[1]->DataSize;
		if(Connection->OutMessage.vDatas[1]->DataFile.length())
			bodysize = GetFileLength(Connection->OutMessage.vDatas[1]->DataFile.c_str());
		if(bodysize)
		{
			char ProviBuf[50];
			sprintf(ProviBuf, "%llu", bodysize);
			Connection->OutMessage.AddHeader("Content-Length", ProviBuf, 1);
		}
		
		// The object name and type should be send in parameters
		Connection->OutMessage.AddHeader("NAME", ServerObjName, 1);
		char Buffer[50];
		sprintf(Buffer, "%d", iObjectType);
		Connection->OutMessage.AddHeader("TYPE", Buffer, 1);
		

		// PM TODO
		// We can call an Encode function here that operates on Command->Data
		// This Encode function can also compress the data stream
	}

	// Write the end of the message
	Data = NVNEW NvData;
	Connection->OutMessage.vDatas.push_back(Data);
	sBoundary += "--";		// last boundary
	Connection->OutMessage.vDatas[2]->sInfo = sBoundary;

	// Set the real body size
	NvInt64 BodySize = Connection->OutMessage.GetWriteBodySize();
	sprintf(buffer, "%llu", BodySize);
	Connection->OutMessage.AddHeader("Content-Length", buffer);

	// Now send the order
	string TempWorkDir = GetMain()->GetNirvaDir();
	AddDirSuffix(TempWorkDir);
	TempWorkDir += "Work";
	RetVal = true;
	int OrderResult = Connection->SendOrder(true, GetMain()->GetHttpMaxMemBody(), TempWorkDir.c_str());

	// we clear the output message
	Connection->OutMessage.Clear();

	if(OrderResult != 1)
	{
		// An error occured
		// we clear the input message
		Connection->InMessage.Clear();

		if(OrderResult == -2)
			Command->Error.SetError("SYSTEM", "HTTP", NV_HTTP_PROXY_AUTHENTICATION_ERROR, Address);
		else if(OrderResult == -1)
			Command->Error.SetError("SYSTEM", "HTTP", NV_HTTP_AUTHENTICATION_ERROR, ConnectedServer);
		else
			Command->Error.SetError("SYSTEM", "REQUEST", NV_REQUEST_NETWORK_ERROR, ConnectedServer);
		if(CurrentConnection != NULL)
			GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
		CurrentConnection = NULL;
		return false;
	}
	
	// Get the Connection->OutMessage here


	// The result should have the same number of parts than the number of commands
	// + 1 for the global result + 1 for the first part 
	if(Connection->InMessage.vDatas.size() != 3)
	{
		// An error occured
		// we clear the input message
		Connection->InMessage.Clear();

		Command->Error.SetError("SYSTEM", "REQUEST", NV_REQUEST_NETWORK_ERROR, ConnectedServer);
		if(CurrentConnection != NULL)
			GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
		CurrentConnection = NULL;
	}

	if((Connection->InMessage.vDatas.size()) < 2)
	{
		// This is not a Nirva server
		// we clear the input message
		Connection->InMessage.Clear();

		Command->Error.SetError("SYSTEM", "REQUEST", NV_REQUEST_NETWORK_ERROR, ConnectedServer);
		if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
		{
			if(CurrentConnection != NULL)
				GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
			CurrentConnection = NULL;
		}
		return false;
	}

	if(!Connection->InMessage.HeaderExist("NV_SESSION_ID", 1))
	{
		// This is not a Nirva server
		// we clear the input message
		Connection->InMessage.Clear();

		Command->Error.SetError("SYSTEM", "REQUEST", NV_REQUEST_NETWORK_ERROR, ConnectedServer);
		if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
		{
			if(CurrentConnection != NULL)
				GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
			CurrentConnection = NULL;
		}
		return false;
	}

	m_sSessionId = Connection->InMessage.GetHeader("NV_SESSION_ID", 1);
	if(m_sSessionId.length() && (m_sSessionId[0] == '.'))
		m_sSessionId = "";

	m_sServerVersion = "";
	if(!Connection->InMessage.HeaderExist("NV_SERVER_VERSION", 1))
		m_sServerVersion = Connection->InMessage.GetHeader("NV_SERVER_VERSION", 1);
	CleanString(m_sServerVersion, true);

	if(Connection->InMessage.GetHeader("NVERROR", 1) == "yes")
	{
		Connection->InMessage.Clear();
		if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
		{
			if(CurrentConnection != NULL)
				GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
			CurrentConnection = NULL;
		}
		return false;
	}

	ServiceHere = Command->m_sService;
	MakeUpper(ServiceHere);			// The service is set to uppercase by the server
	CleanString(ServiceHere);
	if(ServiceHere != Connection->InMessage.GetHeader("NV_SERVICE", 2))
	{
		Command->Error.SetError("SYSTEM", "REQUEST", NV_REQUEST_NETWORK_ERROR, ConnectedServer);
		Connection->InMessage.Clear();
		if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
		{
			if(CurrentConnection != NULL)
				GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
			CurrentConnection = NULL;
		}
		return false;
	}

	// First get the level 1 error code
	string sErrorCodeHere = Connection->InMessage.GetHeader("NV_ERROR_CODE", 1);
	int iErrorCodeHere = atoi(sErrorCodeHere.c_str());
	string sErrorServiceHere = Connection->InMessage.GetHeader("NV_ERROR_SERVICE", 1);
	string sErrorClassHere = Connection->InMessage.GetHeader("NV_ERROR_CLASS", 1);
	string sErrorDescHere = Connection->InMessage.GetHeader("NV_ERROR_DESC", 1);
	string sErrorInfoHere = Connection->InMessage.GetHeader("NV_ERROR_INFO", 1);
	Command->Error.SetError(sErrorServiceHere, sErrorClassHere, iErrorCodeHere, sErrorInfoHere);
	if(iErrorCodeHere != NV_OK)
	{
		Connection->InMessage.Clear();
		if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
		{
			if(CurrentConnection != NULL)
				GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
			CurrentConnection = NULL;
		}
		return false;
	}

	// Then get the level 2 error code
	sErrorCodeHere = Connection->InMessage.GetHeader("NV_ERROR_CODE", 2);
	iErrorCodeHere = atoi(sErrorCodeHere.c_str());
	sErrorServiceHere = Connection->InMessage.GetHeader("NV_ERROR_SERVICE", 2);
	sErrorClassHere = Connection->InMessage.GetHeader("NV_ERROR_CLASS", 2);
	sErrorDescHere = Connection->InMessage.GetHeader("NV_ERROR_DESC", 2);
	sErrorInfoHere = Connection->InMessage.GetHeader("NV_ERROR_INFO", 2);
	Command->Error.SetError(sErrorServiceHere, sErrorClassHere, iErrorCodeHere, sErrorInfoHere);
	if(iErrorCodeHere != NV_OK)
	{
		Connection->InMessage.Clear();
		if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
		{
			if(CurrentConnection != NULL)
				GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
			CurrentConnection = NULL;
		}
		return false;
	}

	// We must set the data for the GetObject command of the system service
	CommandHere = Command->m_sCommand;
	MakeUpper(CommandHere);
	CleanString(CommandHere);

	ClassHere = Command->m_sClass;
	MakeUpper(ClassHere);
	CleanString(ClassHere);
	if((iErrorCodeHere == NV_OK) && (ServiceHere == "SYSTEM") && (ClassHere == "OBJECT") && (CommandHere == "GET"))
	{

		// Get server object name and type that should be in returned parameters
		string ServerObjName = Connection->InMessage.GetHeader("NAME", 2);
		string ObjType = Connection->InMessage.GetHeader("TYPE", 2);
		CleanString(ServerObjName);
		CleanString(ObjType);
		if(!ServerObjName.length())
			ServerObjName = NV_OBJ_DEFAULT_NAME;

		if(!ObjType.length())
		{
			Command->Error.SetError("SYSTEM", "OBJECT", NV_OBJECT_NO_TYPE, ServerObjName);
			Connection->InMessage.Clear();
			if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
			{
				if(CurrentConnection != NULL)
					GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
				CurrentConnection = NULL;
			}
			return false;
		}

		int iObjectType = atoi(ObjType.c_str());
		if((iObjectType < 0) || (iObjectType > NV_OBJTYPE_MAXTYPEVALUE))
			iObjectType = NV_OBJTYPE_UNKNOWN;

		if(iObjectType == NV_OBJTYPE_UNKNOWN)
		{
			Command->Error.SetError("SYSTEM", "OBJECT", NV_OBJECT_UNKNOWN_TYPE, ServerObjName);
			Connection->InMessage.Clear();
			if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
			{
				if(CurrentConnection != NULL)
					GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
				CurrentConnection = NULL;
			}
			return false;
		}

		// Get local object name that may be in parameters
		// If this parameter is not here or blank, the local object name is
		// the same than the server object name
		string LocalObjName = Command->Parameters.GetValue("LNAME");
		CleanString(LocalObjName);
		if(!LocalObjName.length())
			LocalObjName = ServerObjName;

		// We first remove the object if it ever exist in the container
		// In fact, the new object should replace the existing one
		Session->InContainer->RemoveObject(LocalObjName.c_str());

		if(!Session->InContainer->CreateObject(LocalObjName.c_str(), iObjectType))
		{
			Command->Error.SetError("SYSTEM", "OBJECT", NV_OBJECT_CANNOT_CREATE, LocalObjName);
			Connection->InMessage.Clear();
			if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
			{
				if(CurrentConnection != NULL)
					GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
				CurrentConnection = NULL;
			}
			return false;
		}

		if(iObjectType == NV_OBJTYPE_FILE)
		{
			// Check the persist parameter
			string PersistTimeFileInCommand = Command->Parameters.GetValue("PERSIST");
			CleanString(PersistTimeFileInCommand);
			MakeUpper(PersistTimeFileInCommand);
			int iPersistTimeFileInCommand = 0;	// The file is not persistant by default on the server
			if(PersistTimeFileInCommand.length())
				iPersistTimeFileInCommand = atoi(PersistTimeFileInCommand.c_str());
			if(iPersistTimeFileInCommand < 0)
				iPersistTimeFileInCommand = -1;


			string FileExtensionInCommand = Command->Parameters.GetValue("EXTENSION");
			CleanString(FileExtensionInCommand);
			if(!FileExtensionInCommand.length())
			{
				if(iPersistTimeFileInCommand == -1)
					FileExtensionInCommand = "obj";
				else
					FileExtensionInCommand = "tmp";
			}
			string FilePrefixInCommand = Command->Parameters.GetValue("PREFIX");
			CleanString(FilePrefixInCommand);
			if(!FilePrefixInCommand.length())
				FilePrefixInCommand = Session->Application->GetAppNameForFiles();
			string FileSuffixInCommand = Command->Parameters.GetValue("SUFFIX");
			CleanString(FileSuffixInCommand);
			string TargetDir = Session->Application->GetWorkDir();
			if(iPersistTimeFileInCommand == -1)
			{
				// If the file is persistant, it's created in the application File directory
				TargetDir = Session->Application->GetFileDir();
			}

			NvObjFile *Object = (NvObjFile *)(Session->InContainer->GetObject(LocalObjName.c_str()));
			if(Object != NULL)
			{

				if(!Object->SetAutoFileName(TargetDir, FileExtensionInCommand, FilePrefixInCommand, FileSuffixInCommand))
				{
					char *ErrorMessage = (char *)malloc(LocalObjName.length() + TargetDir.length() + FileExtensionInCommand.length()  + FilePrefixInCommand.length()  + FileSuffixInCommand.length() + 100);
					sprintf(ErrorMessage, "obj:%s - dir:%s - ext:%s - prefix:%s - suffix:%s", LocalObjName.c_str(), TargetDir.c_str(), FileExtensionInCommand.c_str(), FilePrefixInCommand.c_str(), FileSuffixInCommand.c_str());
					Command->Error.SetError("SYSTEM", "OBJECT", NV_FILE_CANNOT_CREATE, ErrorMessage);
					free(ErrorMessage);
					Session->InContainer->RemoveObject(LocalObjName.c_str());
					Connection->InMessage.Clear();
					if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
					{
						if(CurrentConnection != NULL)
							GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
						CurrentConnection = NULL;
					}
					return false;
				}
				Object->SetPersistTime(iPersistTimeFileInCommand);
			}
		}
		else
		{
			if(iObjectType == NV_OBJTYPE_INDSTRINGLIST)
			{
				NvObjIndStringList *Object = (NvObjIndStringList *)(Session->InContainer->GetObject(LocalObjName.c_str()));
				if(Object != NULL)
				{
					string CaseSensitive = Command->Parameters.GetValue("CASE_SENSITIVE");
					CleanString(CaseSensitive);
					MakeUpper(CaseSensitive);
					if(CaseSensitive == "YES")
						Object->m_bCaseSensitive = true;
				}
			}
		}


		// PM TODO
		// We can call a Decode function here that operates on Command->Data
		// This Decode function can also uncompress the data stream

		// We populate the object
		bool bPopulateFromData = true;
		if(iObjectType == NV_OBJTYPE_FILE)
		{
			if(Connection->InMessage.vDatas[2]->DataFile.length())
			{
				bPopulateFromData = false;
				// We populate the file object directly from the data file in command
				NvObjFile *Object = (NvObjFile *)(Session->InContainer->GetObject(LocalObjName.c_str()));
				if(Object != NULL)
				{
					string ObjectFileName = Object->GetFilePathName();
					if(!CopyFileEx(Connection->InMessage.vDatas[2]->DataFile.c_str(), ObjectFileName.c_str()))
					{
						Command->Error.SetError("SYSTEM", "OBJECT", NV_OBJECT_CANNOT_POPULATE, LocalObjName);
						Connection->InMessage.Clear();
						if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
						{
							if(CurrentConnection != NULL)
								GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
							CurrentConnection = NULL;
						}
						return false;
					}
				}
			}

			// Check the ASCII parameter
			string sAscii = Command->Parameters.GetValue("ASCII");
			CleanString(sAscii);
			if(sAscii == "YES")
			{
				NvObjFile *Object = (NvObjFile *)(Session->InContainer->GetObject(LocalObjName.c_str()));
				if(Object != NULL)
					Object->SetAscii();
			}
		}
		else
		{
			if(Connection->InMessage.vDatas[2]->DataFile.length())
				bPopulateFromData = false;
		}

		if(bPopulateFromData)
		{
			if(!Session->InContainer->PopulateObject(LocalObjName.c_str(), Connection->InMessage.vDatas[2]->Data, Connection->InMessage.vDatas[2]->DataSize))
			{
				Command->Error.SetError("SYSTEM", "OBJECT", NV_OBJECT_CANNOT_POPULATE, LocalObjName);
				Connection->InMessage.Clear();
				if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
				{
					if(CurrentConnection != NULL)
						GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
					CurrentConnection = NULL;
				}
				return false;
			}
		}
		else if(iObjectType != NV_OBJTYPE_FILE)
		{
			// The work has ever been done for a file object
			if(!Session->InContainer->PopulateObjectFromFile(LocalObjName.c_str(), Connection->InMessage.vDatas[2]->DataFile.c_str(), false))
			{
				Command->Error.SetError("SYSTEM", "OBJECT", NV_OBJECT_CANNOT_POPULATE, LocalObjName);
				Connection->InMessage.Clear();
				if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
				{
					if(CurrentConnection != NULL)
						GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
					CurrentConnection = NULL;
				}
				return false;
			}
		}
	}
	if((iErrorCodeHere == NV_OK) && (Buffer != NULL) && ((ServiceHere != "SYSTEM") || (ClassHere != "OBJECT") || (CommandHere != "GET")))
	{
		// If the command is successful and this is not an objetc:get, we get the result in the output buffer if there is one
		// The incoming data then contain the end of string character
		// The proper encoding for client has been ever set on the server side
		unsigned long InDataSize = Connection->InMessage.vDatas[2]->DataSize;
		char *Data = Connection->InMessage.vDatas[2]->Data;
		if(InDataSize)
			Data[InDataSize-1] = 0;
		string VarName = Command->Parameters.GetValue("NV_VAR");
		if(!VarName.length())
			VarName = "NV_RESULT";
		if(InDataSize)
		{
			Session->m_oVariables.SetValue(VarName.c_str(), Connection->InMessage.vDatas[2]->Data);
			Command->GetStringValue(Connection->InMessage.vDatas[2]->Data, Buffer, BufferSize, DataSize, ENCODING_NONE);
		}
		else
			Session->m_oVariables.SetValue(VarName.c_str(), "");
	}

	// we clear the input message
	Connection->InMessage.Clear();

	if(m_iAutoClose == NV_REQUEST_CLOSE_COMMAND)
	{
		if(CurrentConnection != NULL)
			GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
		CurrentConnection = NULL;
	}

	return true;
}

void NvRequest::SetServerAddress()
{	
	string Address = m_sServer;
	CleanString(Address);

	// The proxy is eventually in the address
	// at this time the address starts with proxy::proxyaddress::...
	// proxyaddress can have some optional user and password information in brackets separated by a ; character
	string AddressHere = Address;
	MakeUpper(AddressHere);
	string::size_type index = AddressHere.find("PROXY::");
	if(index != string::npos)
	{
		Address = Address.substr(7);
		index = Address.find("::");
		if(index != string::npos)
			Address = Address.substr(index+2);
	}
	CleanString(Address);

	m_sServerAddress = Address;
}


bool NvRequest::PrivateCommand(string Command, string& Result)
{
	Result = "";
	string Address = m_sServer;
	CleanString(Address);
	int Port = 1081;
	if(b_UseSSL)
		Port = 1082;
	string::size_type index = Address.find(':');
	if(index != string::npos)
	{
		string Provistr = Address.substr(index+1);
		CleanString(Provistr);
		Port = atoi(Provistr.c_str());
		Address.erase(index);
	}

	// We do not use current connection for this kind of message
	if(CurrentConnection != NULL)
	{
		GetMain()->Connections.FreeConnection((NvHttpConnection*)CurrentConnection);
		CurrentConnection = NULL;
	}
	
	NvHttpConnection *Connection = GetMain()->Connections.GetConnection(Address.c_str(), Port, m_iServerConnectionTimeOut, b_UseSSL);

	if(Connection == NULL)
		return false;
	
	Connection->sSocket.LogFile = "";
	Connection->b_ProxyHttps = false;

	CurrentConnection = Connection;

	// we clear the input message
	Connection->InMessage.Clear();
	
	// we clear the outpput message
	Connection->OutMessage.Clear();

	NvData *Data = NVNEW NvData;
	Data->sInfo = "GET ";
	Data->sInfo += Command;
	Data->sInfo += "/NVS?Private";
	Data->sInfo += " HTTP/1.1";
	Connection->OutMessage.vDatas.push_back(Data);
	
	// by default, we have no content in the body
	Connection->OutMessage.AddHeader("Content-Length", "0");
	
	// For HTTP 1.1 we must include the host header
	char buffer[20];
	sprintf(buffer, "%d", Port);
	Address += ":";
	Address += buffer;
	Connection->OutMessage.AddHeader("Host", Address);

	Connection->OutMessage.AddHeader("Nv_Source", HTTP_CLIENT_STRING);

	// We prepare the output global headers
	

	// Set NvMachineName
	Connection->OutMessage.AddHeader("Nv_Machine_Name", GetMain()->GetMachineName());

	// Set NvMachineUser
	Connection->OutMessage.AddHeader("Nv_Machine_User", GetMain()->GetMachineUser());

	// Now send the order
	string TempWorkDir = GetMain()->GetNirvaDir();
	AddDirSuffix(TempWorkDir);
	TempWorkDir += "Work";
	bool RetVal = true;
	int OrderResult = Connection->SendOrder(true, GetMain()->GetHttpMaxMemBody(), TempWorkDir.c_str());

	// we clear the output message
	Connection->OutMessage.Clear();

	if(OrderResult != 1)
	{
		// An error occured
		// we clear the input message
		Connection->InMessage.Clear();

		GetMain()->Connections.FreeConnection(Connection);
		CurrentConnection = NULL;
		return false;
	}
	
	// Get the Connection->OutMessage here


	// The result should have at least one part 
	if(!Connection->InMessage.vDatas.size())
	{
		// An error occured
		// we clear the input message
		Connection->InMessage.Clear();

		GetMain()->Connections.FreeConnection(Connection);
		CurrentConnection = NULL;
		return false;
	}


	Result = Connection->InMessage.GetHeader("NIRVA-STATUS");


	// we clear the input message
	Connection->InMessage.Clear();

	GetMain()->Connections.FreeConnection(Connection);
	CurrentConnection = NULL;

	return true;
}




/////////////////////////////////////////////////////////////////////////////
// NvHostList

NvHostList::NvHostList()
{
	m_tHosts.InsertColumn("NAME");
	m_tHosts.InsertColumn("DESCRIPTION");
	m_tHosts.InsertColumn("ADDRESSES");
	m_tHosts.InsertColumn("ACTIVE");
	m_tHosts.InsertColumn("MODE");
	m_tHosts.SetPrimaryColumn("NAME", false);
}

NvHostList::~NvHostList()
{
	Clear();
}



void NvHostList::Clear()
{
	Lock();
	m_tHosts.ClearData();
	Unlock();
}

void NvHostList::Load()
{
	NvRegistryContainer *Registry;
	Registry = &(GetMain()->SystemRegistry);

	// load the table
	// The vhosts table is stored in a table object named "vhosts" in the
	// registry key system.parameters
	NvContainer Container;
	if(Registry->GetKey("system.parameters", &Container, NV_CONTAINER_COPY_NO_APPEND, "vhosts", false, true))
	{
		NvObjTable *VhostTable = (NvObjTable *)(Container.GetObject("vhosts"));
		if(VhostTable != NULL)
			m_tHosts.CopyFrom(VhostTable);
	}
}

void NvHostList::Save()
{
	NvRegistryContainer *Registry;
	Registry = &(GetMain()->SystemRegistry);

	NvContainer Container;

	// Creates the vhosts object in the container
	Container.CreateObject("vhosts", NV_OBJTYPE_TABLE);
	NvObjTable *VhostTable = (NvObjTable *)(Container.GetObject("vhosts"));
	VhostTable->CopyFrom(&m_tHosts);

	Registry->SetKey("system.parameters", &Container, NV_CONTAINER_COPY_APPEND);
}


// USed to add a new host or to modify information for an existing host
bool NvHostList::AddHost(const char* Name, const char* Description, const char* Addresses, const char* Mode)
{
	string NameHere = Name;
	CleanString(NameHere);
	if(!NameHere.length())
		return false;
	Lock();
	int Rowindex = m_tHosts.GetRowIndex(NameHere);
	if(Rowindex == -1)
	{
		// insert if not yet existing
		m_tHosts.InsertRow();
		m_tHosts.InsertCellLine(-1,0,NameHere);
	}
	else
	{
		m_tHosts.ClearCell(Rowindex,1);
		m_tHosts.ClearCell(Rowindex,2);
		m_tHosts.ClearCell(Rowindex,3);
		m_tHosts.ClearCell(Rowindex,4);
	}

	m_tHosts.InsertCellLine(Rowindex,1,Description);
	string AddressesHere = Addresses;
	while(AddressesHere.length())
	{
		string Address = AddressesHere;
		string::size_type index = AddressesHere.find('\n');
		if(index != string::npos)
		{
			Address = AddressesHere.substr(0, index);
			AddressesHere = AddressesHere.substr(index+1);
		}
		else
			AddressesHere = "";
		CleanString(Address);
		if(Address.length())
		{
			m_tHosts.InsertCellLine(Rowindex,2,Address);
		}
	}
	m_tHosts.InsertCellLine(Rowindex,3,"-1");
	string ModeHere=Mode;
	CleanString(ModeHere);
	MakeUpper(ModeHere);
	if(ModeHere != "LOAD_BALANCING")
		ModeHere = "FAILOVER";
	m_tHosts.InsertCellLine(Rowindex,4,ModeHere);
	Save();
	Unlock();
	return true;
}


void NvHostList::RemoveHost(const char* Name)
{
	string NameHere = Name;
	CleanString(NameHere);
	if(!NameHere.length())
		return;
	Lock();
	int Rowindex = m_tHosts.GetRowIndex(NameHere);
	if(Rowindex == -1)
	{
		Unlock();
		return;
	}
	m_tHosts.RemoveRow(Rowindex);
	Save();
	Unlock();
}


string NvHostList::GetActiveAddress(const char* Name)
{
	string Result = "";
	string NameHere = Name;
	CleanString(NameHere);
	if(!NameHere.length())
		return Result;
	Lock();
	int Rowindex = m_tHosts.GetRowIndex(NameHere);
	if(Rowindex == -1)
	{
		Unlock();
		return Result;
	}
	int NumAdresses = m_tHosts.GetNumCellLines(Rowindex, 2);
	Result = m_tHosts.GetCellLine(Rowindex, 3);
	if((Result == "-1") && (NumAdresses))
	{
		Result = "0";
		m_tHosts.SetCellLine(Rowindex,3,Result);
	}
	int iResult = atoi(Result.c_str());
	Result = m_tHosts.GetCellLine(Rowindex, 2, iResult);
	Unlock();
	return Result;
}


string NvHostList::GetNextAddress(const char* Name)
{
	string Result = "";
	string NameHere = Name;
	CleanString(NameHere);
	if(!NameHere.length())
		return Result;
	Lock();
	int Rowindex = m_tHosts.GetRowIndex(NameHere);
	if(Rowindex == -1)
	{
		Unlock();
		return Result;
	}
	int NumAdresses = m_tHosts.GetNumCellLines(Rowindex, 2);
	if(!NumAdresses)
	{
		Unlock();
		return Result;
	}

	Result = m_tHosts.GetCellLine(Rowindex, 3);
	int iResult = atoi(Result.c_str());
	iResult++;
	if(iResult < 0)
		iResult = 0;	// first one by default
	if(iResult >= NumAdresses)
		iResult = 0;	// first one if previous was the last one
	Result = m_tHosts.GetCellLine(Rowindex, 2, iResult);

	// It becomes the active one
	char Buf[10];
	sprintf(Buf,"%d",iResult);
	m_tHosts.SetCellLine(Rowindex,3,Buf);
	Unlock();
	return Result;
}


int NvHostList::GetNumAddresses(const char* Name)
{
	string NameHere = Name;
	CleanString(NameHere);
	if(!NameHere.length())
		return 0;
	Lock();
	int Rowindex = m_tHosts.GetRowIndex(NameHere);
	if(Rowindex == -1)
	{
		Unlock();
		return 0;
	}
	int RetVal = m_tHosts.GetNumCellLines(Rowindex, 2);
	Unlock();
	return RetVal;
}


string NvHostList::GetMode(const char* Name)
{
	string Result = "";
	string NameHere = Name;
	CleanString(NameHere);
	if(!NameHere.length())
		return Result;
	Lock();
	int Rowindex = m_tHosts.GetRowIndex(NameHere);
	if(Rowindex == -1)
	{
		Unlock();
		return Result;
	}

	Result = m_tHosts.GetCellLine(Rowindex, 4);
	Unlock();
	return Result;
}


void NvHostList::GetList(NvObjTable *Result)
{
	Lock();
	Result->CopyFrom(&m_tHosts);
	Unlock();
}








