// ***************************************************************************************
// File: scheduler.cpp
// Description: scheduler implementation
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#include "nvs.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;


/////////////////////////////////////////////////////////////////////////////
// NvSchedulerTaskList

NvSchedulerTaskList::NvSchedulerTaskList()
{
	m_sAppName = "";
	bListLoaded = false;
}

NvSchedulerTaskList::~NvSchedulerTaskList()
{
	Clear();
}



void NvSchedulerTaskList::Clear()
{
	NV_LOCKOBJECT(this);
	map<string,void *>::iterator pos;
	for(pos = m_mTasks.begin(); pos != m_mTasks.end(); ++pos)
	{
		NvSchedulerTask *Task = (NvSchedulerTask *)(pos->second);
		if(Task != NULL)
			delete Task;
	}
	m_mTasks.clear();
	NV_UNLOCKOBJECT(this);
}

// Load
// Loads the task list from the system registry
bool NvSchedulerTaskList::Load()
{
	if(!m_sAppName.length())
		return false;

	NV_LOCKOBJECT(this);
	if(bListLoaded)
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}

	NvApplication *Application = GetMain()->Applications.GetApplication(m_sAppName.c_str());
	if(Application == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	NvRegistryContainer *Registry;
	Registry = &(Application->AppRegistry);

	// load the tasks
	// The tasks are individual indstringlist objects in registry key system.tasks
	GetMain()->Logs.Write("Scheduler", "Load task list from registry", "", NV_LOG_INFO, 1, m_sAppName.c_str(), "system");
	NvContainer Container;
	if(Registry->GetKey("system.tasks", &Container))
	{
		map<string,void *>::iterator pos;
		for(pos = Container.m_mObjects.begin(); pos != Container.m_mObjects.end(); ++pos)
		{
			NvObj *Object = (NvObj *)(pos->second);
			if(Object == NULL)
				continue;
			if(Object->GetType() != NV_OBJTYPE_INDSTRINGLIST)
				continue;

			NvObjIndStringList *TaskObj = (NvObjIndStringList*)Object;

			// Get task parameters

			// Get general parameters
			string Value = TaskObj->GetValue("Name");
			CleanString(Value);
			if(!Value.length()) // The task is identified by its name
				continue;
			// Get it again because the name has not to be especially cleaned
			Value = TaskObj->GetValue("Name");
			if(InternalGetTask(Value.c_str()) != NULL)	// Check doubles
				continue;

			// The object name is the MD5 string of the task name
			// SO we check the correspondance
			string Md5Name = MD5_string(Value);
			MakeLower(Md5Name);
			string ObjectName = Object->GetName();
			MakeLower(ObjectName);
			if(Md5Name != ObjectName)
				continue;

			NvSchedulerTask *Task = NVNEW NvSchedulerTask;
			Task->m_sName = Value;

			Value = TaskObj->GetValue("Desc");
			Task->m_sDesc = Value;
			
			Task->m_sAppName = m_sAppName;
			
			Value = TaskObj->GetValue("Status");
			int iValue = atoi(Value.c_str());
			if(iValue < 0)
				iValue = 0;
			if(iValue > NV_SCTASK_MAXSTATUS)
				iValue = NV_SCTASK_MAXSTATUS;
			if(iValue == NV_SCTASK_STATUS_RUNNING)
				iValue = NV_SCTASK_STATUS_STOPPED;	// Has been stopped hardly
			if(iValue == NV_SCTASK_STATUS_INQUEUE)
				iValue = NV_SCTASK_STATUS_NOT_STARTED;	// Has been stopped hardly
			Task->m_iStatus = iValue;
			
			Value = TaskObj->GetValue("User");
			Task->m_sUser = Value;
			
			Value = TaskObj->GetValue("Language");
			Task->m_sLanguage = Value;
			
			Value = TaskObj->GetValue("ProcTask");
			Task->m_sProcTask = Value;
			
			Value = TaskObj->GetValue("ProcFailed");
			Task->m_sProcFailed = Value;
			
			Value = TaskObj->GetValue("ProcSuccess");
			Task->m_sProcSuccess = Value;
			
			Value = TaskObj->GetValue("ProcInit");
			Task->m_sProcInit = Value;
			
			Value = TaskObj->GetValue("ProcExit");
			Task->m_sProcExit = Value;
			
			Value = TaskObj->GetValue("ErrorCode");
			iValue = atoi(Value.c_str());
			if(iValue < 0)
				iValue = NV_OK;
			Task->Error.SetError(TaskObj->GetValue("ErrorService"), TaskObj->GetValue("ErrorClass"), iValue, TaskObj->GetValue("ErrorInfo"));
			
			Value = TaskObj->GetValue("LastStartTime");
			iValue = atoi(Value.c_str());
			Task->tLastStartTime = (time_t)iValue;
			
			Value = TaskObj->GetValue("LastEndTime");
			iValue = atoi(Value.c_str());
			Task->tLastEndTime = (time_t)iValue;
			
			Value = TaskObj->GetValue("Enable");
			iValue = atoi(Value.c_str());
			if(iValue < 0)
				iValue = 0;
			if(iValue == 0)
				Task->m_bEnable = false;
			else
				Task->m_bEnable = true;
			
			Value = TaskObj->GetValue("StartDate");
			iValue = atoi(Value.c_str());
			Task->tStartDate = (time_t)iValue;
			
			Value = TaskObj->GetValue("StartTime");
			iValue = atoi(Value.c_str());
			Task->tStartTime = (time_t)iValue;
			
			Value = TaskObj->GetValue("TimeSpan");
			iValue = atoi(Value.c_str());
			if(iValue < 1)
				iValue = 1;
			if(iValue > 86399)
				iValue = 86399;
			Task->m_iTimeSpan = iValue;
			
			Value = TaskObj->GetValue("FrequencyType");
			iValue = atoi(Value.c_str());
			if(iValue < 0)
				iValue = 0;
			if(iValue > NV_SCTASK_MAXFREQUENCY)
				iValue = NV_SCTASK_MAXFREQUENCY;
			Task->m_iFrequencyType = iValue;
			
			
			
			// Get daily repeat mode parameters
			Value = TaskObj->GetValue("RepeatMode");
			iValue = atoi(Value.c_str());
			if(iValue < 0)
				iValue = 0;
			if(iValue == 0)
				Task->m_bRepeatMode = false;
			else
				Task->m_bRepeatMode = true;
			

			Value = TaskObj->GetValue("Delay");
			iValue = atoi(Value.c_str());
			if(iValue < 1)
				iValue = 1;
			if(iValue > 86399)
				iValue = 86399;
			Task->m_iDelay = iValue;
			
			Value = TaskObj->GetValue("Occurences");
			iValue = atoi(Value.c_str());
			if(iValue < 0)
				iValue = 0;
			Task->m_iOccurences = iValue;
			
			Value = TaskObj->GetValue("ActualOccurence");
			iValue = atoi(Value.c_str());
			if(iValue < 0)
				iValue = 0;
			Task->m_iActualOccurence = iValue;
			
			Value = TaskObj->GetValue("FirstOccurenceStartDate");
			iValue = atoi(Value.c_str());
			Task->tFirstOccurenceStartDate = (time_t)iValue;

			// Get daily frequency parameters
			Value = TaskObj->GetValue("DayEvery");
			iValue = atoi(Value.c_str());
			if(iValue < 1)
				iValue = 1;
			if(iValue > 365)
				iValue = 365;
			Task->m_iDayEvery = iValue;

			// Get weekly frequency parameters
			Value = TaskObj->GetValue("WeekEvery");
			iValue = atoi(Value.c_str());
			if(iValue < 1)
				iValue = 1;
			if(iValue > 52)
				iValue = 52;
			Task->m_iWeekEvery = iValue;

			Value = TaskObj->GetValue("WeekDays");
			iValue = atoi(Value.c_str());
			Task->m_iWeekDays = iValue;

			// Get monthly frequency parameters
			Value = TaskObj->GetValue("MonthMonths");
			iValue = atoi(Value.c_str());
			Task->m_iMonthMonths = iValue;

			Value = TaskObj->GetValue("MonthDayBased");
			iValue = atoi(Value.c_str());
			if(iValue < 0)
				iValue = 0;
			if(iValue == 0)
				Task->m_bMonthDayBased = false;
			else
				Task->m_bMonthDayBased = true;

			Value = TaskObj->GetValue("MonthDays");
			iValue = atoi(Value.c_str());
			Task->m_iMonthDays = iValue;

			Value = TaskObj->GetValue("MonthWeekDayPosition");
			iValue = atoi(Value.c_str());
			if(iValue < 1)
				iValue = 1;
			if(iValue > 5)
				iValue = 5;
			Task->m_iMonthWeekDayPosition = iValue;

			Value = TaskObj->GetValue("MonthWeekDay");
			iValue = atoi(Value.c_str());
			if(iValue < 1)
				iValue = 1;
			if(iValue > 7)
				iValue = 7;
			Task->m_iMonthWeekDay = iValue;

			// Resave for eventual changes
			Task->Save();

			// Add the task to the task list
			string RegisteredTaskName = Task->m_sName;
			MakeUpper(RegisteredTaskName);
			if(!RegisteredTaskName.length())
			{
				Task->Remove();
				delete Task;
				continue;
			}
			m_mTasks[RegisteredTaskName] = (void *)Task;
		}
	}

	GetMain()->Applications.FreeApplication(m_sAppName.c_str());

	bListLoaded = true;
	NV_UNLOCKOBJECT(this);

	return true;
}


NvSchedulerTask* NvSchedulerTaskList::InternalGetTask(const char *Name)
{
	if(Name == NULL)
		return NULL;

	string NameHere = Name;
	
	CleanString(NameHere);
	MakeUpper(NameHere);
	NvSchedulerTask *FoundTask = NULL;

	if(m_mTasks.find(NameHere) == m_mTasks.end())
		return NULL;

	FoundTask = (NvSchedulerTask *)(m_mTasks[NameHere]);

	return FoundTask;
}


// Used to get an existing task
NvSchedulerTask* NvSchedulerTaskList::GetTask(const char *Name)
{
	if(Name == NULL)
		return NULL;

	// Load task list if not yet done
	Load();

	NvSchedulerTask *ReturnedTask = NULL;
	NV_LOCKOBJECT(this);
	ReturnedTask = InternalGetTask(Name);

	if(ReturnedTask != NULL)
		ReturnedTask->IncInuse();

	NV_UNLOCKOBJECT(this);
	return ReturnedTask;
}


bool NvSchedulerTaskList::CreateTask(const char *Name, const char *Desc, const char *User, const char *Language)
{
	string NameHere = "";
	if(Name != NULL)
		NameHere = Name;
	CleanString(NameHere);
	if(!NameHere.length())
		return false;
	NameHere = Name;	// we take the original name as task name

	string DescHere = "";
	if(Desc != NULL)
		DescHere = Desc;

	string UserHere = "";
	if(User != NULL)
		UserHere = User;


	// Load task list if not yet done
	Load();

	NV_LOCKOBJECT(this);

	if(InternalGetTask(Name) != NULL)
	{
		// The task ever exist
		NV_UNLOCKOBJECT(this);
		return false;
	}
	
	NvSchedulerTask *Task = NVNEW NvSchedulerTask;
	if(Task == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	// We don't need to lock the task to access its parameters because it's not yet registered in the task list
	Task->m_sName = NameHere;
	Task->m_sDesc = DescHere;
	Task->m_sUser = UserHere;
	if(Language != NULL)
		Task->m_sLanguage = Language;
	Task->m_sAppName = m_sAppName;
	if(!Task->Save())	// Save the task parameters to application registry
	{
		Task->Remove();
		delete Task;
		NV_UNLOCKOBJECT(this);
		return false;
	}

	string RegisteredTaskName = Task->m_sName;
	CleanString(RegisteredTaskName);
	MakeUpper(RegisteredTaskName);
	if(!RegisteredTaskName.length())
	{
		Task->Remove();
		delete Task;
		NV_UNLOCKOBJECT(this);
		return false;
	}
	
	// Then add it to the list
	m_mTasks[RegisteredTaskName] = (void *)Task;
	GetMain()->Logs.Write("Scheduler", "Create task", RegisteredTaskName.c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");

	NV_UNLOCKOBJECT(this);
	
	return true;
}


bool NvSchedulerTaskList::SetTaskParam(NvCommand *CommandToExecute)
{
	// Load task list of not yet done
	Load();

	string TaskName = CommandToExecute->Parameters.GetValue("NAME");

	NvSchedulerTask *FoundTask = NULL;
	NV_LOCKOBJECT(this);
	FoundTask = InternalGetTask(TaskName.c_str());
	if(FoundTask == NULL)
	{
		CommandToExecute->Error.SetError("SYSTEM", "SCHEDULER", NV_SCHEDULER_INVALID_TASK, TaskName);
		NV_UNLOCKOBJECT(this);
		return false;
	}
	
	NV_LOCKLOCK(&(FoundTask->ProtectInUse));

	if(CommandToExecute->Parameters.KeyExist("DESCRIPTION"))
		FoundTask->m_sDesc = CommandToExecute->Parameters.GetValue("DESCRIPTION");

	if(CommandToExecute->Parameters.KeyExist("USER"))
		FoundTask->m_sUser = CommandToExecute->Parameters.GetValue("USER");

	if(CommandToExecute->Parameters.KeyExist("PROCT"))
		FoundTask->m_sProcTask = CommandToExecute->Parameters.GetValue("PROCT");

	if(CommandToExecute->Parameters.KeyExist("PROCF"))
		FoundTask->m_sProcFailed = CommandToExecute->Parameters.GetValue("PROCF");

	if(CommandToExecute->Parameters.KeyExist("PROCS"))
		FoundTask->m_sProcSuccess = CommandToExecute->Parameters.GetValue("PROCS");

	if(CommandToExecute->Parameters.KeyExist("PROCI"))
		FoundTask->m_sProcInit = CommandToExecute->Parameters.GetValue("PROCI");

	if(CommandToExecute->Parameters.KeyExist("PROCE"))
		FoundTask->m_sProcExit = CommandToExecute->Parameters.GetValue("PROCE");

	if(CommandToExecute->Parameters.KeyExist("START_DATE"))
	{
		string StartDate = CommandToExecute->Parameters.GetValue("START_DATE");
		CleanString(StartDate);
		MakeUpper(StartDate);
		if(StartDate.length())
		{
			if(!ArrangeDate(StartDate))
			{
				CommandToExecute->Error.SetError("SYSTEM", "COMMAND", NV_COMMAND_BAD_DATE_FORMAT, StartDate);
				NV_UNLOCKLOCK(&(FoundTask->ProtectInUse));
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}
		else
		{
			// If start date is not given, we use the actual day
			time_t tTime = time(NULL);
			struct tm tCurrentTime;
			localtime_r(&tTime, &tCurrentTime);
			char Buf[50];
			sprintf(Buf, "%02d.%02d.%04d", tCurrentTime.tm_mday, tCurrentTime.tm_mon + 1, tCurrentTime.tm_year + 1900);
			StartDate = Buf;
		}

		int iStartDay;
		int iStartMonth;
		int iStartYear;
		string Day = "";
		string Month = "";
		string Year = "";
		string::size_type index;
		index = StartDate.find('.');
		if(index != string::npos)
		{
			Day = StartDate.substr(0, index);
			StartDate = StartDate.substr(index+1);
		}
		index = StartDate.find('.');
		if(index != string::npos)
		{
			Month = StartDate.substr(0, index);
			Year = StartDate.substr(index+1);
		}

		iStartDay = atoi(Day.c_str());
		iStartMonth = atoi(Month.c_str());
		iStartMonth--;
		iStartYear = atoi(Year.c_str());
		iStartYear -= 1900;
		struct tm ftime;
		ftime.tm_isdst = 0;
		ftime.tm_sec = 0;
		ftime.tm_min = 0;
		ftime.tm_hour = 0;
		ftime.tm_mday = iStartDay;
		ftime.tm_mon = iStartMonth;
		ftime.tm_year = iStartYear;
		ftime.tm_isdst = -1;
		time_t iStartDate = mktime(&ftime); 
		if(iStartDate == -1)
		{
			CommandToExecute->Error.SetError("SYSTEM", "COMMAND", NV_COMMAND_BAD_DATE_FORMAT);
			NV_UNLOCKLOCK(&(FoundTask->ProtectInUse));
			NV_UNLOCKOBJECT(this);
			return false;
		}

		// Only the date part has meaning in tStartDate
		FoundTask->tStartDate = iStartDate;
	}

	if(CommandToExecute->Parameters.KeyExist("START_TIME"))
	{
		string StartTime = CommandToExecute->Parameters.GetValue("START_TIME");
		CleanString(StartTime);
		MakeUpper(StartTime);
		if(StartTime.length())
		{
			if(!ArrangeHour(StartTime))
			{
				CommandToExecute->Error.SetError("SYSTEM", "COMMAND", NV_COMMAND_BAD_DATE_FORMAT, StartTime);
				NV_UNLOCKLOCK(&(FoundTask->ProtectInUse));
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}
		else
		{
			// If start time is not given, we use the actual time
			time_t tTime = time(NULL);
			struct tm tCurrentTime;
			localtime_r(&tTime, &tCurrentTime);
			char Buf[50];
			sprintf(Buf, "%02d%02d%02d", tCurrentTime.tm_hour, tCurrentTime.tm_min, tCurrentTime.tm_sec);
			StartTime = Buf;
		}

		// We use the date of today for the time functions
		// but in fact, only the time part will have meaning in the tStartTime task member
		time_t tTime = time(NULL);
		struct tm tCurrentTime;
		localtime_r(&tTime, &tCurrentTime);
		char Buf[50];
		sprintf(Buf, "%02d.%02d.%04d", tCurrentTime.tm_mday, tCurrentTime.tm_mon + 1, tCurrentTime.tm_year + 1900);
		string StartDate = Buf;

		int iStartDay;
		int iStartMonth;
		int iStartYear;
		string Day = "";
		string Month = "";
		string Year = "";
		string::size_type index;
		index = StartDate.find('.');
		if(index != string::npos)
		{
			Day = StartDate.substr(0, index);
			StartDate = StartDate.substr(index+1);
		}
		index = StartDate.find('.');
		if(index != string::npos)
		{
			Month = StartDate.substr(0, index);
			Year = StartDate.substr(index+1);
		}

		iStartDay = atoi(Day.c_str());
		iStartMonth = atoi(Month.c_str());
		iStartMonth--;
		iStartYear = atoi(Year.c_str());
		iStartYear -= 1900;
		struct tm ftime;
		ftime.tm_isdst = 0;
		ftime.tm_sec = 1;
		ftime.tm_min = 0;
		ftime.tm_hour = 0;
		ftime.tm_mday = iStartDay;
		ftime.tm_mon = iStartMonth;
		ftime.tm_year = iStartYear;
		ftime.tm_isdst = -1;
		if(StartTime.length() == 6)
		{
			string Hour = StartTime.substr(0, 2);
			string Min = StartTime.substr(2, 2);
			string Sec = StartTime.substr(4, 2);
			ftime.tm_hour = atoi(Hour.c_str());
			ftime.tm_min = atoi(Min.c_str());
			ftime.tm_sec = atoi(Sec.c_str());
		}
		time_t iStartDate = mktime(&ftime); 
		if(iStartDate == -1)
		{
			CommandToExecute->Error.SetError("SYSTEM", "COMMAND", NV_COMMAND_BAD_DATE_FORMAT);
			NV_UNLOCKLOCK(&(FoundTask->ProtectInUse));
			NV_UNLOCKOBJECT(this);
			return false;
		}

		// Only the time part has meaning in tStartTime
		FoundTask->tStartTime = iStartDate;
	}
	
	if(CommandToExecute->Parameters.KeyExist("TIME_SPAN"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("TIME_SPAN");
		CleanString(sValue);
		MakeUpper(sValue);
		int iValue = atoi(sValue.c_str());
		if(iValue == 0)
			iValue = 86399;
		if(iValue < 1)
			iValue = 1;
		if(iValue > 86399)
			iValue = 86399;
		FoundTask->m_iTimeSpan = iValue;
	}
	
	if(CommandToExecute->Parameters.KeyExist("FREQUENCY"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("FREQUENCY");
		CleanString(sValue);
		MakeUpper(sValue);
		int iValue = NV_SCTASK_FREQUENCY_ONETIME;
		if(sValue == "DAILY")
			iValue = NV_SCTASK_FREQUENCY_DAILY;
		if(sValue == "WEEKLY")
			iValue = NV_SCTASK_FREQUENCY_WEEKLY;
		if(sValue == "MONTHLY")
			iValue = NV_SCTASK_FREQUENCY_MONTHLY;
		FoundTask->m_iFrequencyType = iValue;
	}

	if(CommandToExecute->Parameters.KeyExist("REPEAT"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("REPEAT");
		FoundTask->m_bRepeatMode = false;
		CleanString(sValue);
		MakeUpper(sValue);
		if(sValue == "YES")
			FoundTask->m_bRepeatMode = true;
	}
	
	if(CommandToExecute->Parameters.KeyExist("DELAY"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("DELAY");
		CleanString(sValue);
		MakeUpper(sValue);
		int iValue = atoi(sValue.c_str());
		if(!sValue.length())
			iValue = 60;	// default
		if(iValue < 1)
			iValue = 1;
		if(iValue > 86399)
			iValue = 86399;
		FoundTask->m_iDelay = iValue;
	}
	
	if(CommandToExecute->Parameters.KeyExist("OCCURENCES"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("OCCURENCES");
		CleanString(sValue);
		MakeUpper(sValue);
		int iValue = atoi(sValue.c_str());
		if(!sValue.length())
			iValue = 1;	// default
		if(iValue < 0)
			iValue = 0;		// Infinite
		FoundTask->m_iOccurences = iValue;
	}
	
	if(CommandToExecute->Parameters.KeyExist("CLEAR_OCC"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("CLEAR_OCC");
		CleanString(sValue);
		MakeUpper(sValue);
		if(sValue == "YES")
			FoundTask->m_iActualOccurence = 0;
	}
	
	if(CommandToExecute->Parameters.KeyExist("DAY_EVERY"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("DAY_EVERY");
		CleanString(sValue);
		MakeUpper(sValue);
		int iValue = atoi(sValue.c_str());
		if(!sValue.length())
			iValue = 1;	// default
		if(iValue < 1)
			iValue = 1;
		if(iValue > 365)
			iValue = 365;
		FoundTask->m_iDayEvery = iValue;
	}
	
	if(CommandToExecute->Parameters.KeyExist("WEEK_EVERY"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("WEEK_EVERY");
		CleanString(sValue);
		MakeUpper(sValue);
		int iValue = atoi(sValue.c_str());
		if(!sValue.length())
			iValue = 1;	// default
		if(iValue < 1)
			iValue = 1;
		if(iValue > 52)
			iValue = 52;
		FoundTask->m_iWeekEvery = iValue;
	}

	if(CommandToExecute->Parameters.KeyExist("WEEK_DAYS"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("WEEK_DAYS");
		CleanString(sValue);
		MakeUpper(sValue);
		NvObjStringList Values;
		Values.Add(sValue, ";");
		int iFinalValue = 0;
		int i;
		for(i=0; i<Values.GetSize(); i++)
		{
			int iValue = atoi(Values.GetValue(i).c_str());
			if(iValue < 1)
				continue;
			if(iValue > 7)
				continue;
			iValue--;
			iFinalValue |= 1 << iValue;
		}
		FoundTask->m_iWeekDays = iFinalValue;
	}

	if(CommandToExecute->Parameters.KeyExist("MONTH_MONTHS"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("MONTH_MONTHS");
		CleanString(sValue);
		MakeUpper(sValue);
		NvObjStringList Values;
		Values.Add(sValue, ";");
		int iFinalValue = 0;
		int i;
		for(i=0; i<Values.GetSize(); i++)
		{
			int iValue = atoi(Values.GetValue(i).c_str());
			if(iValue < 1)
				continue;
			if(iValue > 12)
				continue;
			iValue--;
			iFinalValue |= 1 << iValue;
		}
		FoundTask->m_iMonthMonths = iFinalValue;
	}


	if(CommandToExecute->Parameters.KeyExist("MONTH_DAY_BASED"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("MONTH_DAY_BASED");
		FoundTask->m_bMonthDayBased = true;
		CleanString(sValue);
		MakeUpper(sValue);
		if(sValue == "NO")
			FoundTask->m_bMonthDayBased = false;
	}

	if(CommandToExecute->Parameters.KeyExist("MONTH_DAYS"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("MONTH_DAYS");
		CleanString(sValue);
		MakeUpper(sValue);
		NvObjStringList Values;
		Values.Add(sValue, ";");
		int iFinalValue = 0;
		int i;
		for(i=0; i<Values.GetSize(); i++)
		{
			int iValue = atoi(Values.GetValue(i).c_str());
			if(iValue < 1)
				continue;
			if(iValue > 31)
				continue;
			iValue--;
			iFinalValue |= 1 << iValue;
		}
		FoundTask->m_iMonthDays = iFinalValue;
	}

	
	if(CommandToExecute->Parameters.KeyExist("MONTH_WEEK"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("MONTH_WEEK");
		CleanString(sValue);
		MakeUpper(sValue);
		int iValue = NV_SCTASK_WEEK_FIRST;
		if(sValue == "SECOND")
			iValue = NV_SCTASK_WEEK_SECOND;
		if(sValue == "THIRD")
			iValue = NV_SCTASK_WEEK_THIRD;
		if(sValue == "FOURTH")
			iValue = NV_SCTASK_WEEK_FOURTH;
		if(sValue == "LAST")
			iValue = NV_SCTASK_WEEK_LAST;
		FoundTask->m_iMonthWeekDayPosition = iValue;
	}
	
	if(CommandToExecute->Parameters.KeyExist("MONTH_WEEK_DAY"))
	{
		string sValue = CommandToExecute->Parameters.GetValue("MONTH_WEEK_DAY");
		CleanString(sValue);
		MakeUpper(sValue);
		int iValue = atoi(sValue.c_str());
		if(!sValue.length())
			iValue = 1;	// default
		if(iValue < 1)
			iValue = 1;
		if(iValue > 7)
			iValue = 7;
		FoundTask->m_iMonthWeekDay = iValue;
	}

	if(!FoundTask->Save())
		return false;

	GetMain()->Logs.Write("Scheduler", "Change task parameters", FoundTask->m_sName.c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");

	NV_UNLOCKLOCK(&(FoundTask->ProtectInUse));

	NV_UNLOCKOBJECT(this);

	return true;
}


void NvSchedulerTaskList::FreeTask(const char *Name)
{

	// Load task list of not yet done
	Load();

	NvSchedulerTask *FoundTask = NULL;
	NV_LOCKOBJECT(this);
	FoundTask = InternalGetTask(Name);
	if(FoundTask != NULL)
	{
		FoundTask->Session = NULL;
		FoundTask->DecInUse();
	}
	NV_UNLOCKOBJECT(this);
}


bool NvSchedulerTaskList::RemoveTask(const char *Name, int WaitTime)
{

	// Load task list of not yet done
	Load();

	NvSchedulerTask *FoundTask = NULL;
	NV_LOCKOBJECT(this);
	FoundTask = InternalGetTask(Name);
	if(FoundTask == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}

	if(FoundTask->GetInUse())
	{
		if(WaitTime > 0)
		{
			// Wait first for the task to be free
			// But a maximum of WaitTime
			// We check every 200 ms

			int i;
			int NumLoop = (1000 * WaitTime)/200;
			for(i=0; i<NumLoop; i++)
			{
				if(!FoundTask->GetInUse())
					break;
				SLEEP(200);
			}
		}
	}

	// We check again
	if(FoundTask->GetInUse())
	{
		NV_UNLOCKOBJECT(this);	// Cannot remove a task while in use
		return false;
	}

	string RegisteredTaskName = FoundTask->GetName();
	CleanString(RegisteredTaskName);
	MakeUpper(RegisteredTaskName);
	m_mTasks.erase(RegisteredTaskName);
	
	NV_LOCKLOCK(&(FoundTask->ProtectInUse));
	FoundTask->Remove();	// Remove all task registry information
	NV_UNLOCKLOCK(&(FoundTask->ProtectInUse));
	delete FoundTask;

	GetMain()->Logs.Write("Scheduler", "Remove task", RegisteredTaskName.c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");
	NV_UNLOCKOBJECT(this);

	return true;
}


bool NvSchedulerTaskList::EnableTask(const char *Name, bool Enable)
{

	// Load task list of not yet done
	Load();

	NvSchedulerTask *FoundTask = NULL;
	NV_LOCKOBJECT(this);
	FoundTask = InternalGetTask(Name);
	if(FoundTask == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	
	NV_LOCKLOCK(&(FoundTask->ProtectInUse));
	bool bOldValue = FoundTask->m_bEnable;
	FoundTask->m_bEnable = Enable;
	if(FoundTask->m_bEnable != bOldValue)
		FoundTask->Save();
	NV_UNLOCKLOCK(&(FoundTask->ProtectInUse));
	if(Enable)
		GetMain()->Logs.Write("Scheduler", "Enable task", FoundTask->GetName().c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");
	else
		GetMain()->Logs.Write("Scheduler", "Disable task", FoundTask->GetName().c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");

	NV_UNLOCKOBJECT(this);

	return true;
}


bool NvSchedulerTaskList::RunTask(const char *Name)
{

	// Load task list if not yet done
	Load();

	NvSchedulerTask *FoundTask = NULL;
	NV_LOCKOBJECT(this);
	FoundTask = InternalGetTask(Name);
	if(FoundTask == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	
	NV_LOCKLOCK(&(FoundTask->ProtectInUse));
	FoundTask->m_bImmediateRun = true;
	NV_UNLOCKLOCK(&(FoundTask->ProtectInUse));

	GetMain()->Logs.Write("Scheduler", "Run task immediatly", FoundTask->GetName().c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");

	NV_UNLOCKOBJECT(this);

	return true;
}


bool NvSchedulerTaskList::TaskExist(const char *Name)
{

	// Load task list of not yet done
	Load();

	NvSchedulerTask *FoundTask = NULL;
	NV_LOCKOBJECT(this);
	FoundTask = InternalGetTask(Name);
	NV_UNLOCKOBJECT(this);
	if(FoundTask == NULL)
		return false;

	return true;
}

bool NvSchedulerTaskList::GetInfo(NvObjTable *TaskInfo, const char *Language)
{

	// Load task list if not yet done
	Load();

	TaskInfo->Clear();

	NV_LOCKOBJECT(this);

	NvSchedulerTask *Task = NULL;


	TaskInfo->InsertColumn("NAME");
	TaskInfo->InsertColumn("DESCRIPTION");
	TaskInfo->InsertColumn("STATUS");
	TaskInfo->InsertColumn("ENABLE");
	TaskInfo->InsertColumn("FREQUENCY");
	TaskInfo->InsertColumn("LASTSTART");
	TaskInfo->InsertColumn("LASTEND");
	TaskInfo->InsertColumn("ERROR");


	int j;
	map<string,void *>::iterator pos;
	for(pos = m_mTasks.begin(), j=0; pos != m_mTasks.end(); ++pos)
	{
		Task = (NvSchedulerTask *)(pos->second);
		if(Task == NULL)
			continue;

		TaskInfo->InsertRow();
		NV_LOCKLOCK(&(Task->ProtectInUse));
		TaskInfo->InsertCellLine(j, 0, Task->m_sName);
		TaskInfo->InsertCellLine(j, 1, Task->m_sDesc);
		string sStatus = "";
		switch(Task->m_iStatus)
		{
			case NV_SCTASK_STATUS_NOT_STARTED:
				sStatus = "NOT_STARTED";
				break;
			case NV_SCTASK_STATUS_INQUEUE:
				sStatus = "IN_QUEUE";
				break;
			case NV_SCTASK_STATUS_RUNNING:
				sStatus = "RUNNING";
				break;
			case NV_SCTASK_STATUS_FAILED:
				sStatus = "FAILED";
				break;
			case NV_SCTASK_STATUS_SUCCESS:
				sStatus = "SUCCESS";
				break;
			case NV_SCTASK_STATUS_STOPPED:
				sStatus = "STOPPED";
				break;
		}
		TaskInfo->InsertCellLine(j, 2, sStatus);

		if(Task->m_bEnable)
			TaskInfo->InsertCellLine(j, 3, "YES");
		else
			TaskInfo->InsertCellLine(j, 3, "NO");

		string sFrequency = "";
		switch(Task->m_iFrequencyType)
		{
			case NV_SCTASK_FREQUENCY_ONETIME:
				sStatus = "ONETIME";
				break;
			case NV_SCTASK_FREQUENCY_DAILY:
				sStatus = "DAILY";
				break;
			case NV_SCTASK_FREQUENCY_WEEKLY:
				sStatus = "WEEKLY";
				break;
			case NV_SCTASK_FREQUENCY_MONTHLY:
				sStatus = "MONTHLY";
				break;
		}
		TaskInfo->InsertCellLine(j, 4, sStatus);

		time_t tTime = time(NULL);
		struct tm tCurrentTime;
		localtime_r(&tTime, &tCurrentTime);
		int ActualDay = tCurrentTime.tm_mday;
		int ActualMonth = tCurrentTime.tm_mon;
		int ActualYear = tCurrentTime.tm_year;
		time_t tTime2 = Task->tLastStartTime;
		localtime_r(&tTime2, &tCurrentTime);
		char Buf[100];
		if((tCurrentTime.tm_year == ActualYear) && (tCurrentTime.tm_mon == ActualMonth) && (tCurrentTime.tm_mday == ActualDay))
		{
			sprintf(Buf, "%02d:%02d:%02d",
							tCurrentTime.tm_hour,
							tCurrentTime.tm_min,
							tCurrentTime.tm_sec); 
		}
		else
		{
			sprintf(Buf, "%4d/%02d/%02d %02d:%02d:%02d",
							tCurrentTime.tm_year + 1900,
							tCurrentTime.tm_mon + 1,
							tCurrentTime.tm_mday,
							tCurrentTime.tm_hour,
							tCurrentTime.tm_min,
							tCurrentTime.tm_sec); 
		}
		if(tTime2 == 0)
			strcpy(Buf, "");
		TaskInfo->InsertCellLine(j, 5, Buf);
		
		tTime2 = Task->tLastEndTime;
		localtime_r(&tTime2, &tCurrentTime);
		if((tCurrentTime.tm_year == ActualYear) && (tCurrentTime.tm_mon == ActualMonth) && (tCurrentTime.tm_mday == ActualDay))
		{
			sprintf(Buf, "%02d:%02d:%02d",
							tCurrentTime.tm_hour,
							tCurrentTime.tm_min,
							tCurrentTime.tm_sec); 
		}
		else
		{
			sprintf(Buf, "%4d/%02d/%02d %02d:%02d:%02d",
							tCurrentTime.tm_year + 1900,
							tCurrentTime.tm_mon + 1,
							tCurrentTime.tm_mday,
							tCurrentTime.tm_hour,
							tCurrentTime.tm_min,
							tCurrentTime.tm_sec); 
		}
		if(tTime2 == 0)
			strcpy(Buf, "");
		TaskInfo->InsertCellLine(j, 6, Buf);
		
		string sError = "";
		if((Task->m_iStatus == NV_SCTASK_STATUS_FAILED) && (Task->Error.m_iErrorCode != NV_OK))
		{
			sError = Task->Error.m_sErrorService;
			sError += " - ";
			sError += Task->Error.m_sErrorClass;
			char Buffer[20];
			sprintf(Buffer, "%d", Task->Error.m_iErrorCode);
			sError += " - ";
			sError += Buffer;
			if(Task->Error.m_sErrorInfo.length())
			{
				sError += " - ";
				sError += Task->Error.m_sErrorInfo;
			}
			string Provistr = Task->Error.GetErrorDesc(Language);
			if(Provistr.length())
			{
				sError += " - ";
				sError += Provistr;
			}
		}
		TaskInfo->InsertCellLine(j, 7, sError);
		NV_UNLOCKLOCK(&(Task->ProtectInUse));
		j++;
	}
	NV_UNLOCKOBJECT(this);

	return true;
}



bool NvSchedulerTaskList::GetInfoEx(const char *Name, NvObjIndStringList *TaskInfo, const char *Language)
{
	// Load task list of not yet done
	Load();

	if(TaskInfo == NULL)
		return false;

	TaskInfo->RemoveAll();

	if(Name == NULL)
		return false;

	if(!strlen(Name))
		return false;

	NV_LOCKOBJECT(this);

	NvSchedulerTask *FoundTask = NULL;
	
	// Try to get the task
	FoundTask = InternalGetTask(Name);
	if(FoundTask == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	NV_LOCKLOCK(&(FoundTask->ProtectInUse));
	TaskInfo->SetValue("NAME", FoundTask->m_sName);
	TaskInfo->SetValue("DESCRIPTION", FoundTask->m_sDesc);
	
	string sStatus = "";
	switch(FoundTask->m_iStatus)
	{
		case NV_SCTASK_STATUS_NOT_STARTED:
			sStatus = "NOT_STARTED";
			break;
		case NV_SCTASK_STATUS_INQUEUE:
			sStatus = "IN_QUEUE";
			break;
		case NV_SCTASK_STATUS_RUNNING:
			sStatus = "RUNNING";
			break;
		case NV_SCTASK_STATUS_FAILED:
			sStatus = "FAILED";
			break;
		case NV_SCTASK_STATUS_SUCCESS:
			sStatus = "SUCCESS";
			break;
		case NV_SCTASK_STATUS_STOPPED:
			sStatus = "STOPPED";
			break;
	}
	TaskInfo->SetValue("STATUS", sStatus);
	
	if(FoundTask->m_bEnable)
		TaskInfo->SetValue("ENABLE", "YES");
	else
		TaskInfo->SetValue("ENABLE", "NO");

	time_t tTime = time(NULL);
	struct tm tCurrentTime;
	localtime_r(&tTime, &tCurrentTime);
	int ActualDay = tCurrentTime.tm_mday;
	int ActualMonth = tCurrentTime.tm_mon;
	int ActualYear = tCurrentTime.tm_year;
	time_t tTime2 = FoundTask->tLastStartTime;
	localtime_r(&tTime2, &tCurrentTime);
	char Buf[100];
	if((tCurrentTime.tm_year == ActualYear) && (tCurrentTime.tm_mon == ActualMonth) && (tCurrentTime.tm_mday == ActualDay))
	{
		sprintf(Buf, "%02d:%02d:%02d",
						tCurrentTime.tm_hour,
						tCurrentTime.tm_min,
						tCurrentTime.tm_sec); 
	}
	else
	{
		sprintf(Buf, "%4d/%02d/%02d %02d:%02d:%02d",
						tCurrentTime.tm_year + 1900,
						tCurrentTime.tm_mon + 1,
						tCurrentTime.tm_mday,
						tCurrentTime.tm_hour,
						tCurrentTime.tm_min,
						tCurrentTime.tm_sec); 
	}
	TaskInfo->SetValue("LAST_START_TIME", Buf);
	if(tTime2 == 0)
		TaskInfo->SetValue("LAST_START_TIME", "");
	
	tTime2 = FoundTask->tLastEndTime;
	localtime_r(&tTime2, &tCurrentTime);
	if((tCurrentTime.tm_year == ActualYear) && (tCurrentTime.tm_mon == ActualMonth) && (tCurrentTime.tm_mday == ActualDay))
	{
		sprintf(Buf, "%02d:%02d:%02d",
						tCurrentTime.tm_hour,
						tCurrentTime.tm_min,
						tCurrentTime.tm_sec); 
	}
	else
	{
		sprintf(Buf, "%4d/%02d/%02d %02d:%02d:%02d",
						tCurrentTime.tm_year + 1900,
						tCurrentTime.tm_mon + 1,
						tCurrentTime.tm_mday,
						tCurrentTime.tm_hour,
						tCurrentTime.tm_min,
						tCurrentTime.tm_sec); 
	}
	TaskInfo->SetValue("LAST_END_TIME", Buf);
	if(tTime2 == 0)
		TaskInfo->SetValue("LAST_END_TIME", "");
	
	tTime2 = FoundTask->tStartDate;
	localtime_r(&tTime2, &tCurrentTime);
	sprintf(Buf, "%4d/%02d/%02d",
					tCurrentTime.tm_year + 1900,
					tCurrentTime.tm_mon + 1,
					tCurrentTime.tm_mday); 
	TaskInfo->SetValue("START_DATE", Buf);
	
	tTime2 = FoundTask->tStartTime;
	localtime_r(&tTime2, &tCurrentTime);
	sprintf(Buf, "%02d:%02d:%02d",
						tCurrentTime.tm_hour,
						tCurrentTime.tm_min,
						tCurrentTime.tm_sec); 
	TaskInfo->SetValue("START_TIME", Buf);
	
	sprintf(Buf, "%d", FoundTask->m_iTimeSpan); 
	TaskInfo->SetValue("TIME_SPAN", Buf);

	string sError = "";
	if((FoundTask->m_iStatus == NV_SCTASK_STATUS_FAILED) && (FoundTask->Error.m_iErrorCode != NV_OK))
	{
		sError = FoundTask->Error.m_sErrorService;
		sError += " - ";
		sError += FoundTask->Error.m_sErrorClass;
		char Buffer[20];
		sprintf(Buffer, "%d", FoundTask->Error.m_iErrorCode);
		sError += " - ";
		sError += Buffer;
		if(FoundTask->Error.m_sErrorInfo.length())
		{
			sError += " - ";
			sError += FoundTask->Error.m_sErrorInfo;
		}
		string Provistr = FoundTask->Error.GetErrorDesc(Language);
		if(Provistr.length())
		{
			sError += " - ";
			sError += Provistr;
		}
	}
	TaskInfo->SetValue("ERROR", sError);

	TaskInfo->SetValue("USER", FoundTask->m_sUser);
	
	TaskInfo->SetValue("PROCT", FoundTask->m_sProcTask);
	TaskInfo->SetValue("PROCF", FoundTask->m_sProcFailed);
	TaskInfo->SetValue("PROCS", FoundTask->m_sProcSuccess);
	TaskInfo->SetValue("PROCI", FoundTask->m_sProcInit);
	TaskInfo->SetValue("PROCE", FoundTask->m_sProcExit);

	string sFrequency = "ONETIME";
	switch(FoundTask->m_iFrequencyType)
	{
		case NV_SCTASK_FREQUENCY_ONETIME:
			sFrequency = "ONETIME";
			break;
		case NV_SCTASK_FREQUENCY_DAILY:
			sFrequency = "DAILY";
			break;
		case NV_SCTASK_FREQUENCY_WEEKLY:
			sFrequency = "WEEKLY";
			break;
		case NV_SCTASK_FREQUENCY_MONTHLY:
			sFrequency = "MONTHLY";
			break;
	}
	TaskInfo->SetValue("FREQUENCY", sFrequency);

	
	if(FoundTask->m_bRepeatMode)
		TaskInfo->SetValue("REPEAT", "YES");
	else
		TaskInfo->SetValue("REPEAT", "NO");


	sprintf(Buf, "%d", FoundTask->m_iDelay); 
	TaskInfo->SetValue("DELAY", Buf);

	sprintf(Buf, "%d", FoundTask->m_iOccurences); 
	TaskInfo->SetValue("OCCURENCES", Buf);

	sprintf(Buf, "%d", FoundTask->m_iDayEvery); 
	TaskInfo->SetValue("DAY_EVERY", Buf);

	sprintf(Buf, "%d", FoundTask->m_iWeekEvery); 
	TaskInfo->SetValue("WEEK_EVERY", Buf);

	int i;
	string Value = "";
	for(i=0; i<7; i++)
	{
		int ValueToCheck = 1 << i;
		if(FoundTask->m_iWeekDays & ValueToCheck)
		{
			if(Value.length())
				Value += ";";
			sprintf(Buf, "%d", i+1);
			Value += Buf;
		}
	}
	TaskInfo->SetValue("WEEK_DAYS", Value);

	Value = "";
	for(i=0; i<12; i++)
	{
		int ValueToCheck = 1 << i;
		if(FoundTask->m_iMonthMonths & ValueToCheck)
		{
			if(Value.length())
				Value += ";";
			sprintf(Buf, "%d", i+1);
			Value += Buf;
		}
	}
	TaskInfo->SetValue("MONTH_MONTHS", Value);

	if(FoundTask->m_bMonthDayBased)
		TaskInfo->SetValue("MONTH_DAY_BASED", "YES");
	else
		TaskInfo->SetValue("MONTH_DAY_BASED", "NO");

	Value = "";
	for(i=0; i<31; i++)
	{
		int ValueToCheck = 1 << i;
		if(FoundTask->m_iMonthDays & ValueToCheck)
		{
			if(Value.length())
				Value += ";";
			sprintf(Buf, "%d", i+1);
			Value += Buf;
		}
	}
	TaskInfo->SetValue("MONTH_DAYS", Value);

	Value = "FIRST";
	switch(FoundTask->m_iMonthWeekDayPosition)
	{
		case NV_SCTASK_WEEK_FIRST:
			Value = "FIRST";
			break;
		case NV_SCTASK_WEEK_SECOND:
			Value = "SECOND";
			break;
		case NV_SCTASK_WEEK_THIRD:
			Value = "THIRD";
			break;
		case NV_SCTASK_WEEK_FOURTH:
			Value = "FOURTH";
			break;
		case NV_SCTASK_WEEK_LAST:
			Value = "LAST";
			break;
	}
	TaskInfo->SetValue("MONTH_WEEK", Value);

	sprintf(Buf, "%d", FoundTask->m_iMonthWeekDay); 
	TaskInfo->SetValue("MONTH_WEEK_DAY", Buf);

	NV_UNLOCKLOCK(&(FoundTask->ProtectInUse));

	NV_UNLOCKOBJECT(this);

	return true;
}


bool NvSchedulerTaskList::GetTasksToProcess(NvObjStringList *TasksToProcess)
{

	// Load task list if not yet done
	Load();

	TasksToProcess->RemoveAll();

	// Get the tasks with start time in the past and that are not in queue
	// We also get tasks having the immediate run flag set
	NV_LOCKOBJECT(this);
	map<string,void *>::iterator pos;
	time_t ActualTime = time(NULL);
	for(pos = m_mTasks.begin(); pos != m_mTasks.end(); ++pos)
	{
		NvSchedulerTask *Task = (NvSchedulerTask *)(pos->second);
		if(Task == NULL)
			continue;	// bad task

		NV_LOCKLOCK(&(Task->ProtectInUse));

		GetMain()->Logs.Write("Scheduler", "Check processing", Task->m_sName.c_str(), NV_LOG_INFO, 3, m_sAppName.c_str(), "system");

		if(!Task->m_bImmediateRun && !Task->m_bEnable)
		{
			GetMain()->Logs.Write("Scheduler", "Check failed - Task is disabled", Task->m_sName.c_str(), NV_LOG_INFO, 3, m_sAppName.c_str(), "system");
			NV_UNLOCKLOCK(&(Task->ProtectInUse));
			continue;	// disabled
		}

		if(!Task->m_sProcTask.length())
		{
			GetMain()->Logs.Write("Scheduler", "Check failed - Task has no procedure", Task->m_sName.c_str(), NV_LOG_INFO, 3, m_sAppName.c_str(), "system");
			Task->m_bImmediateRun=false;
			NV_UNLOCKLOCK(&(Task->ProtectInUse));
			continue;	// no task procedure
		}

		if(Task->m_bInQueue)
		{
			GetMain()->Logs.Write("Scheduler", "Check failed - Task ever in queue", Task->m_sName.c_str(), NV_LOG_INFO, 3, m_sAppName.c_str(), "system");
			Task->m_bImmediateRun=false;
			NV_UNLOCKLOCK(&(Task->ProtectInUse));
			continue;	// ever in scheduler queue
		}

		if(Task->m_iStatus == NV_SCTASK_STATUS_INQUEUE)
		{
			GetMain()->Logs.Write("Scheduler", "Check failed - Task status is in queue", Task->m_sName.c_str(), NV_LOG_INFO, 3, m_sAppName.c_str(), "system");
			Task->m_bImmediateRun=false;
			NV_UNLOCKLOCK(&(Task->ProtectInUse));
			continue;	// ever in scheduler queue
		}

		if(Task->m_iStatus == NV_SCTASK_STATUS_RUNNING)
		{
			GetMain()->Logs.Write("Scheduler", "Check failed - Task ever running", Task->m_sName.c_str(), NV_LOG_INFO, 3, m_sAppName.c_str(), "system");
			Task->m_bImmediateRun=false;
			NV_UNLOCKLOCK(&(Task->ProtectInUse));
			continue;	// ever running
		}

		// Lets check if we are in the time range
		// The function also retrieves the actual date at 00h00 (can be the day before)
		if(!Task->m_bImmediateRun)
		{
			time_t ActualDate;
			if(!Task->CheckTimeRange(&ActualDate))
			{
				// Not ready
				GetMain()->Logs.Write("Scheduler", "Check failed - Task outside time range", Task->m_sName.c_str(), NV_LOG_INFO, 3, m_sAppName.c_str(), "system");
				Task->ResetRepeatCounter();
				NV_UNLOCKLOCK(&(Task->ProtectInUse));
				continue; 
			}

			// Now, check the frequency
			// in order to know if we can run the task today
			if(!Task->CheckFrequency(ActualDate))
			{
				// Not ready
				GetMain()->Logs.Write("Scheduler", "Check failed - Task outside date range", Task->m_sName.c_str(), NV_LOG_INFO, 3, m_sAppName.c_str(), "system");
				Task->ResetRepeatCounter();
				NV_UNLOCKLOCK(&(Task->ProtectInUse));
				continue; 
			}

			// Finaly, check the repeat mode
			// in order to know if the task has ever been ran today
			// and if we can repeat it
			int RepeatRet = Task->CheckRepeat(ActualDate);
			if(RepeatRet)
			{
				// Not ready
				if(RepeatRet == 1)
					GetMain()->Logs.Write("Scheduler", "Check failed - Max number of occurences reached", Task->m_sName.c_str(), NV_LOG_INFO, 3, m_sAppName.c_str(), "system");
				else if(RepeatRet == 2)
					GetMain()->Logs.Write("Scheduler", "Check failed - Repeat delay not reached", Task->m_sName.c_str(), NV_LOG_INFO, 3, m_sAppName.c_str(), "system");
				else // Should not occur
					GetMain()->Logs.Write("Scheduler", "Check failed - Task already ran in time range", Task->m_sName.c_str(), NV_LOG_INFO, 3, m_sAppName.c_str(), "system");
				NV_UNLOCKLOCK(&(Task->ProtectInUse));
				continue; 
			}
		}

		// The task seems a good candidate
		// So we take it
		Task->m_bInQueue = true;
		int OldStatus = Task->m_iStatus;
		Task->m_iStatus = NV_SCTASK_STATUS_INQUEUE;
		GetMain()->Logs.Write("Scheduler", "Turn status to in queue", Task->m_sName.c_str(), NV_LOG_INFO, 2, m_sAppName.c_str(), "system");
		
		// Save the new task status
		if(!Task->Save())
		{
			Task->m_iStatus = NV_SCTASK_STATUS_FAILED;
			Task->Error.SetError("SYSTEM", "SCHEDULER", NV_SCHEDULER_EXCEPTION_IN_TASK, "Cannot save task to registry");
			GetMain()->Logs.Write("Scheduler", "Turn status to failed", Task->m_sName.c_str(), NV_LOG_INFO, 2, m_sAppName.c_str(), "system");
			Task->m_bInQueue = false;
			Task->m_bImmediateRun=false;
			NV_UNLOCKLOCK(&(Task->ProtectInUse));
			continue;
		}

		Task->m_bImmediateRun=false;
		TasksToProcess->Add(Task->m_sName);
		NV_UNLOCKLOCK(&(Task->ProtectInUse));
	}
	NV_UNLOCKOBJECT(this);

  return true;

}

bool NvSchedulerTaskList::VerifyTasksInQueue(string& AppName, NvSchedulerServer *Server)
{
	// Load task list if not yet done
	Load();

	// Get the tasks with start time in the past and that are not in queue
	// We also get tasks having the immediate run flag set
	NV_LOCKOBJECT(this);
	map<string,void *>::iterator pos;
	time_t ActualTime = time(NULL);
	for(pos = m_mTasks.begin(); pos != m_mTasks.end(); ++pos)
	{
		NvSchedulerTask *Task = (NvSchedulerTask *)(pos->second);
		if(Task == NULL)
			continue;	// bad task

		NV_LOCKLOCK(&(Task->ProtectInUse));

		if(Task->m_bInQueue || Task->m_iStatus == NV_SCTASK_STATUS_INQUEUE)
		{
			// Check if the task is really in the queue
			string tIdent = AppName;
			tIdent += ";";
			tIdent += Task->m_sName;
			MakeUpper(tIdent);
			if(!Server->IsInProcess(tIdent))
			{
				list<string>::iterator pos;
				bool bFound = false;
				for(pos = Server->m_lTasksToProcess.begin(); pos != Server->m_lTasksToProcess.end(); ++pos)
				{
					if(*pos == tIdent)
					{
						bFound = true;
						break;
					}
				}
				if(!bFound)
				{
					// the task was marked in queue but was not in the queue list
					// We set it to a NV_SCTASK_STATUS_NOT_STARTED status
					Task->m_bInQueue = false;
					Task->m_bImmediateRun=false;
					Task->m_iStatus = NV_SCTASK_STATUS_INQUEUE;
					Task->Save();
					GetMain()->Logs.Write("Scheduler", "Task removed from queue", Task->m_sName.c_str(), NV_LOG_INFO, 1, AppName.c_str(), "system");
				}
			}
		}
		NV_UNLOCKLOCK(&(Task->ProtectInUse));
	}
	NV_UNLOCKOBJECT(this);

  return true;

}


/////////////////////////////////////////////////////////////////////////////
// NvSchedulerTask

NvSchedulerTask::NvSchedulerTask()
{
	Session = NULL;
	InUse = 0;
	Error.SetError();
	m_iStatus = NV_SCTASK_STATUS_NOT_STARTED;
	m_bInQueue = false;
	m_bImmediateRun = false;
	m_sName = "";
	m_sDesc = "";
	m_sAppName = ""; 
	m_sUser = ""; 
	m_sLanguage = ""; 
	m_sProcTask = ""; 
	m_sProcFailed = ""; 
	m_sProcSuccess = "";
	m_sProcInit = "";
	m_sProcExit = "";
	tLastStartTime = 0;
	tLastEndTime = 0;
	m_iTimeSpan = 3600;
	m_bEnable = true;
	tStartDate = time(NULL);
	tStartTime = time(NULL);
	m_iFrequencyType = NV_SCTASK_FREQUENCY_ONETIME;
	m_bRepeatMode = false;
	m_iDelay = 60;
	m_iOccurences = 1;
	m_iActualOccurence = 0;
	tFirstOccurenceStartDate = 0;
	m_iDayEvery = 1;
	m_iWeekEvery = 1;
	m_iWeekDays = NV_SCTASK_WDAY_MON;
	m_iMonthMonths = NV_SCTASK_MONTH_ALL;
	m_bMonthDayBased = true;
	m_iMonthDays = NV_SCTASK_MDAY_01;
	m_iMonthWeekDayPosition = NV_SCTASK_WEEK_FIRST;
	m_iMonthWeekDay = 1;
}

NvSchedulerTask::~NvSchedulerTask()
{
}


bool NvSchedulerTask::IncInuse()
{
	NV_LOCKLOCK(&ProtectInUse);
	InUse++;
	NV_UNLOCKLOCK(&ProtectInUse);

	return true;
}

void NvSchedulerTask::DecInUse()
{
	NV_LOCKLOCK(&ProtectInUse);
	if(InUse)
		InUse--;
	NV_UNLOCKLOCK(&ProtectInUse);
}

int NvSchedulerTask::GetInUse()
{
	int RetVal = 0;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = InUse;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}



string NvSchedulerTask::GetName()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sName;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvSchedulerTask::GetAppName()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sAppName;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}

void NvSchedulerTask::SetInQueue(bool InQueue)
{
	NV_LOCKLOCK(&ProtectInUse);
	m_bInQueue = InQueue;
	NV_UNLOCKLOCK(&ProtectInUse);
}

bool NvSchedulerTask::GetInQueue()
{
	bool RetVal = false;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_bInQueue;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


// Saves the task parameters to the application registry
// Not locked. Must be externaly locked
bool NvSchedulerTask::Save()
{
	if(!m_sAppName.length())
	{
		GetMain()->Logs.Write("Scheduler", "Error saving task to registry", m_sName.c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");
		return false;
	}

	NvContainer Container;
	string Md5Name = m_sName;
	CleanString(Md5Name);
	Md5Name = MD5_string(Md5Name);
	MakeLower(Md5Name);
	Container.CreateObject(Md5Name.c_str(), NV_OBJTYPE_INDSTRINGLIST);
	NvObjIndStringList *TaskObj = (NvObjIndStringList *)(Container.GetObject(Md5Name.c_str()));
	if(TaskObj == NULL)
	{
		GetMain()->Logs.Write("Scheduler", "Error saving task to registry", m_sName.c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");
		return false;	// Should not occur
	}

	
	// Set task parameters

	// Set general parameters
	TaskObj->SetValue("Name", m_sName);
	TaskObj->SetValue("Desc", m_sDesc);
	char Buf[100];
	sprintf(Buf, "%d", m_iStatus);
	TaskObj->SetValue("Status", Buf);
	TaskObj->SetValue("User", m_sUser);
	TaskObj->SetValue("Language", m_sLanguage);
	TaskObj->SetValue("ProcTask", m_sProcTask);
	TaskObj->SetValue("ProcFailed", m_sProcFailed);
	TaskObj->SetValue("ProcSuccess", m_sProcSuccess);
	TaskObj->SetValue("ProcInit", m_sProcInit);
	TaskObj->SetValue("ProcExit", m_sProcExit);
	sprintf(Buf, "%d", Error.m_iErrorCode);
	TaskObj->SetValue("ErrorCode", Buf);
	TaskObj->SetValue("ErrorService", Error.m_sErrorService);
	TaskObj->SetValue("ErrorClass", Error.m_sErrorClass);
	TaskObj->SetValue("ErrorInfo", Error.m_sErrorInfo);
	sprintf(Buf, "%ld", (unsigned long)tLastStartTime);
	TaskObj->SetValue("LastStartTime", Buf);
	sprintf(Buf, "%ld", (unsigned long)tLastEndTime);
	TaskObj->SetValue("LastEndTime", Buf);
	if(m_bEnable)
		TaskObj->SetValue("Enable", "1");
	else
		TaskObj->SetValue("Enable", "0");
	sprintf(Buf, "%ld", (unsigned long)tStartDate);
	TaskObj->SetValue("StartDate", Buf);
	sprintf(Buf, "%ld", (unsigned long)tStartTime);
	TaskObj->SetValue("StartTime", Buf);
	sprintf(Buf, "%d", m_iTimeSpan);
	TaskObj->SetValue("TimeSpan", Buf);
	sprintf(Buf, "%d", m_iFrequencyType);
	TaskObj->SetValue("FrequencyType", Buf);
	
	
	// Set daily repeat mode parameters
	if(m_bRepeatMode)
		TaskObj->SetValue("RepeatMode", "1");
	else
		TaskObj->SetValue("RepeatMode", "0");
	sprintf(Buf, "%d", m_iDelay);
	TaskObj->SetValue("Delay", Buf);
	sprintf(Buf, "%d", m_iOccurences);
	TaskObj->SetValue("Occurences", Buf);
	sprintf(Buf, "%d", m_iActualOccurence);
	TaskObj->SetValue("ActualOccurence", Buf);
	sprintf(Buf, "%ld", (unsigned long)tFirstOccurenceStartDate);
	TaskObj->SetValue("FirstOccurenceStartDate", Buf);


	// Set daily frequency parameters
	sprintf(Buf, "%d", m_iDayEvery);
	TaskObj->SetValue("DayEvery", Buf);


	// Set weekly frequency parameters
	sprintf(Buf, "%d", m_iWeekEvery);
	TaskObj->SetValue("WeekEvery", Buf);
	sprintf(Buf, "%d", m_iWeekDays);
	TaskObj->SetValue("WeekDays", Buf);

	// Set monthly frequency parameters
	sprintf(Buf, "%d", m_iMonthMonths);
	TaskObj->SetValue("MonthMonths", Buf);
	if(m_bMonthDayBased)
		TaskObj->SetValue("MonthDayBased", "1");
	else
		TaskObj->SetValue("MonthDayBased", "0");
	sprintf(Buf, "%d", m_iMonthDays);
	TaskObj->SetValue("MonthDays", Buf);
	sprintf(Buf, "%d", m_iMonthWeekDayPosition);
	TaskObj->SetValue("MonthWeekDayPosition", Buf);
	sprintf(Buf, "%d", m_iMonthWeekDay);
	TaskObj->SetValue("MonthWeekDay", Buf);



	NvApplication *Application = GetMain()->Applications.GetApplication(m_sAppName.c_str());
	if(Application == NULL)
	{
		GetMain()->Logs.Write("Scheduler", "Error saving task to registry", m_sName.c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");
		return false;
	}


	NvRegistryContainer *Registry;
	Registry = &(Application->AppRegistry);

	// Save the task
	if(!Registry->SetKey("system.tasks", &Container))
	{
		GetMain()->Applications.FreeApplication(m_sAppName.c_str());
		GetMain()->Logs.Write("Scheduler", "Error saving task to registry", m_sName.c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");
		return false;
	}
	GetMain()->Applications.FreeApplication(m_sAppName.c_str());
	return true;
}

// Removes the task information from registry
// Not locked. Must be externaly locked
void NvSchedulerTask::Remove()
{
	string Md5Name = m_sName;
	CleanString(Md5Name);
	Md5Name = MD5_string(Md5Name);
	MakeLower(Md5Name);
	
	NvApplication *Application = GetMain()->Applications.GetApplication(m_sAppName.c_str());
	if(Application == NULL)
		return;


	NvRegistryContainer *Registry;
	Registry = &(Application->AppRegistry);

	// Remove the task
	Registry->RemoveKey("system.tasks", Md5Name.c_str());
	GetMain()->Applications.FreeApplication(m_sAppName.c_str());
}







// Check that we are in the window time range
// Doesn't check the day itself (made in CheckFrequency function)
// The function also retrieves the actual date at 00h00 (can be the day before)
bool NvSchedulerTask::CheckTimeRange(time_t *ActualDate)
{
	// Check if the task is in time window range

	int LogLevel = GetMain()->Logs.GetLevel("Scheduler");

	struct tm tTime1;
	localtime_r(&tStartTime, &tTime1);
	int iStartTimeWindow = (tTime1.tm_hour * 3600) + (tTime1.tm_min * 60) + tTime1.tm_sec;
	int iStartEndWindow = iStartTimeWindow + m_iTimeSpan;
	if(LogLevel == 6)
	{
		string LogMess = "NvSchedulerTask::CheckTimeRange - Start time=";
		char Buff[100];
		sprintf(Buff, "%d:%d:%d", tTime1.tm_hour, tTime1.tm_min, tTime1.tm_sec);
		LogMess += Buff;
		LogMess += " - StartTimeWindow=";
		sprintf(Buff, "%d", iStartTimeWindow);
		LogMess += Buff;
		LogMess += " - TimeSpan=";
		sprintf(Buff, "%d", m_iTimeSpan);
		LogMess += Buff;
		LogMess += " - StartEndWindow=";
		sprintf(Buff, "%d", iStartEndWindow);
		LogMess += Buff;
		GetMain()->Logs.Write("Scheduler", LogMess.c_str(), m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");
	}
	iStartEndWindow = iStartEndWindow % (24*60*60);
	if(LogLevel == 6)
	{
		string LogMess = "NvSchedulerTask::CheckTimeRange - Adjusted StartEndWindow=";
		char Buff[100];
		sprintf(Buff, "%d", iStartEndWindow);
		LogMess += Buff;
		GetMain()->Logs.Write("Scheduler", LogMess.c_str(), m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");
	}
	bool bMidnigthInWindow = false;
	if(iStartEndWindow < iStartTimeWindow)
		bMidnigthInWindow = true;

	if(LogLevel == 6)
	{
		if(bMidnigthInWindow)
			GetMain()->Logs.Write("Scheduler", "Midnight in window: yes", m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");
		else
			GetMain()->Logs.Write("Scheduler", "Midnight in window: no", m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");
	}

	time_t tActualTime = time(NULL);
	localtime_r(&tActualTime, &tTime1);
	int iActualTime = (tTime1.tm_hour * 3600) + (tTime1.tm_min * 60) + tTime1.tm_sec;
	if(bMidnigthInWindow)
	{
		if(iActualTime < iStartTimeWindow)
		{
			if(iActualTime > iStartEndWindow)
			{
				if(LogLevel == 6)
					GetMain()->Logs.Write("Scheduler", "Not in time range (MidnigthInWindow)", m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");
				return false;
			}
		}
	}
	else
	{
		if(iActualTime < iStartTimeWindow)
		{
			if(LogLevel == 6)
				GetMain()->Logs.Write("Scheduler", "Not in time range (iActualTime < iStartTimeWindow)", m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");
			return false;
		}
		if(iActualTime > iStartEndWindow)
		{
			if(LogLevel == 6)
				GetMain()->Logs.Write("Scheduler", "Not in time range (iActualTime > iStartEndWindow)", m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");
			return false;
		}
	}

	if(LogLevel == 6)
		GetMain()->Logs.Write("Scheduler", "In time range", m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");

	// OK, we are in the time range

	// We will now check the day
	// So we get the actual day
	// the actual day is the one at the start
	// of the window time range so its the actual day
	// or the day before
	bool bDayBefore = false;
	if(bMidnigthInWindow)
		if(iActualTime < iStartEndWindow)
			bDayBefore = true;

	if(bDayBefore)
		GetMain()->Logs.Write("Scheduler", "Day before: yes", m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");
	else
		GetMain()->Logs.Write("Scheduler", "Day before: no", m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");

	if(bDayBefore)
		tActualTime -= (24*60*60);

	// Check the start date
	localtime_r(&tActualTime, &tTime1);
	int ActualYear = tTime1.tm_year;
	int ActualMonth = tTime1.tm_mon;
	int ActualDay = tTime1.tm_mday;
	localtime_r(&tStartDate, &tTime1);
	int StartYear = tTime1.tm_year;
	int StartMonth = tTime1.tm_mon;
	int StartDay = tTime1.tm_mday;
	if(LogLevel == 6)
	{
		string LogMess = "NvSchedulerTask::CheckTimeRange - ActualDate=";
		char Buff[100];
		sprintf(Buff, "%d:%d:%d", ActualYear+1900, ActualMonth+1, ActualDay);
		LogMess += Buff;
		LogMess += " - StartDate=";
		sprintf(Buff, "%d:%d:%d", StartYear+1900, StartMonth+1, StartDay);
		LogMess += Buff;
		GetMain()->Logs.Write("Scheduler", LogMess.c_str(), m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");
	}
	if(StartYear > ActualYear)
	{
		if(LogLevel == 6)
			GetMain()->Logs.Write("Scheduler", "Not in time range (StartYear > ActualYear)", m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");
		return false;
	}
	if(StartYear == ActualYear)
	{
		if(StartMonth > ActualMonth)
		{
			if(LogLevel == 6)
				GetMain()->Logs.Write("Scheduler", "Not in time range (StartMonth > ActualMonth)", m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");
			return false;
		}
		if(StartMonth == ActualMonth)
		{
			if(StartDay > ActualDay)
			{
				if(LogLevel == 6)
					GetMain()->Logs.Write("Scheduler", "Not in time range (StartDay > ActualDay)", m_sName.c_str(), NV_LOG_INFO, 6, m_sAppName.c_str(), "system");
				return false;
			}
		}
	}

	time_t tActualDateHour0 = tActualTime;
	localtime_r(&tActualTime, &tTime1);
	tActualDateHour0 -= (tTime1.tm_hour * 3600) + (tTime1.tm_min * 60) + tTime1.tm_sec;


	*ActualDate = tActualDateHour0;

	// Ok, we are in the time range
	// And the start date is not a future date
	return true;
}



// Check that we can run the task today following the frequency mode
bool NvSchedulerTask::CheckFrequency(time_t ActualDate)
{

	time_t tActualDateHour0 = ActualDate;
	time_t tStartDateHour0 = tStartDate;
	struct tm tTime1;
	localtime_r(&tStartDate, &tTime1);
	tStartDateHour0 -= (tTime1.tm_hour * 3600) + (tTime1.tm_min * 60) + tTime1.tm_sec;


	// Now we can check following the frequency type
	if(m_iFrequencyType == NV_SCTASK_FREQUENCY_ONETIME)
	{
		// We must be the required day
		if(tStartDateHour0 != tActualDateHour0)
			return false;

		// Ok for this frequency type
		return true;
	}

	if(m_iFrequencyType == NV_SCTASK_FREQUENCY_DAILY)
	{
		// We must check the m_iDayEvery parameter
		// starting from the start date
		int NumDays = (int)(tActualDateHour0 - tStartDateHour0);
		NumDays = NumDays / (24*60*60);
		if((NumDays % m_iDayEvery) != 0)
			return false;

		// Ok for this frequency type
		return true;
	}

	if(m_iFrequencyType == NV_SCTASK_FREQUENCY_WEEKLY)
	{
		// We must check the m_iWeekEvery parameter
		// starting from the start date
		time_t tStartDateHour0Monday = tStartDateHour0;
		localtime_r(&tStartDateHour0Monday, &tTime1);
		int DayOfWeek = tTime1.tm_wday;
		if(DayOfWeek == 0)
			DayOfWeek = 7;
		DayOfWeek--;	//0=monday, 6=sunday
		tStartDateHour0Monday -= (DayOfWeek * 24 * 60 * 60);

		time_t tActualDateHour0Monday = tActualDateHour0;
		localtime_r(&tActualDateHour0Monday, &tTime1);
		DayOfWeek = tTime1.tm_wday;
		if(DayOfWeek == 0)
			DayOfWeek = 7;
		DayOfWeek--;	//0=monday, 6=sunday
		tActualDateHour0Monday -= (DayOfWeek * 24 * 60 * 60);

		int NumWeeks = (int)(tActualDateHour0Monday - tStartDateHour0Monday);
		NumWeeks = NumWeeks / (7*24*60*60);

		if((NumWeeks % m_iWeekEvery) != 0)
			return false;	// bad week

		// We check now the week day
		DayOfWeek = 1 << DayOfWeek;
		if(!(m_iWeekDays & DayOfWeek))
			return false;	// bad week day

		// Ok for this frequency type
		return true;
	}

	if(m_iFrequencyType == NV_SCTASK_FREQUENCY_MONTHLY)
	{
		// We must check the m_iMonthMonths parameter
		localtime_r(&tActualDateHour0, &tTime1);
		int MonthOfYear = tTime1.tm_mon;
//		MonthOfYear--;
		MonthOfYear = 1 << MonthOfYear;
		if(!(m_iMonthMonths & MonthOfYear))
			return false;	// bad month

		if(m_bMonthDayBased)
		{
			// Month day based
			int DayOfMonth = tTime1.tm_mday;
			DayOfMonth--;
			DayOfMonth = 1 << DayOfMonth;
			if(!(m_iMonthDays & DayOfMonth))
				return false;	// bad month
		}
		else
		{
			int DayOfWeek = tTime1.tm_wday;
			if(DayOfWeek == 0)
				DayOfWeek = 7;

			int DayOfWeekPosition = tTime1.tm_mday;
			DayOfWeekPosition = (DayOfWeekPosition+6)/7;

			int MonthWeekDayPosition = m_iMonthWeekDayPosition;
			if(tTime1.tm_mon == 1)	// February (special because max weeks is generally 4
			{
				// if not bisextil year february has 4 weeks;
				if(((tTime1.tm_year + 1900) % 4) != 0)
				{
					if(MonthWeekDayPosition == NV_SCTASK_WEEK_LAST)
						MonthWeekDayPosition = NV_SCTASK_WEEK_FOURTH;
				}
			}
			if(DayOfWeek != m_iMonthWeekDay)
				return false;	// bad week day
			if(MonthWeekDayPosition != DayOfWeekPosition)
				return false;	// bad week position

		}

		// Ok for this frequency type
		return true;
	}

	return true;
}



// Check the repeat mode
// in order to know if the task has ever been ran today
// and if we can repeat it
// it increases the occurence counter by one if OK
// ret values:
// 0 no repeat, authorized to run
// 1 max number of occurences reached
// 2 repeat delay not reached
int NvSchedulerTask::CheckRepeat(time_t ActualDate)
{
	if(!m_iActualOccurence)
	{
		if(tFirstOccurenceStartDate != ActualDate)
			tFirstOccurenceStartDate = ActualDate;
		m_iActualOccurence++;
		return 0;
	}

	// We have an occurence
	// We verify that it's the same day
	if(tFirstOccurenceStartDate != ActualDate)
	{
		// not same day
		// so we reset the occurence counter to 1
		m_iActualOccurence = 1;
		tFirstOccurenceStartDate = ActualDate;
		return 0;
	}

	// Ok, now, we ever have an occurence and it's the same day
	// Check if we have reach the maximum number of occurences
	int MaxOccurences = 1;
	if(m_bRepeatMode)
		MaxOccurences = m_iOccurences;

	if((MaxOccurences != 0) && (m_iActualOccurence >= MaxOccurences))
	{
		// We have reach the maximum number of occurences
		// We do not change anything because the user has the possibility to
		// change the max number of occurences while in the time range window
		return 1;
	}

	// OK, we can eventually run a new occurence
	// But we have first to test the m_iDelay parameter

	time_t ActualTime = time(NULL);
	time_t MinimumNextStartTime = tLastEndTime + m_iDelay;
	if(MinimumNextStartTime > ActualTime)
	{
		// Not ready for the next occurence
		// We do not change anything because the user has the possibility to
		// change the delay between occurences while in the time range window
		return 2;
	}

	// Ok for this occurence
	m_iActualOccurence++;
	return 0;
}


// Resets the tFirstOccurenceStartDate and m_iActualOccurence if necessary
// Saves the task if there is some change
void NvSchedulerTask::ResetRepeatCounter()
{
	bool bChanged = false;
	if(m_iActualOccurence > 0)
	{
		m_iActualOccurence = 0;
		bChanged = true;
	}
	if(tFirstOccurenceStartDate > 0)
	{
		tFirstOccurenceStartDate = 0;
		bChanged = true;
	}

	if(bChanged)
		Save();
}


// Run the task
bool NvSchedulerTask::Run()
{
	// Get necessary parameters from the task
	NV_LOCKLOCK(&ProtectInUse);

	string PasswordHere ="";
	string UserName = m_sUser;
	string AppName = m_sAppName;
	string Language = m_sLanguage;
	string ProcTask = m_sProcTask;
	string ProcFailed = m_sProcFailed;
	string ProcSuccess = m_sProcSuccess;
	string ProcInit = m_sProcInit;
	string ProcExit = m_sProcExit;

	// Also set the start time
	tLastStartTime = time(NULL);

	// And the status
	m_iStatus = NV_SCTASK_STATUS_RUNNING;
	GetMain()->Logs.Write("Scheduler", "Turn status to running", m_sName.c_str(), NV_LOG_INFO, 2, m_sAppName.c_str(), "system");

	// The error is reset
	Error.SetError();

	// Save the task
	Save();
	NV_UNLOCKLOCK(&ProtectInUse);

	// We must create a new session
	if(GetMain()->CheckOverTime())
	{
		NV_LOCKLOCK(&ProtectInUse);
		Error.SetError("SYSTEM", "SYSTEM", NV_SYSTEM_OVER_TIME);
		m_iStatus = NV_SCTASK_STATUS_FAILED;
		GetMain()->Logs.Write("Scheduler", "Your license does not allow Nirva to run at this time", m_sName.c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");
		GetMain()->Logs.Write("Scheduler", "Turn status to failed", m_sName.c_str(), NV_LOG_INFO, 2, m_sAppName.c_str(), "system");
		tLastEndTime = time(NULL);
		Save();
		NV_UNLOCKLOCK(&ProtectInUse);
		return false;
	}
	Session = GetMain()->Sessions.GetSession(NULL);
	if(Session == NULL)
	{
		// Error creating the session
		NV_LOCKLOCK(&ProtectInUse);
		Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOTCREATE);
		int MaxSessions = GetMain()->GetMaxSessions();
		if(MaxSessions != -1)
		{
			if(GetMain()->Sessions.GetNumSessions() >= MaxSessions)
				Error.SetError("SYSTEM", "SYSTEM", NV_SYSTEM_MAX_SESSIONS);
		}
		m_iStatus = NV_SCTASK_STATUS_FAILED;
		GetMain()->Logs.Write("Scheduler", "Task failed - Max sessions reached", m_sName.c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");
		GetMain()->Logs.Write("Scheduler", "Turn status to failed", m_sName.c_str(), NV_LOG_INFO, 2, m_sAppName.c_str(), "system");
		tLastEndTime = time(NULL);
		Save();
		NV_UNLOCKLOCK(&ProtectInUse);
		return false;
	}

	// m_bFromTransaction is used to discard password control
	// when login the application
	Session->m_bFromTransaction = true;
	Session->m_sOpenProc = ProcInit;
	Session->m_sCloseProc = ProcExit;

	GetMain()->AttachRequestList.Get(Session, ATTACH_REQUEST_TYPE_SCHEDULER, AppName);
	if(Session->Login(UserName.c_str(), PasswordHere.c_str(), AppName.c_str(), SESSION_TYPE_SCHEDULER) != 1)
	{
		// Login failed
		// We prepare error information
		string Info = UserName;
		Info += " - ";
		Info += AppName;
		NV_LOCKLOCK(&ProtectInUse);
		Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOTLOGIN, Info);
		m_iStatus = NV_SCTASK_STATUS_FAILED;
		GetMain()->Logs.Write("Scheduler", "Task failed - Cannot login", m_sName.c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");
		GetMain()->Logs.Write("Scheduler", "Turn status to failed", m_sName.c_str(), NV_LOG_INFO, 2, m_sAppName.c_str(), "system");
		tLastEndTime = time(NULL);
		Save();
		NV_UNLOCKLOCK(&ProtectInUse);

		// We free the session
		GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

		// And remove it
		GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
		Session = NULL;
		return false;
	}

	string UserHereForCheck = UserName;
	CleanString(UserHereForCheck, true);
	MakeLower(UserHereForCheck);
	if(UserHereForCheck == "nvadmin")
		Session->m_bAdministrator = true; // Administrator user

	// Now we have the session

	// We prepare a command
	NvCommand Command;
	Command.m_iCommandSource = NV_SOURCE_SERVER;
	Command.m_sService = "SYSTEM";
	Command.m_sClass = "MISC";
	Command.m_sCommand = "NOP";
	Command.m_sInContainer = NV_CONTAINER_DEFAULT_NAME;
	Command.m_sOutContainer = NV_CONTAINER_DEFAULT_NAME;
	Command.m_sLanguage = Language;
	Command.Parameters.SetValue("NV_PROC", ProcTask);
	
	
	// Run the task here
	bool SessionRetVal = true;
	bool bExceptionOccured = false;
	try
	{
		SessionRetVal = Session->Command(&Command);
	}
	catch(...)
	{
		GetMain()->IncNumExceptions();
		GetMain()->Logs.Write("Exceptions", "Exception when running scheduler procedure", ProcTask.c_str(), NV_LOG_ERROR);
		string Message = Command.m_sService;
		Message += ":";
		Message += Command.m_sClass;
		Message += ":";
		Message += Command.m_sCommand;
		Command.Error.SetError("SYSTEM", "COMMAND", NV_COMMAND_EXCEPTION, Message);
		GetMain()->Logs.Write("Scheduler", "Task failed - exception occured", GetName().c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");
		SessionRetVal = false;
		bExceptionOccured = true;
	}
	if(!SessionRetVal)
	{
		// Error in task procedure
		NV_LOCKLOCK(&ProtectInUse);
		Error.SetError(Command.Error.m_sErrorService, Command.Error.m_sErrorClass, Command.Error.m_iErrorCode, Command.Error.m_sErrorInfo);
		m_iStatus = NV_SCTASK_STATUS_FAILED;
		if(!bExceptionOccured)
			GetMain()->Logs.Write("Scheduler", "Task failed - task procedure failed", m_sName.c_str(), NV_LOG_INFO, 1, m_sAppName.c_str(), "system");
		GetMain()->Logs.Write("Scheduler", "Turn status to failed",  m_sName.c_str(), NV_LOG_INFO, 2, m_sAppName.c_str(), "system");
		tLastEndTime = time(NULL);
		Save();
		NV_UNLOCKLOCK(&ProtectInUse);
		if(ProcFailed.length())
		{
			NvCommand ErrorCommand;
			ErrorCommand.m_iCommandSource = NV_SOURCE_SERVER;
			ErrorCommand.m_sService = "SYSTEM";
			ErrorCommand.m_sClass = "MISC";
			ErrorCommand.m_sCommand = "NOP";
			ErrorCommand.m_sInContainer = NV_CONTAINER_DEFAULT_NAME;
			ErrorCommand.m_sOutContainer = NV_CONTAINER_DEFAULT_NAME;
			ErrorCommand.m_sLanguage = Language;
			ErrorCommand.Parameters.SetValue("NV_PROC", ProcFailed);
			char Buf[20];
			sprintf(Buf, "%d", Command.Error.m_iErrorCode);
			ErrorCommand.Parameters.SetValue("NV_ERROR_CODE", Buf);
			ErrorCommand.Parameters.SetValue("NV_ERROR_SERVICE", Command.Error.m_sErrorService.c_str());
			ErrorCommand.Parameters.SetValue("NV_ERROR_CLASS", Command.Error.m_sErrorClass.c_str());
			ErrorCommand.Parameters.SetValue("NV_ERROR_DESC", Command.Error.GetErrorDesc(Language.c_str()).c_str());
			ErrorCommand.Parameters.SetValue("NV_ERROR_INFO", Command.Error.m_sErrorInfo.c_str());
			try
			{
				Session->Command(&ErrorCommand);
			}
			catch(...)
			{
				GetMain()->IncNumExceptions();
				GetMain()->Logs.Write("Exceptions", "Exception when running error scheduler procedure", ProcFailed.c_str(), NV_LOG_ERROR);
			}
		}

		GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());
		GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
		Session = NULL;
		return false;	
	}

	// The task procedure has been completed sucessfully
	NV_LOCKLOCK(&ProtectInUse);
	m_iStatus = NV_SCTASK_STATUS_SUCCESS;
	GetMain()->Logs.Write("Scheduler", "Turn status to success",  m_sName.c_str(), NV_LOG_INFO, 2, m_sAppName.c_str(), "system");
	tLastEndTime = time(NULL);
	Save();
	NV_UNLOCKLOCK(&ProtectInUse);

	// We remove the session a the end
	GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());
	GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
	Session = NULL;

	return true;
}





NvSchedulerServer::NvSchedulerServer()
{
	b_Initialized = false;
	m_iMaxTasks = 10;
	m_iCheckEvery = 10000;
}

// Scheduler server main thread function
int NvSchedulerServer::Main()
{
	time_t VerifyTasksInQueueInterval = time(NULL);
	while(1)
	{
		int MaxTasks = GetMaxTasks();
		int CheckEvery = GetCheckEvery() * 1000;
		if((time(NULL) - VerifyTasksInQueueInterval) > 60)
		{
			// Each minute we verify if the tasks are really in queue
			// This is because we got support case were the task was marked to be in queue but in fact
			// was not in queue (case occured after a problem on a virtual machine)
			VerifyTasksInQueue();
			VerifyTasksInQueueInterval = time(NULL);
		}
		GetTasksToProcess();
		while(m_lTasksToProcess.size())
		{
			if(AskToTerminate.IsSignaled())
				break;

			// Do nothing if max simultaneous task requests reached
			if(GetInProcessSize() >= MaxTasks)
			{
				SLEEP(CheckEvery);
				continue;
			}

			string Tid = m_lTasksToProcess.front();
			m_lTasksToProcess.pop_front();
			AddInProcess(Tid);
			
			NvSchedulerClient *Client = NVNEW NvSchedulerClient;
			Client->m_sTaskId = Tid;
			Client->Server = this;
			
			// we run the client thread
			if(!Client->Run(false, true))
			{
				RemoveInProcess(Tid);
			}
		}
		if(AskToTerminate.IsSignaled())
			break;
		SLEEP(CheckEvery);

	}
	return 0;
}


// called to stop the server
void NvSchedulerServer::Stop()
{
	// we just the AskToTerminate event to signaled
	AskToTerminate.Signal();
}


// called when an exception occurs (see class NvTask)
void NvSchedulerServer::OnException(string sInfo)
{
	GetMain()->IncNumExceptions();
	GetMain()->Logs.Write("Exceptions", "Exception in scheduler server", NULL, NV_LOG_ERROR);
	// We must signal the AskToTerminate event for all client threads
	AskToTerminate.Signal();
}


// called before the Main function (see class NvTask)
bool NvSchedulerServer::OnInit()
{
	NvTask::OnInit();	// call the base class
	
	if(!m_bSameThread)
		GetMain()->Threads.Add(THREAD_TYPE_SCHEDULER_SERVER);

	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Init thread", "Type SCHEDULER SERVER", NV_LOG_INFO, 1, sThreadId, "system");
	
	AskToTerminate.Reset();		// we reset the asktoterminate event in case it is set

	b_Initialized = true;
	// Implement your code here
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread started", "Type SCHEDULER SERVER", NV_LOG_INFO, 1, sThreadId, "system");
	return true;
}


// called after the Main function (see class NvTask)
void NvSchedulerServer::OnCleanup()
{
	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Cleanup thread", "Type SCHEDULER SERVER", NV_LOG_INFO, 1, sThreadId, "system");

	if(b_Initialized)
	{
		// we have to wait for all clients to close themselves

		while(1)
		{
			if(!GetInProcessSize())
				break;	// the client size is empty meaning that all the clients have beeen closed properly

			// let other threads do their work (in milliseconds)
			SLEEP(500);
		}
	}
	
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread stopped", "Type SCHEDULER SERVER", NV_LOG_INFO, 1, sThreadId, "system");
	if(!m_bSameThread)
		GetMain()->Threads.Remove();
	NvTask::OnCleanup();	// call the base class
}


// Look at the tasks to process in applications
// and populates the FIFO m_lTasks;
void NvSchedulerServer::GetTasksToProcess()
{

	// Do nothing if the list is greater than the max size
	if((int)(m_lTasksToProcess.size()) >= m_iMaxTasks)
		return;

	NvObjTable ApplicationInfo;
	GetMain()->Applications.GetInfo(NULL, &ApplicationInfo);
	int ColNameIndex = ApplicationInfo.GetColumnIndex("NAME");
	int i;
	for(i=0; i<ApplicationInfo.GetNumRow(); i++)
	{
		string AppName = ApplicationInfo.GetCellLine(i, ColNameIndex);
		CleanString(AppName);
		if(!GetMain()->Applications.IsApplicationEnable(AppName.c_str()))
			continue;
		NvApplication *Application = GetMain()->Applications.GetApplication(AppName.c_str());
		if(Application == NULL)
			continue;
		if(!Application->IsEnable())
		{
			GetMain()->Applications.FreeApplication(AppName.c_str());
			continue;
		}
		if(Application->IsInStartStop())
		{
			// Do not accept application in start or stop step
			GetMain()->Applications.FreeApplication(AppName.c_str());
			continue;
		}
		NvObjStringList TasksToProcess;
		Application->Tasks.GetTasksToProcess(&TasksToProcess);
		int j;
		for(j=0; j<TasksToProcess.GetSize(); j++)
		{
			string tIdent = AppName;
			tIdent += ";";
			tIdent += TasksToProcess.GetValue(j);
			MakeUpper(tIdent);
			m_lTasksToProcess.push_back(tIdent);
//			AddInProcess(tIdent);
		}
		GetMain()->Applications.FreeApplication(AppName.c_str());
	}

}


// Verify tasks in queue
void NvSchedulerServer::VerifyTasksInQueue()
{
	NvObjTable ApplicationInfo;
	GetMain()->Applications.GetInfo(NULL, &ApplicationInfo);
	int ColNameIndex = ApplicationInfo.GetColumnIndex("NAME");
	int i;
	for(i=0; i<ApplicationInfo.GetNumRow(); i++)
	{
		string AppName = ApplicationInfo.GetCellLine(i, ColNameIndex);
		CleanString(AppName);
		if(!GetMain()->Applications.IsApplicationEnable(AppName.c_str()))
			continue;
		NvApplication *Application = GetMain()->Applications.GetApplication(AppName.c_str());
		if(Application == NULL)
			continue;
		if(!Application->IsEnable())
		{
			GetMain()->Applications.FreeApplication(AppName.c_str());
			continue;
		}
		Application->Tasks.VerifyTasksInQueue(AppName, this);
		GetMain()->Applications.FreeApplication(AppName.c_str());
	}

}

bool NvSchedulerServer::IsInProcess(string tIdent)
{
	bool RetVal = false;
	MakeUpper(tIdent);
	NV_LOCKOBJECT(this);
	if(m_lTasksInProcess.find(tIdent) != m_lTasksInProcess.end())
		RetVal = true;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvSchedulerServer::AddInProcess(string tIdent)
{
	MakeUpper(tIdent);
	NV_LOCKOBJECT(this);
	m_lTasksInProcess.insert(tIdent);
	NV_UNLOCKOBJECT(this);
	return true;
}


bool NvSchedulerServer::RemoveInProcess(string tIdent)
{
	MakeUpper(tIdent);
	NV_LOCKOBJECT(this);
	m_lTasksInProcess.erase(tIdent);
	NV_UNLOCKOBJECT(this);
	return true;
}

int NvSchedulerServer::GetInProcessSize()
{
	int RetVal = 0;
	NV_LOCKOBJECT(this);
	RetVal = (int)(m_lTasksInProcess.size());
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

void NvSchedulerServer::SetMaxTasks(int MaxTasks)
{
	NV_LOCKOBJECT(this);
	m_iMaxTasks = MaxTasks;
	if(m_iMaxTasks < 1)
		m_iMaxTasks = 1;
	if(m_iMaxTasks > 500)
		m_iMaxTasks = 500;
	NV_UNLOCKOBJECT(this);
}

int NvSchedulerServer::GetMaxTasks()
{
	int RetVal = 0;
	NV_LOCKOBJECT(this);
	RetVal = m_iMaxTasks;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

void NvSchedulerServer::SetCheckEvery(int CheckEvery)
{
	NV_LOCKOBJECT(this);
	m_iCheckEvery = CheckEvery;
	if(m_iCheckEvery < 1)
		m_iCheckEvery = 1;
	NV_UNLOCKOBJECT(this);
}

int NvSchedulerServer::GetCheckEvery()
{
	int RetVal = 0;
	NV_LOCKOBJECT(this);
	RetVal = m_iCheckEvery;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}



NvSchedulerClient::NvSchedulerClient()
{
	b_Initialized = false;
	m_tTask = NULL;
	m_bTaskRunning = false;
}

// Client thread main function
int NvSchedulerClient::Main()
{
	// Just get and run the task
	RunTask();
	return 0;
}


// called when an exception occurs (see class NvTask)
void NvSchedulerClient::OnException(string sInfo)
{
	GetMain()->IncNumExceptions();
	if(m_tTask != NULL)
		GetMain()->Logs.Write("Exceptions", "Exception in scheduler task", m_tTask->m_sName.c_str(), NV_LOG_ERROR);
	else
		GetMain()->Logs.Write("Exceptions", "Exception in scheduler task", NULL, NV_LOG_ERROR);
//	throw this;
}


// called before the Main function (see class NvTask)
bool NvSchedulerClient::OnInit()
{
	NvTask::OnInit();	// call the base class
	
	if(!m_bSameThread)
		GetMain()->Threads.Add(THREAD_TYPE_SCHEDULER_TASK);

	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Init thread", "Type TASK", NV_LOG_INFO, 1, sThreadId, "system");

	// Initialize the random generator for this thread
#ifdef WIN32
	int ThreadHandle = GetCurrentThreadId();						// Thread handle for window
#else
	pthread_t ThreadHandle = pthread_self();					// Thread handle for unix
#endif

	unsigned int randinit = (unsigned int)ThreadHandle;
	randinit *= (unsigned int)time(NULL);
	srand(randinit);
	

	b_Initialized = true;
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread started", "Type TASK", NV_LOG_INFO, 1, sThreadId, "system");

	return true;
}


// called after the Main function (see class NvTask)
void NvSchedulerClient::OnCleanup()
{
	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Cleanup thread", "Type TASK", NV_LOG_INFO, 1, sThreadId, "system");

	if(b_Initialized)
	{
		// free the task if not made
		// and set it not in queue
		NvSession *Session = NULL;
		if(m_bTaskRunning)
		{
			// This is a crash condition
			if(m_tTask != NULL)
			{
				m_tTask->SetInQueue(false);
				NvApplication *Application = GetMain()->Applications.GetApplication(m_tTask->GetAppName().c_str());
				if(Application != NULL)
				{
					NV_LOCKLOCK(&(m_tTask->ProtectInUse));
					m_tTask->Error.SetError("SYSTEM", "SCHEDULER", NV_SCHEDULER_EXCEPTION_IN_TASK);
					m_tTask->m_iStatus = NV_SCTASK_STATUS_FAILED;
					GetMain()->Logs.Write("Scheduler", "Task failed - exception occured for unknown reason (cleanup)", m_tTask->m_sName.c_str(), NV_LOG_INFO, 1, m_tTask->m_sAppName.c_str(), "system");
					GetMain()->Logs.Write("Scheduler", "Turn status to failed", m_tTask->m_sName.c_str(), NV_LOG_INFO, 2, m_tTask->m_sAppName.c_str(), "system");
					m_tTask->tLastEndTime = time(NULL);
					m_tTask->Save();
					NV_UNLOCKLOCK(&(m_tTask->ProtectInUse));
					Application->Tasks.FreeTask(m_tTask->GetName().c_str());
					GetMain()->Applications.FreeApplication(m_tTask->GetAppName().c_str());
				}

				Session = m_tTask->Session;
			}
		}
	
		// Free the session if not made
		if(Session != NULL)
		{
			if(Session->ActualCommand != NULL)
			{
				// The session is in a command so it's probably a crash
				if(Session->ActualCommand->m_sService != "SYSTEM")
					GetMain()->Services.FreeService(Session->ActualCommand->m_sService.c_str());
				Session->ActualCommand = NULL;
				Session->OriginalCommand = NULL;
				Session->InContainer = NULL;
				Session->OutContainer = NULL;
				Session->SetActualCommand("");
				Session->SetOriginalCommand("");
				NV_UNLOCKOBJECT(Session);
			}

			GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

			// We always remove the session because it has been created only for the task duration
			GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
		}

		m_bTaskRunning = false;


		// Detach thread from java interface if not yet made
		if(!m_bSameThread)
			GetMain()->JavaInterface.DetachJavaThread();
	}

	// remove this client from the server client list
	Server->RemoveInProcess(m_sTaskId);

	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread stopped", "Type TASK", NV_LOG_INFO, 1, sThreadId, "system");
	if(!m_bSameThread)
		GetMain()->Threads.Remove();
	NvTask::OnCleanup();	// call the base class
}


bool NvSchedulerClient::RunTask()
{
	// First try to get the task
	string AppName = m_sTaskId;
	string::size_type index;
	index = AppName.find(';');
	if(index == string::npos)
		return false;	// should not occur
	string RealTaskId = AppName.substr(index+1);
	AppName = AppName.substr(0, index);
	CleanString(AppName);
	NvApplication *Application = GetMain()->Applications.GetApplication(AppName.c_str());
	if(Application == NULL)
		return false;	// maybe the application has been removed

	m_tTask = Application->Tasks.GetTask(RealTaskId.c_str());
	if(m_tTask == NULL)
	{
		// The task has probably been removed
		GetMain()->Applications.FreeApplication(AppName.c_str());
		return false;
	}

	GetMain()->Logs.Write("Scheduler", "Start task", m_tTask->GetName().c_str(), NV_LOG_INFO, 1, m_tTask->m_sAppName.c_str(), "system");
	m_bTaskRunning = true;
	m_tTask->Run();
	m_bTaskRunning = false;
	GetMain()->Logs.Write("Scheduler", "End task", m_tTask->GetName().c_str(), NV_LOG_INFO, 1, m_tTask->m_sAppName.c_str(), "system");

	m_tTask->SetInQueue(false);

	// In order to be sure that the task is properly closed
	// we set an error if status is not success or error
	NV_LOCKLOCK(&(m_tTask->ProtectInUse));
	if((m_tTask->m_iStatus != NV_SCTASK_STATUS_FAILED) && (m_tTask->m_iStatus != NV_SCTASK_STATUS_SUCCESS))
	{
		m_tTask->Error.SetError("SYSTEM", "SCHEDULER", NV_SCHEDULER_EXCEPTION_IN_TASK, "Unknown exception in task");
		m_tTask->m_iStatus = NV_SCTASK_STATUS_FAILED;
		GetMain()->Logs.Write("Scheduler", "Task failed - exception occured for unknown reason (RunTask)", m_tTask->m_sName.c_str(), NV_LOG_INFO, 1, m_tTask->m_sAppName.c_str(), "system");
		GetMain()->Logs.Write("Scheduler", "Turn status to failed", m_tTask->m_sName.c_str(), NV_LOG_INFO, 2, m_tTask->m_sAppName.c_str(), "system");
		m_tTask->tLastEndTime = time(NULL);
		m_tTask->Save();
	}
	NV_UNLOCKLOCK(&(m_tTask->ProtectInUse));

	Application->Tasks.FreeTask(RealTaskId.c_str());

	GetMain()->Applications.FreeApplication(AppName.c_str());
	m_tTask = NULL;

	return true;
}

