// ***************************************************************************************
// File: scheduler.h
// Description: scheduler classes
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifndef _SCHEDULER_H
#define _SCHEDULER_H


// Classes declared in this file
class NvSchedulerTask;
class NvSchedulerTaskList;
class NvSchedulerServer;
class NvSchedulerClient;

#include "task.h"
#include "util.h"
#include "container.h"
#include "session.h"
#include <string>
#include <list>
#include <set>


// task status
#define NV_SCTASK_STATUS_NOT_STARTED			0
#define NV_SCTASK_STATUS_INQUEUE				1
#define NV_SCTASK_STATUS_RUNNING				2
#define NV_SCTASK_STATUS_FAILED					3
#define NV_SCTASK_STATUS_SUCCESS				4
#define NV_SCTASK_STATUS_STOPPED				5
#define NV_SCTASK_MAXSTATUS						5

// task repetition frequency type
#define NV_SCTASK_FREQUENCY_ONETIME				0
#define NV_SCTASK_FREQUENCY_DAILY				1
#define NV_SCTASK_FREQUENCY_WEEKLY				2
#define NV_SCTASK_FREQUENCY_MONTHLY				3
#define NV_SCTASK_MAXFREQUENCY					3

// month in m_iMonthMonths
#define NV_SCTASK_MONTH_JAN				0x0001
#define NV_SCTASK_MONTH_FEB				0x0002
#define NV_SCTASK_MONTH_MAR				0x0004
#define NV_SCTASK_MONTH_APR				0x0008
#define NV_SCTASK_MONTH_MAY				0x0010
#define NV_SCTASK_MONTH_JUN				0x0020
#define NV_SCTASK_MONTH_JUL				0x0040
#define NV_SCTASK_MONTH_AUG				0x0080
#define NV_SCTASK_MONTH_SEP				0x0100
#define NV_SCTASK_MONTH_OCT				0x0200
#define NV_SCTASK_MONTH_NOV				0x0400
#define NV_SCTASK_MONTH_DEC				0x0800
#define NV_SCTASK_MONTH_ALL				0xFFFF

// day in m_iMonthDays
#define NV_SCTASK_MDAY_01				0x00000001
#define NV_SCTASK_MDAY_02				0x00000002
#define NV_SCTASK_MDAY_03				0x00000004
#define NV_SCTASK_MDAY_04				0x00000008
#define NV_SCTASK_MDAY_05				0x00000010
#define NV_SCTASK_MDAY_06				0x00000020
#define NV_SCTASK_MDAY_07				0x00000040
#define NV_SCTASK_MDAY_08				0x00000080
#define NV_SCTASK_MDAY_09				0x00000100
#define NV_SCTASK_MDAY_10				0x00000200
#define NV_SCTASK_MDAY_11				0x00000400
#define NV_SCTASK_MDAY_12				0x00000800
#define NV_SCTASK_MDAY_13				0x00001000
#define NV_SCTASK_MDAY_14				0x00002000
#define NV_SCTASK_MDAY_15				0x00004000
#define NV_SCTASK_MDAY_16				0x00008000
#define NV_SCTASK_MDAY_17				0x00010000
#define NV_SCTASK_MDAY_18				0x00020000
#define NV_SCTASK_MDAY_19				0x00040000
#define NV_SCTASK_MDAY_20				0x00080000
#define NV_SCTASK_MDAY_21				0x00100000
#define NV_SCTASK_MDAY_22				0x00200000
#define NV_SCTASK_MDAY_23				0x00400000
#define NV_SCTASK_MDAY_24				0x00800000
#define NV_SCTASK_MDAY_25				0x01000000
#define NV_SCTASK_MDAY_26				0x02000000
#define NV_SCTASK_MDAY_27				0x04000000
#define NV_SCTASK_MDAY_28				0x08000000
#define NV_SCTASK_MDAY_29				0x10000000
#define NV_SCTASK_MDAY_30				0x20000000
#define NV_SCTASK_MDAY_31				0x40000000
#define NV_SCTASK_MDAY_ALL				0xFFFFFFFF

// day in m_iWeekDays
#define NV_SCTASK_WDAY_MON				0x01
#define NV_SCTASK_WDAY_TUE				0x02
#define NV_SCTASK_WDAY_WED				0x04
#define NV_SCTASK_WDAY_THU				0x08
#define NV_SCTASK_WDAY_FRI				0x10
#define NV_SCTASK_WDAY_SAT				0x20
#define NV_SCTASK_WDAY_SUN				0x40
#define NV_SCTASK_WDAY_ALL				0xFF

// week in m_iMonthWeekDayPosition
#define NV_SCTASK_WEEK_FIRST			1
#define NV_SCTASK_WEEK_SECOND			2
#define NV_SCTASK_WEEK_THIRD			3
#define NV_SCTASK_WEEK_FOURTH			4
#define NV_SCTASK_WEEK_LAST				5


/////////////////////////////////////////////////////////////////////////////
// NvSchedulerTask class
class NvSchedulerTask
{

public:
		
	// General parameters
	NvMutex ProtectInUse;						// protect locking of task for updating the in use counter and other variables
	NvSession *Session;							// Current session
	NvError Error;								// Error object (used when task fails)
	int InUse;									// In use counter
	int m_iStatus;								// task status
	bool m_bInQueue;							// set to true when the task is in the server queue for processing
	bool m_bImmediateRun;						// Immediate run mode
	std::string m_sName;						// Task name
	std::string m_sDesc;						// Task description
 	std::string m_sAppName;						// Name of the application that owns the task
	std::string m_sUser;						// User name who runs the task
	std::string m_sLanguage;					// Language used when creating the task
	std::string m_sProcTask;					// Task procedure
	std::string m_sProcFailed;					// Failed procedure
	std::string m_sProcSuccess;					// Success procedure
	std::string m_sProcInit;					// Init procedure (same as session_open with no restrictions)
	std::string m_sProcExit;					// Exit procedure
	time_t tLastStartTime;						// Time of last task start
	time_t tLastEndTime;						// Time of last task end
	bool m_bEnable;								// Enable flag
	time_t tStartDate;							// Start date. Only the date is used in the structure.
	time_t tStartTime;							// Start time in day. Only the time is used in the structure.
	int m_iTimeSpan;							// In seconds from 1 to 82799 (1 second to 23h59)
	int m_iFrequencyType;						// Type of frequency Monthly, weekly, daily or one time.

	// Daily repeat mode parameters
	bool m_bRepeatMode;							// Daily repeat mode
	int m_iDelay;								// Delay between each occurence in seconds (from 1 to 82799)
	int m_iOccurences;							// Max number of occurences (starting from 1, 0=infinite)
	int m_iActualOccurence;						// For save only
	time_t tFirstOccurenceStartDate;			// For save only

	// Daily frequency parameters
	int m_iDayEvery;							// Frequency. 1 to 365 days.

	// Weekly frequency parameters
	int m_iWeekEvery;							// Frequency. 1 to 52 weeks
	int m_iWeekDays;							// Days of week. OR combination

	// Monthly frequency parameters
	int m_iMonthMonths;							// Months of year. OR combination
	bool m_bMonthDayBased;						// If true, the m_iMonthDays parameter is taken in care. Otherwise m_iMonthWeekPosition and m_iMonthWeekDay
	int m_iMonthDays;							// Days of month. OR combination
	int m_iMonthWeekDayPosition;				// Week of month. 1 to 5. 1=first and 5=last.
	int m_iMonthWeekDay;						// Day of week. 1 to 7.

	NvSchedulerTask();
	virtual ~NvSchedulerTask();
	
	bool Start(NvCommand *CommandToExecute);
	bool Stop(NvCommand *CommandToExecute);

	void SetError(NvCommand *CommandToExecute);

	
	bool IncInuse();
	void DecInUse();
	int GetInUse();

	std::string GetName();
	std::string GetAppName();
	void SetInQueue(bool InQueue);
	bool GetInQueue();

	// Saves the task parameters to the application registry
	bool Save();

	// Removes the task information from registry
	void Remove();

	// Check that we are in the window time range
	// Doesn't check the day itself (made in CheckFrequency function)
	// The function also retrieves the actual date at 00h00
	// The actual date can be the day before
	// If the time range contains midnigth
	bool CheckTimeRange(time_t *ActualDate);

	// Check that we can run the task today following the frequency mode
	bool CheckFrequency(time_t ActualDate);

	// Check the repeat mode
	// in order to know if the task has ever been ran today
	// and if we can repeat it
	int CheckRepeat(time_t ActualDate);

	// Resets the tFirstOccurenceStartDate and m_iActualOccurence if necessary
	// Saves the task if there is some change
	void ResetRepeatCounter();

	// Run the task
	bool Run();
};



/////////////////////////////////////////////////////////////////////////////
// NvSchedulerTaskList class
class NvSchedulerTaskList : public NvObject
{
private:
	NvSchedulerTask* InternalGetTask(const char *Name);

public:
	std::map<std::string,void *> m_mTasks;			// task list
	std::string m_sAppName;							// Name of the application that owns the task list
	bool bListLoaded;								// Flag to load the task list only on request
	
	NvSchedulerTaskList();
	virtual ~NvSchedulerTaskList();

	// Removes all tasks from the list
	void Clear();
	
	// used to get an existing task
	NvSchedulerTask* GetTask(const char *Name);

	// Load the task list from registry
	bool Load();

	// Creates a new task
	bool CreateTask(const char *Name, const char *Desc, const char *User, const char *Language = NULL);

	// Set task parameters
	bool SetTaskParam(NvCommand *CommandToExecute);

	// used to free an existing task
	// The function decrements the in use counter of the corresponding task
	void FreeTask(const char *Name);
	
	// Remove the given task from the list if not in use
	bool RemoveTask(const char *Name, int WaitTime = 10);

	// Enable or disable a task
	bool EnableTask(const char *Name, bool Enable);

	// Immediatly run the task
	bool RunTask(const char *Name);

	// Checks if a task exists in the list
	bool TaskExist(const char *Name);

	// Get task list info
	bool GetInfo(NvObjTable *TaskInfo, const char *Language = NULL);

	// Get task specific info
	bool GetInfoEx(const char *Name, NvObjIndStringList *TaskInfo, const char *Language = NULL);

	// Get scheduled task to process
	bool GetTasksToProcess(NvObjStringList *TasksToProcess);

	// Verify that tasks are in queue
	bool VerifyTasksInQueue(std::string& AppName, NvSchedulerServer *Server);
};





	
// Scheduler server class
class NvSchedulerServer : public NvTask
{
// Attributes
public:
	NvEvent AskToTerminate;
	bool b_Initialized;
	int m_iMaxTasks;
	int m_iCheckEvery;
	std::list<std::string> m_lTasksToProcess;		// task list to process (FIFO)
	std::set<std::string> m_lTasksInProcess;		// task in process

	NvSchedulerServer();
	void GetTasksToProcess();
	void VerifyTasksInQueue();
	bool IsInProcess(std::string tIdent);
	bool AddInProcess(std::string tIdent);
	bool RemoveInProcess(std::string tIdent);
	int GetInProcessSize();
	void SetMaxTasks(int MaxTasks);
	int GetMaxTasks();
	void SetCheckEvery(int CheckEvery);
	int GetCheckEvery();


// Overridable callbacks
protected:
	virtual int Main();
	virtual void OnException(std::string sInfo);
	virtual bool OnInit();
	virtual void OnCleanup();

public:
	void Stop();

};



// Scheduler client class
class NvSchedulerClient : public NvTask
{
// Attributes
public:
	NvSchedulerServer *Server;
	bool b_Initialized;
	std::string m_sTaskId;				// Task identifier
	NvSchedulerTask *m_tTask;			// Current task
	bool m_bTaskRunning;				// Used for crash condition testing

	NvSchedulerClient();


// Overridable callbacks
protected:
	virtual int Main();
	virtual void OnException(std::string sInfo);
	virtual bool OnInit();
	virtual void OnCleanup();

public:

	// Run the task
	bool RunTask();
};



#endif //_SCHEDULER_H
