
package com.allaire.cfx ;


/**
* Interface to response generated from a <code>{@link CustomTag}</code>.
*
* This interface includes methods for writing output, generating queries,
* and setting variables within the calling page.
*
*  @author J.J. Allaire
*/
public interface Response 
{	
	/**
	* Outputs text back to the user.
	*
	* <P>The following example outputs the value of the DESTINATION
	* attribute: </P>
	*
	* <pre><code>   response.write( &quot;DESTINATION = &quot; + request.getAttribute(&quot;DESTINATION&quot;) ) ;</code></pre>
	*
	* @param output text to output
	*/
	public void write( String output ) ;


	/**
	* Sets a variable in the calling template. If the variable name specified already
	* exists in the template then its value is replaced. If it does not already exist 
	* then a new variable is created. 
	*
	* <P>For example, this code sets the value of a variable
	* named 'MessageSent' based on the success of an operation
	* performed by the custom tag: <P>
	*	
	* <pre><code>   boolean bMessageSent ;
	*
	*   ...attempt to send the message...
	*
	*   if ( bMessageSent == true )
	*   {
	*      response.setVariable( &quot;MessageSent&quot;, &quot;Yes&quot; ) ;
	*   }
	*   else
	*   {
	*      response.setVariable( &quot;MessageSent&quot;, &quot;No&quot; ) ;
	*   } </code></pre>
	*
	* @param name name of variable to set
	* @param value value to set variable to
	*
	* @throws <code>IllegalArgumentException</code> if the <code>name</code> parameter is not a valid CFML variable name
	*/
	public void setVariable( String name, String value ) throws IllegalArgumentException ;


	/**
	* Adds a query to the calling template. This query can
	* then be accessed by CFML tags within the template. 
	*
	* Note that after calling <code>addQuery</code> the query exists 
	* but is empty (i.e. it has 0 rows). To populate the query with data you 
	* should call the <code>{@link Query Query}</code> member functions 
	* <code>{@link Query#addRow addRow}</code> and <code>{@link Query#setData setData}</code>.
	*
	* <P>The following example adds a <code>Query</code> named 'People' to the calling template.
	* The query has two columns ('FirstName' and 'LastName') and 2 rows:</P>
	*
	* <pre><code>   // Create string array with column names (also track columns indexes)
	*   String[] columns = { &quot;FirstName&quot;, &quot;LastName&quot; } ;
	*   int iFirstName = 1, iLastName = 2 ;
	*
	*   // Create a query which contains these columns
	*   Query query = response.addQuery( &quot;People&quot;, columns ) ;
	*					
	*   // Add data to the query
	*   int iRow = query.addRow() ;
	*   query.setData( iRow, iFirstName, &quot;John&quot; ) ;
	*   query.setData( iRow, iLastName, &quot;Smith&quot; ) ;
	*   iRow = query.addRow() ;
	*   query.setData( iRow, iFirstName, &quot;Jane&quot; ) ;
	*   query.setData( iRow, iLastName, &quot;Doe&quot; ) ; </code></pre>
   *
	* @param name name of query to add to the template
	* @param columns column names to be used in the query
	*
	* @return The <code>Query</code> that was added to the template.
	*
	* @throws <code>IllegalArgumentException</code> if the <code>name</code> parameter is not a valid CFML variable name
	*
	* @see Query#addRow		Query.addRow
	* @see Query#setData		Query.setData
	*/
	public Query addQuery( String name, String[] columns ) throws IllegalArgumentException ;


	/**
	* Writes text output into the debug stream. This text is only displayed
	* to the end-user if the tag contains the <code>DEBUG</code> attribute 
	* (you can check for this attribute using the <code>{@link Request#debug Request.debug}</code>
	* member function).
	*
	* <P>The following example checks to see whether the <code>DEBUG</code> attribute
	* is present, and if it is then it writes a brief debug message: </P>
	*
	* <pre><code>   if ( request.debug() )
	*   {
	*      response.writeDebug( &quot;debug info&quot; ) ;
	*   } </pre></code>
	*
	* @param output text to output
	*
	* @see Request#debug Request.debug
	*/
	public void writeDebug( String output ) ;

}


